require 'net/http'
require 'uri'
require 'base64'
require 'json'
require 'cgi'

require 'sketchup.rb'
require 'socket'
require 'set'
require 'date'
require 'digest'
require 'zlib'

if Sketchup.platform == :platform_win
    require 'win32/registry'
    require 'fiddle'
    require 'fiddle/types'
    require 'fiddle/import'
end

#require 'baroquesoftware_vrsketch/ui'
#require 'baroquesoftware_vrsketch/clipboard'

module Baroque
    module VRSketch
        extend self
        attr_accessor :allowed_actions
        attr_accessor :snap_scale
        attr_accessor :teleport_tool
        attr_accessor :model_tracker
        attr_accessor :error_monitor
        attr_accessor :last_used_directory
        attr_accessor :prev_used_directory
    
        MAIN_VERSION = "11.9.4"
        RB_VERSION_ID = 110904
        CLOUD_PAGE = "/cloud.html"
        PUBLIC_CLOUD_SERVER = "https://vrsketch.eu/m/"
        IS_WINDOWS = Sketchup.platform == :platform_win

        # This DIGEST_FILES list is produced by running 'python digest.py' in the MasterServer directory.
        # The files listed after the 'nil' separator are checked later than the first part.
        # If there is already a waiting Unity program listening, typically started manually from the
        # Unity editor in play mode, then the files after 'nil' are not checked at all.
        DIGEST_FILES = [
            ["about.html", "35926f23c763394eb64e06863f441e80"],
            ["BeretGuy16.png", "06657c45b5a1c8024c6d00151a1d779b"],
            ["BeretGuy24.png", "c544cd288a0c9fcee54b0beb9a48a926"],
            ["BeretGuyQ16.png", "42f35a9f68210918ecbe44c9473e0152"],
            ["BeretGuyQ24.png", "d0b153890d30939118f14813a498344a"],
            ["cloud.html", "ecab4b1f0566eee21a343909e874fdfa"],
            ["Cloud.skp", "0075ac7c77a45706d2cdd6e6b298032c"],
            ["Config16.png", "438c0431576a3c82a410bc2abd65b533"],
            ["Config24.png", "a2b10a411c7d6a7dee1025a7a6610036"],
            ["CopyVRloc16.png", "1391de63afc9d4ae74cad6f5693e410a"],
            ["CopyVRloc24.png", "384045c483abe4407c739d1f653c781c"],
            ["ICSharpCode.SharpZipLib.dll", "cc547565785085d77197950305ff88d6"],
            ["vrsketch-sketchlib11.1.0.dll", "5df856ee9b05a8aa5613a8b5d2ee1dcb"],
            nil,  # separator
            ["VRSketch11.1.0.exe", "01a6a085ae717acd6d6c6de1c9fc0df6"],
            ["VRSketch11.1.0_Data/Managed/Assembly-CSharp-firstpass.dll", "98abbda5d0bdf4f4bab55267681c83e6"],
            ["VRSketch11.1.0_Data/Managed/Assembly-CSharp-firstpass.pdb", "1a409aa80233c62102bf1acc51c20e51"],
            ["VRSketch11.1.0_Data/Managed/Assembly-CSharp.dll", "54e2a1520ac69ecb76df86d386d97722"],
            ["VRSketch11.1.0_Data/Managed/Assembly-CSharp.pdb", "4845c18eda6e06da634416c8a4c0f50f"],
            ["VRSketch11.1.0_Data/Managed/I18N.West.dll", "961f486c969e9b2f4708b12f55b340b5"],
            ["VRSketch11.1.0_Data/Managed/I18N.dll", "c90ac617b6e0758d3119961cb312a2f0"],
            ["VRSketch11.1.0_Data/Managed/ICSharpCode.SharpZipLib.dll", "cc547565785085d77197950305ff88d6"],
            ["VRSketch11.1.0_Data/Managed/Mono.Posix.dll", "5cecad569657777c0025e3c01a836d1a"],
            ["VRSketch11.1.0_Data/Managed/Mono.Security.dll", "8fdbc1cbe022a6e5c668b266b5d35c6e"],
            ["VRSketch11.1.0_Data/Managed/Oculus.VR.dll", "aa04c10e88dd7e6bf2234c98549e25dd"],
            ["VRSketch11.1.0_Data/Managed/Oculus.VR.pdb", "f578a6f0156063bb49c2a4f52399e49d"],
            ["VRSketch11.1.0_Data/Managed/SteamVR.dll", "a195284524d8646b48ea6515d2124781"],
            ["VRSketch11.1.0_Data/Managed/SteamVR.pdb", "b63661a7b91e61ae208f1c3a15fd1a39"],
            ["VRSketch11.1.0_Data/Managed/SteamVR_Actions.dll", "edc3f59d49d1d33f643f3cbcfb2c5e05"],
            ["VRSketch11.1.0_Data/Managed/SteamVR_Actions.pdb", "2f5776a6b3520de02a735a90f9b9ae7a"],
            ["VRSketch11.1.0_Data/Managed/System.Configuration.dll", "9b6e727ebdda78aafd8e53cefcc3e8fc"],
            ["VRSketch11.1.0_Data/Managed/System.Core.dll", "cc2deb9fb18b7e16a7ea5754ac30b291"],
            ["VRSketch11.1.0_Data/Managed/System.Runtime.Serialization.dll", "c8a5467ccb541b6fa5b283691f968762"],
            ["VRSketch11.1.0_Data/Managed/System.Security.dll", "24549630886df150504cab15d3fdcf59"],
            ["VRSketch11.1.0_Data/Managed/System.ServiceModel.Internals.dll", "626cf0f349d38c4b35de4fc3cc5747c0"],
            ["VRSketch11.1.0_Data/Managed/System.Xml.dll", "986dda64948a4aa789a378ac28e5279e"],
            ["VRSketch11.1.0_Data/Managed/System.dll", "8148f5992eddfe66e36a67ac34b96bd1"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.AIModule.dll", "6dde798e18b9209d2423f0d5e6ccb13d"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.AIModule.dll.mdb", "b9fb384127cde8458f0724a499f029b4"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.ARModule.dll", "287664afeae25d1a170160970e65e4aa"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.ARModule.dll.mdb", "7319d31d23ad809bb7cdac8c93f3696b"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.AccessibilityModule.dll", "9ba7cc4663f13df546c7ff0a6017c84a"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.AccessibilityModule.dll.mdb", "c2043c9cce9524988b2637c2de79963d"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.AnimationModule.dll", "6d771b39977c344ba3ed706c6c5617df"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.AnimationModule.dll.mdb", "6b2d6beaf27c8023cd5df59d630849aa"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.AssetBundleModule.dll", "66e06162d7609e924a0bc01509096864"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.AssetBundleModule.dll.mdb", "699e1561e8f189c4c14b8f6acaf9b44a"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.AudioModule.dll", "bb34ea187f951043986a9a4bfb7467aa"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.AudioModule.dll.mdb", "4cac46497a3dc00a86f96fb0c1fa37ea"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.ClothModule.dll", "db2dffd1bf22cdce45267be7430dac88"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.ClothModule.dll.mdb", "8760c30f36526b0a9fa7f78f789d4b92"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.ClusterInputModule.dll", "924f7c4ec0f2fb9854f0142c68358903"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.ClusterInputModule.dll.mdb", "f5d22f8d91d63ed541f2408ab096fc3c"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.ClusterRendererModule.dll", "7a3a75e4a9ffeb29f8806d9b9e25d654"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.ClusterRendererModule.dll.mdb", "1a91d1fb4cd783282f97f715b6b38e9a"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.CoreModule.dll", "b5edd8d6e80c4950b810146dab8382fc"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.CoreModule.dll.mdb", "9801b62697babb2a653ed985b8afcd8c"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.CrashReportingModule.dll", "902616604bdb70439b2d1028de41417d"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.CrashReportingModule.dll.mdb", "0ce5827a7516c4dddda2f141c5b37552"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.DirectorModule.dll", "1f7506c85af47849eb4fad2a03de0d14"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.DirectorModule.dll.mdb", "7f6b1760dd7af242f2e3e7ea62749adb"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.FileSystemHttpModule.dll", "492fcd886a7aebe47178af4542ff2a0d"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.FileSystemHttpModule.dll.mdb", "45f0723a555461b9dcb73d6ffa848964"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.GameCenterModule.dll", "4e335b0f0d0f165e40aed20e4f45a923"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.GameCenterModule.dll.mdb", "16c4b1463f5a04693d4431304951c4fd"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.GridModule.dll", "f5574fd749aa97240e9630bdb9acafe8"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.GridModule.dll.mdb", "fd1775eaae2ffb37905c9c48ab19bba3"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.HotReloadModule.dll", "1b6a57580299df1480c5e09381de5c28"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.HotReloadModule.dll.mdb", "b9df3782b50d7111943b15176e551299"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.IMGUIModule.dll", "5ed74d6bfd66bfc7cc6fa283f5c3eb1d"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.IMGUIModule.dll.mdb", "619e6b80e68b3ae0a68c3651f4af8b0a"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.ImageConversionModule.dll", "812941240f9dc642b0f611381bbed630"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.ImageConversionModule.dll.mdb", "7ee0dd60c4f4199947c1296742e25162"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.InputModule.dll", "325f0540698f922db4f72dafcbf28738"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.InputModule.dll.mdb", "59af7fb41ae5b7669a11a01be12ec807"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.JSONSerializeModule.dll", "cc02b9d28d7f99a0837aa6ec93707484"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.JSONSerializeModule.dll.mdb", "6d101ff4540380d9f143072aa8d465ff"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.LocalizationModule.dll", "3e2faf37abd1b96bd7e2e4fb046c589c"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.LocalizationModule.dll.mdb", "58a8e3fa16e094fa1a8f7e98d8c92418"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.ParticleSystemModule.dll", "53292d7c13d79dc4b4da373d59bbb7bc"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.ParticleSystemModule.dll.mdb", "b4e539bfffa833266c124c2459ec242e"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.PerformanceReportingModule.dll", "ed2e6f7acc351a2c2fbe2e4cfe40148c"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.PerformanceReportingModule.dll.mdb", "8966b6d20f678913be2dcf053dfda1f6"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.Physics2DModule.dll", "b1eb87afe99f53fd194ec4e004b8d351"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.Physics2DModule.dll.mdb", "986dc6c17ea59e667f427f3aa7b898a7"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.PhysicsModule.dll", "938469e1f429c1c2eb5b8302f6450ccc"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.PhysicsModule.dll.mdb", "9ab3faa3727f2dc46241535ec45ed1a2"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.ProfilerModule.dll", "bd4194bf7316a489cfb3cc53cb98aa69"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.ProfilerModule.dll.mdb", "012e61be3ac847b7880672fce48357e1"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.ScreenCaptureModule.dll", "afabbc5754adfc1458a2bdaedb6459f5"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.ScreenCaptureModule.dll.mdb", "a8bea3eb09071fb029620eaea60c0d39"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.SharedInternalsModule.dll", "e664114ec481c0794b9991a79b48e606"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.SharedInternalsModule.dll.mdb", "9a1f3a9d28643f96ca1048095843cfee"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.SpriteMaskModule.dll", "49e9e88244404f88a9fe144522ef798c"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.SpriteMaskModule.dll.mdb", "5a66bcf299ef613f523ce5b7ed77454e"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.SpriteShapeModule.dll", "178e3e6cca44573ae8d3b05faa3ddcf0"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.SpriteShapeModule.dll.mdb", "f36ed1547c62c9a8935be0cd61a4f1f7"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.StreamingModule.dll", "b3c1323d1c01ca13686e6b8e85cc3c2e"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.StreamingModule.dll.mdb", "d4b9a217b8a3d67d1437fd41cec06fd4"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.StyleSheetsModule.dll", "9b80b80e064907de3d6b15198912314f"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.StyleSheetsModule.dll.mdb", "724340c36a211f965dbfa54e809c56fd"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.SubstanceModule.dll", "5c898e77dd6804c3e5e827c0f36430cd"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.SubstanceModule.dll.mdb", "e18ba5383c16a99145aae1f13d1001b4"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.TLSModule.dll", "d032f31a4472d27dada7f9a3e368a5e2"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.TLSModule.dll.mdb", "ecf503aeb331b3ddb15fb476cdd2248e"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.TerrainModule.dll", "f138e72bf0a172e364e88b37f2036f77"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.TerrainModule.dll.mdb", "9a8b716d41705754182630e669b7e3d1"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.TerrainPhysicsModule.dll", "29578d6c8c23d5a6bab6c2f92df4305c"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.TerrainPhysicsModule.dll.mdb", "743a67322915c5983a8df14b08d3c709"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.TextCoreModule.dll", "a397bd3801ddc7327a1a71977187b65c"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.TextCoreModule.dll.mdb", "4942645e2925095b8bdf9fffb4b794c7"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.TextRenderingModule.dll", "6aa071ee68ad4935bc5316f9114cd0c0"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.TextRenderingModule.dll.mdb", "89a57636e656a834cc657507be923b5c"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.TilemapModule.dll", "433208a8dd424d526896a365a19a7969"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.TilemapModule.dll.mdb", "8d22d5306131b50f1cfe38f5b2ade53b"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UI.dll", "998563724e55b7844be2d264267638c7"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UI.dll.mdb", "187f0995821f5277b01a9ae6d2bcab41"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UIElementsModule.dll", "5272dfa938d788a1a12cf11042f2c104"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UIElementsModule.dll.mdb", "56c5fed9fe833088fef5b7559c033946"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UIModule.dll", "b0167b9f42d01817d6668bf139cdf9bc"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UIModule.dll.mdb", "9a13c4553771d4aab40e49ef1bfb02cd"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UNETModule.dll", "ae15b95b6f2845f9830f25039c7f732c"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UNETModule.dll.mdb", "9737c77eebf3095c03c1eb56a4cb4d1c"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UmbraModule.dll", "e2cdfd32cff17f8ceb32edd5c37b547f"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UmbraModule.dll.mdb", "724c8d0f511e97a1cdb1e627d3808cbe"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UnityAnalyticsModule.dll", "2c98aeb2fbd0cca2a5afcb07a1b2ad35"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UnityAnalyticsModule.dll.mdb", "080f67989621499921ac69cf70fe11a8"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UnityConnectModule.dll", "eac6ae968afcbbfb7d21c56598bc9a47"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UnityConnectModule.dll.mdb", "07d26b8ba8535c3351eb51166350357c"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UnityTestProtocolModule.dll", "2256f5821b6b5e63bb049cd2f15c8205"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UnityTestProtocolModule.dll.mdb", "1cf0d4b699318eab60a5d9aeb1403aa0"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UnityWebRequestAssetBundleModule.dll", "203f0f29b0ef451f48485b849525fdf4"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UnityWebRequestAssetBundleModule.dll.mdb", "2089cdde73b5b1cab2612c2377d3a176"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UnityWebRequestAudioModule.dll", "019a1ac2d23a9ec5b93e6b667fdee91f"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UnityWebRequestAudioModule.dll.mdb", "825ea516c3ab759abb3f7f1ccf4d52b1"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UnityWebRequestModule.dll", "04c4ed0ffd12f87785de1ba1f7a9ccfd"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UnityWebRequestModule.dll.mdb", "254a6c33e7f4a53e6115363bd3b96f3f"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UnityWebRequestTextureModule.dll", "dac3f358a83501cf53215c7808571ed3"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UnityWebRequestTextureModule.dll.mdb", "19ccd4da5b79515e5a1c58c99a84bae6"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UnityWebRequestWWWModule.dll", "dcf4a73f2d00cacf505e89e00761b8f5"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.UnityWebRequestWWWModule.dll.mdb", "997ff14ac5e4a85380ebcb7875675e48"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.VFXModule.dll", "5faaa7321958bb51fdff7e591d20c7e4"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.VFXModule.dll.mdb", "894c8427376e6ecf6d16d55a4578bcc1"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.VRModule.dll", "413d54a9adbcdbafab862826db6b5a77"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.VRModule.dll.mdb", "f54b11b431d9127a0f386e8180856b11"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.VehiclesModule.dll", "62259324b774ea2a0fb4fec2acb37edc"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.VehiclesModule.dll.mdb", "6da9fecf90a93ffb273116b37ead35d1"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.VideoModule.dll", "294a3edfa03d854f06df529a353d222e"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.VideoModule.dll.mdb", "b883e3d3f4f53b8aeb6d3083b16e35ce"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.WindModule.dll", "2b68e4c4631fe62d4c32faee8dd9a67e"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.WindModule.dll.mdb", "152d5c3b5fa60c8533dd085070cd52cc"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.XRModule.dll", "faa3bc1a78008fba65a05279dc952b88"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.XRModule.dll.mdb", "3c549be96b79e609bf4e10d74b4cc074"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.dll", "b3cdec2df435f6508c8e44b6fc1c5ec8"],
            ["VRSketch11.1.0_Data/Managed/UnityEngine.dll.mdb", "b485ed2debc51013237cf1d69a5c23d5"],
            ["VRSketch11.1.0_Data/Managed/Valve.Newtonsoft.Json.dll", "bacf01c5e0270cbaee0d0b2dc602a49f"],
            ["VRSketch11.1.0_Data/Managed/mscorlib.dll", "2004887de7a92bf0c73539a8363971f3"],
            ["VRSketch11.1.0_Data/Managed/websocket-sharp.dll", "537b06511f5f8c9676b7b72e57c88c43"],
            ["VRSketch11.1.0_Data/Plugins/AudioPluginDissonance.dll", "47c5b7e536b02d5a02425667e5099aab"],
            ["VRSketch11.1.0_Data/Plugins/AudioPluginOculusSpatializer.dll", "f7d75248a62e56fdb6ee058c28da46a1"],
            ["VRSketch11.1.0_Data/Plugins/OVRPlugin.dll", "c49908a16192ec0dbaf13d3856b39201"],
            ["VRSketch11.1.0_Data/Plugins/VideoCaptureLib.dll", "35064efb93f0d64b78b156a4c4b8771e"],
            ["VRSketch11.1.0_Data/Plugins/openvr_api.dll", "e85622ffa2fc95b618f86a5b8e2d15b8"],
            ["VRSketch11.1.0_Data/Plugins/opus.dll", "7e5bcd6c49309f707ce6913d435d9b99"],
            ["VRSketch11.1.0_Data/Resources/unity default resources", "a86ef2c6f1996bbd86e9cf3c5a1f43f1"],
            ["VRSketch11.1.0_Data/Resources/unity_builtin_extra", "73a76e03a4c40d898a70bca7a6511106"],
            ["VRSketch11.1.0_Data/StreamingAssets/FFmpegOut/Windows/README.txt", "ad80045985c401b4ab9967fb2e16acf1"],
            ["VRSketch11.1.0_Data/StreamingAssets/FFmpegOut/Windows/ffmpeg.exe", "cef0e7343b30e05046894acd0040d226"],
            ["VRSketch11.1.0_Data/StreamingAssets/FFmpegOut/Windows/gpl-2.0.txt", "b234ee4d69f5fce4486a80fdaf4a4263"],
            ["VRSketch11.1.0_Data/StreamingAssets/FFmpegOut/Windows/libwinpthread-1.dll", "af6f46fea27266c77a9612d4d9dabfd3"],
            ["VRSketch11.1.0_Data/StreamingAssets/GrammarNumbers.xml", "8f952f496edf9b6a827bb1bc3400ee1b"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/actions.json", "ca3fccff083c42913256709e9c4c6360"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/actions.json~", "df211261b664ae2bd2474e08ea63596b"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/binding_holographic_hmd.json", "cc92067caf16f5183963e95f0b0c21e9"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/binding_index_hmd.json", "984ab237c4d0880b8e616fc22262a560"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/binding_rift.json", "82a9d571a31b41d0dc35d2479a163099"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/binding_vive.json", "c9e11e72dcd30b40be11cc22e22b9d4e"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/binding_vive_cosmos.json", "3ff84953a818c831cf05c3c4fe71e382"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/binding_vive_pro.json", "114bb073523c40ad6a1675840022d439"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/binding_vive_tracker_camera.json", "5898b2a80822f6352347b946516b53e5"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/bindings_holographic_controller.json", "770ae31b721683519541074cca8c4256"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/bindings_holographic_controller.json~", "ff1e5a0510934a2e0d53d18ff445320a"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/bindings_knuckles.json", "2bd6e472ac34a2a50c2603411055152f"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/bindings_logitech_stylus.json", "062ff08d68766a72e6605735196b6820"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/bindings_logitech_stylus.json~", "062ff08d68766a72e6605735196b6820"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/bindings_oculus_touch.json", "485a7045c4489eb0d12e80aa461af184"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/bindings_vive_controller.json", "2553e41f34afaca3b15c1b10d4f5d8ba"],
            ["VRSketch11.1.0_Data/StreamingAssets/SteamVR/bindings_vive_cosmos_controller.json", "25fe0814b4d578bbb9996471a1ad3d1c"],
            ["VRSketch11.1.0_Data/app.info", "bcaab484a5fd728af08397e70a597861"],
            ["VRSketch11.1.0_Data/boot.config", "5958bc393e4f777157e78226d4fea5da"],
            ["VRSketch11.1.0_Data/globalgamemanagers", "7d57a2ee05748ec0498dd2b35657ce99"],
            ["VRSketch11.1.0_Data/globalgamemanagers.assets", "e8668625d48054c52cbe211f84dd0bc4"],
            ["VRSketch11.1.0_Data/level0", "91296a301d7f9efed99e29caefe35655"],
            ["VRSketch11.1.0_Data/level1", "b51367491f7539fd3daf695594b0f414"],
            ["VRSketch11.1.0_Data/level1.resS", "34f842a07b4e0240e7b63e47077d00f3"],
            ["VRSketch11.1.0_Data/resources.assets", "a88a1a715aba3832d395a516ab341b87"],
            ["VRSketch11.1.0_Data/resources.assets.resS", "fac72a9893c53d6c0a897b5b3c56b8fb"],
            ["VRSketch11.1.0_Data/sharedassets0.assets", "1753738722a15963b326dc6f638e6fbc"],
            ["VRSketch11.1.0_Data/sharedassets0.assets.resS", "395e46152b4b82732a0dc12f631ffb9d"],
            ["VRSketch11.1.0_Data/sharedassets1.assets", "e1f725faae1e6e99318ac88228f495f7"],
            ["VRSketch11.1.0_Data/sharedassets1.assets.resS", "ccfe94ff9842fc915401e64be564c918"],
            ["MonoBleedingEdge/EmbedRuntime/MonoPosixHelper.dll", "1fe695edcad867c3c155b9637d457b20"],
            ["MonoBleedingEdge/EmbedRuntime/mono-2.0-bdwgc.dll", "fc6789679b8527d2b988e31a7fd20de9"],
            ["MonoBleedingEdge/etc/mono/2.0/Browsers/Compat.browser", "0d831c1264b5b32a39fa347de368fe48"],
            ["MonoBleedingEdge/etc/mono/2.0/DefaultWsdlHelpGenerator.aspx", "f7be9f1841ff92f9d4040aed832e0c79"],
            ["MonoBleedingEdge/etc/mono/2.0/machine.config", "cad24142abba464dd90777c3d347ef88"],
            ["MonoBleedingEdge/etc/mono/2.0/settings.map", "22c818a23169e12bd3c8587b6394c731"],
            ["MonoBleedingEdge/etc/mono/2.0/web.config", "2b6303c4f12762b71051db6e947f90a4"],
            ["MonoBleedingEdge/etc/mono/4.0/Browsers/Compat.browser", "0d831c1264b5b32a39fa347de368fe48"],
            ["MonoBleedingEdge/etc/mono/4.0/DefaultWsdlHelpGenerator.aspx", "f7be9f1841ff92f9d4040aed832e0c79"],
            ["MonoBleedingEdge/etc/mono/4.0/machine.config", "24c866ce8037fcdca2287234eddff637"],
            ["MonoBleedingEdge/etc/mono/4.0/settings.map", "ba17ade8a8e3ee221377534c8136f617"],
            ["MonoBleedingEdge/etc/mono/4.0/web.config", "b127480ee9f0b8dab6a3f73ad79dd332"],
            ["MonoBleedingEdge/etc/mono/4.5/Browsers/Compat.browser", "0d831c1264b5b32a39fa347de368fe48"],
            ["MonoBleedingEdge/etc/mono/4.5/DefaultWsdlHelpGenerator.aspx", "f7be9f1841ff92f9d4040aed832e0c79"],
            ["MonoBleedingEdge/etc/mono/4.5/machine.config", "0869544722561f5aff0eefc83fc7b001"],
            ["MonoBleedingEdge/etc/mono/4.5/settings.map", "ba17ade8a8e3ee221377534c8136f617"],
            ["MonoBleedingEdge/etc/mono/4.5/web.config", "08101241b15b53ef0ab908f6d388881f"],
            ["MonoBleedingEdge/etc/mono/browscap.ini", "378be809df7d15aac75a175693e25fbb"],
            ["MonoBleedingEdge/etc/mono/config", "3f09772a4f6f3d1932918169d2050e2b"],
            ["MonoBleedingEdge/etc/mono/mconfig/config.xml", "f34b330f20dce1bdcce9058fca287099"],
            ["UnityCrashHandler64.exe", "88bb39c750a2cc85ca2993044d5374f9"],
            ["UnityPlayer.dll", "0e6cbc4b753b20fd6977a8d5247a83c7"],
            ["WinPixEventRuntime.dll", "e697f3f1b02990cc4cfe6f666b40dc5d"],
        ]

        EPSILON_VECTOR = Geom::Vector3d.new(0.003, 0.003, 0.003)      # in inches
        EPSILON_LENGTH = 0.001.to_l

        class ErrorMonitor
            ERROR_URL = "http://vrsketch.eu:8925/error"
            def monitor
                begin
                    yield
                rescue Exception => e
                    if ((@already_sent_once == nil) && (!e.to_s.end_with?(" - send(2)", " - recvfrom(2)")))
                        @already_sent_once = true
                        begin
                            tb = e.to_s + "\n"
                            tb += "PROPERTY version: " + RB_VERSION_ID.to_s + "\n"
                            begin
                                mydir = File.dirname(__FILE__)
                                head = File.read(mydir + "/../../.git/HEAD").split(/\s+/)
                                if head[0] == "ref:"
                                    head = File.read(mydir + "/../../.git/" + head[1]).split(/\s+/)
                                end
                                tb += ".GIT HEAD: " + head[0] + "\n"
                            rescue
                            end
                            if e.backtrace != nil
                                e.backtrace.each { |line| tb += "\n" + line.to_s }
                            end
                            Net::HTTP.post_form(URI(ERROR_URL), {tb: tb})
                            puts("---The following (first) error was automatically reported---")
                        rescue
                        end
                    end
                    raise e
                end
            end
        end

        def initialize()
            @initialized = false
            @error_monitor = ErrorMonitor.new
            @model_tracker = nil
            @waiting_sock = nil
            @allowed_actions = [1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0]
            @debug_auto_reconnect = nil
            @config_dialog = nil
            @license_dialog = nil
            @cloud_server = "https://vrsketch.eu"
            @upload_vrsketch_dialog = nil
            @about_dialog = nil
            @snap_scale = 22

            @all_dialogs = {}

            read_config
            debug_autoreconnect
        end


        if Baroque::VRSketch::IS_WINDOWS
            module Kernel32
                extend Fiddle::Importer
                dlload 'kernel32.dll'
                include Fiddle::Win32Types
                PROCESS_INFORMATION = struct [
                  "HANDLE hProcess",
                  "HANDLE hThread",
                  "DWORD  dwProcessId",
                  "DWORD  dwThreadId"
                ]
                extern 'DWORD GetCurrentThreadId()'
                extern 'BOOL CreateProcessA(LPCSTR, PVOID, PVOID, PVOID, BOOL, DWORD, PVOID, PVOID, const char *, struct PROCESS_INFORMATION *)'
                extern 'BOOL CloseHandle(HANDLE)'
                extern 'BOOL GetExitCodeProcess(HANDLE, DWORD *)'
                extern 'HANDLE LoadLibraryExW(PVOID, HANDLE, DWORD)'
            end

            module User32
                extend Fiddle::Importer
                dlload 'user32.dll'
                include Fiddle::Win32Types
                extern 'HWND GetAncestor(HWND, UINT)'
                #extern 'BOOL EnumThreadWindows(DWORD, WNDENUMPROC, LPARAM)'
                extern 'BOOL EnumThreadWindows(DWORD, PVOID, PVOID)'
            end

            def hack_prepare_open_group()
                require 'Win32API'
                @hack_SetForegroundWindow = Win32API.new("user32.dll", "SetForegroundWindow", ['L'], 'I')

                # the rest is only used in SketchUp <= 2019
                return if Sketchup.version_number >= 2000000000

                @hack_GetMenu = Win32API.new("user32.dll", "GetMenu", ['L'], 'L')
                @hack_GetMenuItemCount = Win32API.new("user32.dll", "GetMenuItemCount", ['L'], 'I')
                @hack_GetMenuItemID = Win32API.new("user32.dll", "GetMenuItemID", ['L', 'I'], 'I')
                @hack_GetSubMenu = Win32API.new("user32.dll", "GetSubMenu", ['L', 'I'], 'L')
                @hack_SendMessage = Win32API.new("user32.dll", "SendMessage", ['L', 'I', 'L', 'L'], 'L')

                # the fiddle parts are based on https://github.com/SketchUp/testup-2/blob/master/src/testup/win32.rb
                thread_id = Kernel32.GetCurrentThreadId()
                main_hwnd = 0

                enumWindowsProc = Fiddle::Closure::BlockCaller.new(Fiddle::TYPE_INT,
                  [Fiddle::TYPE_VOIDP, Fiddle::TYPE_VOIDP]) { |hwnd, lparam|
                    hwnd = User32.GetAncestor(hwnd, 3)   # GA_ROOTOWNER
                    hmenu = @hack_GetMenu.call(hwnd)
                    next 1 if hmenu == 0
                    heditmenu = @hack_GetSubMenu.call(hmenu, 1)
                    next 1 if heditmenu == 0
                    menulength = @hack_GetMenuItemCount.call(heditmenu)
                    contextmenu = @hack_GetSubMenu.call(heditmenu, menulength - 1)
                    next 1 if contextmenu == 0
                
                    main_hwnd = hwnd
                    next 0
                }
                User32.EnumThreadWindows(thread_id, enumWindowsProc, 0)

                UI.messagebox("Cannot get the top-level SketchUp window's Edit menu.  Opening groups in VR will not work") if main_hwnd == 0
                @hack_hwnd = main_hwnd
            end

            def hack_SetForegroundWindow()
                if connected
                    if model_tracker.hwnd != nil
                        @hack_SetForegroundWindow.call(model_tracker.hwnd)
                    end
                end
            end
        else
            # OSX
            def hack_prepare_open_group()
            end

            def hack_SetForegroundWindow()
            end
        end

        def hack_open_group(model, grp)
            if (grp.is_a? Sketchup::Group) || (grp.is_a? Sketchup::ComponentInstance)
                if model.active_entities.member? grp
                    
                    # detect locked component.  For now we can't edit them easily from VR Sketch
                    if grp.locked?
                        return {cmd: "text_on_controller", name: "Locked"}
                    end
                    if grp.definition.instances.any? { |grp1| grp1.locked? }
                        return {cmd: "text_on_controller", name: "Other\nLocked"}
                    end

                    model.selection.clear
                    if Sketchup.version_number >= 2000000000
                        path = model.active_path
                        path = [] if path == nil
                        model.active_path = path + [grp]
                    elsif Baroque::VRSketch::IS_WINDOWS
                        model.selection.add(grp)

                        hwnd = @hack_hwnd
                        hmenu = @hack_GetMenu.call(hwnd)
                        @hack_SendMessage.call(hwnd, 0x0116, hmenu, 0)
                        heditmenu = @hack_GetSubMenu.call(hmenu, 1)
                        @hack_SendMessage.call(hwnd, 0x0117, heditmenu, 1)
                        menulength = @hack_GetMenuItemCount.call(heditmenu)
                        contextmenu = @hack_GetSubMenu.call(heditmenu, menulength - 1)
                        id = @hack_GetMenuItemID.call(contextmenu, 0)
                        # at this point we should have got in 'id' the correct id.  Very occasionally,
                        # something goes wrong and we end up with a negative value (-1 or 0, I guess).
                        # In that case, do nothing and let the VR user try again...
                        if id > 0
                            Sketchup.send_action(id)
                        else
                            puts("open_close_group: failed to locate the menu entry to click.  Try again..?")
                        end
                    else
                        warn("OS X: you need SketchUp 2020 in order to open groups from VR. The Ruby API of SketchUp 2019 (and older) doesn't support this operation.")
                    end
                else
                    puts("open_close_group: group not in active_entities")
                end
            else
                puts("open_close_group: not a group id")
            end
            return nil
        end

        def sendall(s, data)
            return if s == nil
            while data.bytesize > 0
                count = s.send(data, 0)
                if count == 0
                    throw "disconnected?"
                end
                data = data.byteslice(count, data.bytesize - count)
            end
        end


        CSS = 'body { font-family: Arial, Helvetica, sans-serif; background-color: "f8f8ff"; }
                    p { margin: 0.3em; }
                    p.vskip { margin: 0.5em; }
                    p.small { font-size: 100%; }
                    a.h { color: black; text-decoration:none; }
                    span.slightlysmall { font-size: 90%; }
                    form { border-style: outset; border-color: white; border-width: 1;
                            margin-left: 0.5em; margin-right: 0.5em; padding: 0.4em; }
                    input.text { height: 2em; }
					
					.button {
						background-color: #4CAF50;
						border: none;
						color: white;
						padding: 15px 25px;
						text-align: center;
						font-size: 16px;
						cursor: pointer;
					}
					.button:hover {
						background-color: green;
					}'

        def fill_template(content, h)
            h.each do |k, v|
                # may contain the template several times
                content = content.gsub("{{ " + k.to_s + " }}", v)
            end
            return content
        end

        def get_html(name, template)
            mydir = File.dirname(__FILE__)
            f = open(mydir + "/html/" + name + '.html')
            res = f.read
            template[:style] = CSS
            res = fill_template(res, template)
            f.close
            return res
        end

        # public api

        def show_dialog(name, title, width, height, template: {})
            if @all_dialogs[name] != nil
                @all_dialogs[name].close
                @all_dialogs[name] = nil
            end

            dialog = UI::WebDialog.new(title, true,
                "za.co.VRSketch_" + name, width, height, 200, 250, true)
            
            dialog.set_html(get_html(name, template))

            dialog.show
            @all_dialogs[name] = dialog
            return dialog
        end

        if Baroque::VRSketch::IS_WINDOWS
            module Clipboard

                CF_UNICODETEXT = 13

                require 'fiddle'
                require 'fiddle/types'
                require 'fiddle/import'

                module User32
                    extend Fiddle::Importer
                    dlload 'user32.dll'
                    include Fiddle::Win32Types
                    extern 'BOOL OpenClipboard(HWND)'
                    extern 'BOOL CloseClipboard()'
                    extern 'HANDLE GetClipboardData(UINT)'
                end

                module Kernel32
                    extend Fiddle::Importer
                    dlload 'kernel32'
                    include Fiddle::Win32Types
                    extern 'void *GlobalLock(HANDLE)'
                    extern 'BOOL GlobalUnlock(HANDLE)'
                end

                # see http://www.codeproject.com/KB/clipboard/archerclipboard1.aspx
                def self.paste_latin1()
                    data = ""
                    if 0 != User32.OpenClipboard(0)
                        hclip = User32.GetClipboardData(CF_UNICODETEXT)
                        if hclip && 0 != hclip
                            ptr = Kernel32.GlobalLock(hclip)
                            # pff pff pff
                            i = 0
                            lst = []
                            while ptr[i] != 0
                                if ptr[i + 1] != 0
                                    lst.clear
                                    break
                                end
                                lst.push(ptr[i])
                                i += 2
                            end
                            Kernel32.GlobalUnlock(hclip)
                            data = lst.pack("C*")
                        end
                        User32.CloseClipboard()
                    end
                    return data
                end
            end
        end

        # ----- Class ModelTracker -----
        #
        # This class holds everything needed to track one model.  As we're only supporting
        # Windows for now, there can only be one instance of it at the same time.
        #

        class ModelTracker
            attr_accessor :model
            attr_accessor :hwnd
            attr_accessor :tutorial
            attr_accessor :on_quest
            attr_accessor :point_cloud_base_path
            attr_accessor :sock
            attr_accessor :notify_closing_queue

            def initialize(model, sock, on_quest)
                @model = model
                @sock_buffer = ''
                if sock.class == Array
                    @sock_buffer = sock[1]
                    sock = sock[0]
                end
                @sock = sock
                @on_quest = on_quest
                @server_active = false

                @pending_updates = nil       # dict {cdef: [list-of-entities-or-nil-for-all]}
                @pending_removes = nil       # dict {cdef: [list-of-id-of-deleted-entity]}
                @pending_removes_cdef_id = nil    # list [id-of-deleted-cdef]
                
                @model_observer = MyModelObserver.new(self)
                @vertex_observer = MyVertexObserver.new(self)
                @group_observer = MyGroupObserver.new(self)
                @entities_observer = MyEntitiesObserver.new(self)
                @definitions_observer = MyDefinitionsObserver.new(self, model)
                @materials_observer = MyMaterialsObserver.new(self)
                @shadow_info_observer = MyShadowInfoObserver.new(self)
                @layers_observer = MyLayersObserver.new(self, @group_observer)
                @units_options_observer = MyUnitsOptionsObserver.new(self)
                @rendering_options_observer = MyRenderingOptionsObserver.new(self)
                @pages_observer = MyPagesObserver.new(self)
                
                @active_path = @model.active_path
                @active_path = [] if @active_path == nil

                @warnings = {}
                @keep_selection = nil    # a list of entities OR a list of ids
                @pending_mat_msgs = []
                @undo_history = []       # { msg:, optionally oldsel: }
                @undone_operations = 0
                @general_pending_ops = {}

                @connected_to_localhost = ((@sock.respond_to? :recv_nonblock) && !@on_quest)
                if @connected_to_localhost
                    @timer = UI.start_timer(0.1, true) { Baroque::VRSketch.error_monitor.monitor { run_timer } }
                end
                @sock_edit_cmd = []
                @cancel_wait_id = 0
                @frame_change_speed = 0

                @tracker_initialized = Time.now.to_f
            end

            def send_model_now()
                return if Baroque::VRSketch.model_tracker != self
                @server_active = true
				begin
					upload_whole_model
				rescue
					close_socket
					Baroque::VRSketch.model_tracker = nil
					raise
				end
                @model.add_observer(@model_observer)
                @model.entities.add_observer(@entities_observer)
                @model.definitions.add_observer(@definitions_observer)
                @model.materials.add_observer(@materials_observer)
                @model.shadow_info.add_observer(@shadow_info_observer)
                @model.layers.add_observer(@layers_observer)
                @model.options["UnitsOptions"].add_observer(@units_options_observer)
                @model.rendering_options.add_observer(@rendering_options_observer)
                @model.pages.add_observer(@pages_observer)

                @model.select_tool(Baroque::VRSketch.teleport_tool)
            end

            def connected(mode)
                if mode == "Server"
                    return false if !@server_active
                end
                return @sock != nil
            end

            def warn(html_text, title="Warning", height=200, &block)
                if !(@warnings.has_key? html_text)
                    html_details = ''
                    details_msg = nil
                    if block
                        html_details = '&nbsp;&nbsp;&nbsp;
                        <button type="button" onclick="window.location.href=\'skp:details\'">Details...</button>
                        '
                        details_msg = block.call
                    end
                    Baroque::VRSketch.delay(0.15) {
                        dlg = UI::WebDialog.new(title, true, nil, 500, height, 620, 200)
                        dlg.set_html('<html><head><style>
                                body { font-family: Arial, Helvetica, sans-serif; background-color: "f8f8ff"; }
                                form { border-style: outset; border-color: white; border-width: 1;
                                        margin-left: 0.5em; margin-right: 0.5em; padding: 0.35em; }
                            </style></head><body>&nbsp;
                            <form>' + html_text + '</form>&nbsp;
                            <button type="button" onclick="window.location.href=\'skp:close\'">Close</button>
                            ' + html_details + '</body></html>')
                        dlg.add_action_callback("close") { |dlg, param|
                            Baroque::VRSketch.error_monitor.monitor {
                                dlg.close
                            }
                        }
                        if details_msg
                            dlg.add_action_callback("details") { |dlg, param|
                                Baroque::VRSketch.error_monitor.monitor {
                                    UI.messagebox(details_msg)
                                }
                            }
                        end
                        dlg.show
                    }
                    @warnings[html_text] = true
                end
            end

            # ----- SERIALIZING ENTITIES -----

            def _enable_updates()
                if @pending_updates == nil
                    #puts("===== delay...")
                    Baroque::VRSketch.delay(0.02) { flush_transaction_changes }
                    @pending_updates = {}
                    @pending_removes = {}
                #else
                #    puts(@pending_updates.length.to_s + " updates already")
                end
            end

            def _enable_updates_cdef(cdef)
                _enable_updates
                if ((cdef != @model) && !(cdef.is_a? Sketchup::ComponentDefinition))
                    throw ({BadCDef: cdef}.to_s)
                end
                if !(@pending_updates.has_key?(cdef))
                    @pending_updates[cdef] = []
                end
                return @pending_updates[cdef]
            end

            def push_update(obj)
                return if obj.deleted?
                if ((obj.parent == @model) || (obj.parent.is_a? Sketchup::ComponentDefinition))
                    upd = _enable_updates_cdef(obj.parent)
                    #puts({:push_update => obj})
                    upd.push(obj) if upd != nil
                end
            end

            def push_removal_entity(id, cdef)
                _enable_updates_cdef(cdef)
                #puts({:push_removal_entity => [id, cdef]})
                if !(@pending_removes.has_key?(cdef))
                    @pending_removes[cdef] = []
                end
                @pending_removes[cdef].push(id)     # cdef is always in @pending_updates too
            end

            def push_removal_cdef_id(cdef, cdef_id)
                _enable_updates
                #puts({:push_removal_cdef_id => cdef_id})
                @pending_removes_cdef_id = [] if @pending_removes_cdef_id == nil
                @pending_removes_cdef_id.push(cdef_id)
                @pending_updates.delete(cdef)   # may show up again later if another cdef with the same id appears
            end

            def push_all_entities_from(cdef)
                _enable_updates
                #puts({:push_all_entities_from => cdef})
                @pending_updates[cdef] = nil
            end

            def push_update_whole_current_group()
                # this is needed because undo/redo don't cause all entities to have the
                # event fired.  For example, if you undo a move of a detached edge, then
                # no EntitiesObserver is fired.  Or, if you open/close a group.
                push_all_entities_from(@model.active_entities.parent)
            end

            def register_cdef(cdef)
                if !@definitions_observer.registered?(cdef)
                    @definitions_observer.add_component_definition(cdef)
                    cdef.entities.add_observer(@entities_observer)
                    push_all_entities_from(cdef)
                end
            end

            def push_edges(result, loop1)
                # !?@! An automatic error reports that 'loop1' is a Sketchup::Curve, which is
                # impossible according to the documentation and my attempts to reproduce it.
                # Here's a random blind fix attempt, assuming a closed curve with no reversed
                # edges.  Actually one report says that 'loop1' is a Sketchup::OptionsProvider,
                # which is absolute nonsense.  I bet SketchUp is buggy in the way the Ruby
                # objects are mapped to C objects.  Meh.
                if !(loop1.is_a? Sketchup::Loop)
                    if loop1.is_a? Sketchup::Curve
                        loop1.each_edge { |edge| result.push(edge.entityID) }
                    end
                    return
                end
                loop1.edgeuses.each {
                    |edgeuse|
                    # In the same vein as above, one report says that "edgeuse" is a
                    # Sketchup::AttributeDictionary.  I give up.
                    num = edgeuse.edge.entityID
                    if edgeuse.reversed?
                        num = -num
                    end
                    result.push(num)
                }
            end

            def entityID_of(cdef)
                return 0 if cdef == @model
                return cdef.entityID
            end

            def mat_index(material)
                if (material == nil) || (!(@material_to_index.has_key? material))
                    return -1
                end
                return @material_to_index[material]
            end

            def mat_filename_index(filename)
                if !(@material_filename_to_index.has_key? filename)
                    return -1
                end
                return @material_filename_to_index[filename]
            end

            def matuv(result, uvgetter, face, is_front)
                origin = uvgetter.call(Geom::Point3d.new(0,0,0))   # -> UVQ
                org = origin.to_a
                ptx = uvgetter.call(Geom::Point3d.new(1,0,0)).to_a
                pty = uvgetter.call(Geom::Point3d.new(0,1,0)).to_a
                ptz = uvgetter.call(Geom::Point3d.new(0,0,1)).to_a

                if ((origin == [0, 0, 1]) &&    # equal within tolerance
						Geom::Vector3d.new(ptx.z, pty.z, ptz.z) == [1, 1, 1] &&
                        (face.get_texture_projection(is_front) == nil))
                    org = [1e300, 1e300, 1e300]     # HACK, detected in Unity
                end
                result.push(org[0], org[1], org[2],
                            ptx[0], ptx[1], ptx[2],
                            pty[0], pty[1], pty[2],
                            ptz[0], ptz[1], ptz[2])
            end

            def serialize_edge(result, edge, incremental_removes=nil)
                kind = 1.0 + (8 << 10)
                if edge.hidden?
                    kind += 32.0
                end
                if edge.soft?
                    kind += 128.0
                end
                if incremental_removes != nil
                    if edge.faces.index { |face| !face.hidden? } != nil
                        kind += 64.0    # "not standalone" flag
                    end
                end
                #puts("    EDGE " + edge.entityID.to_s + "  flag " + kind.to_s)
                result.push(kind, edge.entityID)
                result.concat(edge.start.position.to_a)
                result.concat(edge.end.position.to_a)
            end

            def serialize_face(result, entity)
                mat1 = entity.material
                mat2 = entity.back_material
                uvhelper = entity.get_UVHelper
                if uvhelper.is_a? Sketchup::UVHelper
                    uvhelper_front = lambda {|p| uvhelper.get_front_UVQ(p)}
                    uvhelper_back  = lambda {|p| uvhelper.get_back_UVQ(p)}
                else
                    # no clue how it is possible!  It seems to be, though.
                    uvhelper_front = nil
                    uvhelper_back = nil
                end
                flag_loc = result.length
                kind = 5.0    # fixed below
                result.push(kind, entity.entityID)
                result.concat(entity.normal.to_a)
                result.push(mat_index(mat1), mat_index(mat2))
                # we always try to send at least one set of UVs.  If the material is "default", it
                # is useful to have it in case this face is part of a group that is itself painted
                # with a texture.  If the material is some solid color, it is useful in the corner
                # case where we load a texture into that material later.
                if uvhelper_front != nil
                    kind += 32.0
                    matuv(result, uvhelper_front, entity, true)
                end
                if (mat1 == nil) && (mat2 == nil)
                    # don't need to re-send the same UV twice, for two sides with "default".
                else#if (mat2 == nil) || (mat2.materialType != Sketchup::Material::MATERIAL_SOLID)
                    if uvhelper_back != nil
                        kind += 64.0
                        matuv(result, uvhelper_back, entity, false)
                    end
                end
                #puts("    FACE " + entity.entityID.to_s)
                push_edges(result, entity.outer_loop)
                if entity.loops.length > 1
                    entity.loops.each {
                        |inner_loop|
                        if inner_loop != entity.outer_loop
                            result.push(0.5)
                            push_edges(result, inner_loop)
                        end
                    }
                end
                result.push(0.0)
                result[flag_loc] = kind + ((result.length - flag_loc) << 10)
            end

            SPEAKER_UNICODE = "\xf0\x9f\x94\x88"

            def serialize_text(result, entity)
                flag_loc = result.length
                kind = 6.0    # fixed below
                kind += 32.0 if entity.display_leader?
                kind += 64.0 if entity.arrow_type != 0
                kind += 128.0 if entity.arrow_type == 2
                has_sound = entity.attribute_dictionary("Baroque_VRSketch_audio") != nil
                kind += 256.0 if has_sound
                result.push(kind, entity.entityID)
                result.concat(entity.point.to_a)
                result.concat((entity.point + entity.vector).to_a)
                chars = entity.text.chars
                chars.shift if (has_sound && (chars[0] == SPEAKER_UNICODE))
                result.push(chars.length)
                chars.each { |char|
                    result.push(char.ord)
                }
                result[flag_loc] = kind + ((result.length - flag_loc) << 10)
            end

            def _dynamic_attr(group_instance, attrname)
                result = group_instance.get_attribute("dynamic_attributes", attrname)
                if result == nil
                    result = group_instance.definition.get_attribute("dynamic_attributes", attrname)
                end
                return result
            end

            def _set_dynamic_attr(group_instance, attrname, value)
                group_instance.set_attribute("dynamic_attributes", attrname, value)
            end

            def _set_default_dynamic_attr(group_instance, attrname, value)
                if _dynamic_attr(group_instance, attrname) == nil
                    _set_dynamic_attr(group_instance, attrname, value)
                end
            end

            def get_animation_func(entity)
                onclick = _dynamic_attr(entity, "onclick")
                if onclick != nil
                    searchtext = onclick.strip.upcase
                    match = /^(?<func>ANIMATE\w*)\s*\(\s*"?(?<varname>\w+)"?(?<numbers>[\s\d.,-]+)\)\s*;?$/ =~ searchtext
                    if match != nil
                        values = []
                        numbers.split(",").each { |number|
                            values.push(number.to_f) if number.strip != ""
                        }
                        return [func, varname, values]
                    end
                    puts("dynamic animation: onclick too complex: " + onclick)
                end
                return [nil, nil, nil]
            end

            def get_animation_unit(entity, create_if_missing)
                unit = _dynamic_attr(entity, "_lengthunits")
                if unit == "CENTIMETERS"
                    result = 1 / 2.54
                elsif unit == "INCHES"
                    result = 1.0
                else
                    if @model.options["UnitsOptions"]["LengthFormat"] == 0
                        unit = "CENTIMETERS"
                        result = 1 / 2.54
                    else
                        unit = "INCHES"
                        result = 1.0
                    end
                    _set_dynamic_attr(entity, "_lengthunits", unit) if create_if_missing
                end
                return result
            end

            def get_animation(entity)
                # Try to decode the dynamic animation parameters
                func, varname, values = get_animation_func(entity)
                if func != nil
                    varindex = ["X", "Y", "Z", "ROTX", "ROTY", "ROTZ"].index(varname)
                    if varindex != nil
                        current = _dynamic_attr(entity, varname.downcase)
                        current = 0.0 if !((current.is_a? Integer) || (current.is_a? Float))
                        if func.start_with? "ANIMATECUSTOM"
                            speedvalues = values[0...3]
                            values.shift(3)    # drop the first three elements
                        elsif func.start_with? "ANIMATESLOW"
                            speedvalues = [1, 0, 100]
                        elsif func.start_with? "ANIMATEFAST"
                            speedvalues = [0.25, 0, 100]
                        else
                            speedvalues = []   # default: 0.5, 0, 100
                        end
                        if values.min != values.max
                            if varindex < 3    # "x", "y" or "z"
                                unit = get_animation_unit(entity, false)
                                # the unit of distance on the socket is always inches
                            else
                                unit = 1.0    # for angles, we send degrees
                            end
                            # note: "current" is ignored if varindex is "x" or "y" or "z"
                            animation = [varindex, current * unit, values.min * unit, values.max * unit]
                            animation.concat(speedvalues)
                            if values.length >= 3
                                animation.push(0.5, 0, 100) if speedvalues.length == 0
                                animation.push(values.length)
                                values.each { |x| animation.push(x * unit) }
                            end
                            return animation
                        end
                    else
                        puts("dynamic animation: only supports animating directly X, Y, Z, RotX, RotY or RotZ, but got " + varname)
                    end
                end
                return nil
            end

            def serialize_group(result, entity)
                flag = 3.0
                flag += 32.0 if entity.is_a? Sketchup::ComponentInstance
                definition = entity.definition
                animation = get_animation(entity)
                flag += 64.0 if animation != nil
                mat1 = entity.material
                flag += 20 << 10
                flag += animation.length << 10 if animation != nil
                result.push(flag, entity.entityID, definition.entityID, mat_index(mat1))
                result.concat(entity.transformation.to_a)
                result.concat(animation) if animation != nil
                #
                register_cdef(entity.definition)
                #
                if ((entity.is_a? Sketchup::ComponentInstance) && entity.glued_to &&
                        entity.definition.behavior.cuts_opening?)
                    warn("This model is using components with the 'Cut opening' feature.  This cuts a hole " +
                         "in the face that the component touches.  In VR, this feature is not implemented " +
                         "and such holes won't appear.") {
                        names = ["Components using 'Cut opening':"]
                        @model.definitions.each { |cdef|
                            names.push("\n    " + cdef.name) if cdef.behavior.cuts_opening?
                        }
                        names.join
                    }
                end
            end

            def hidden_entity(entity, incremental_removes)
                incremental_removes.push(entity.entityID) if incremental_removes != nil
            end

            def serialize_entity(result, entity, incremental_removes, toplevel)
                # fill in a list of floats with a serialized format from the entity:
                #
                # - edge:  [1+flags, id, x1, y1, z1, x2, y2, z2]
                #       flags: 32 = hidden edge
                # - face:  [5+flags, id, normx, normy, normz, mat1, mat2, (uv1, uv2), edges...]
                #       flags: 32 = uv1 present
                #              64 = uv2 present
                # - group: [3+flags, id, definition_id, mat, transformation(4x4) (, animation data)]
                #       flags: 32 = it is a component
                #              64 = animated
                # - image: [4, id, mat, transformation(4x4)]
                # - text: [6, id, x1, y1, z1, x2, y2, z2, text_length, chars...]
                # - end marker: [0]
                #
                # the uv1/uv2 are 12 numbers each, but are only present if the flag 32/64 is given
                #
                # the list of edges contains a list of:
                #     edge_id        add an edge
                #     (-edge_id)     add the reverse of an edge
                #     0.5            start of another loop (a hole)
                #     0              end of list
                #
                if entity.is_a? Sketchup::Edge
                    serialize_edge(result, entity, incremental_removes)
                    faces = entity.faces
                    if (faces == nil) || (faces.length == 0)
                        @vertex_observer.track(entity.start)
                        @vertex_observer.track(entity.end)
                    end
                    return
                end

                if entity.is_a? Sketchup::Face
                    return hidden_entity(entity, incremental_removes) if entity.hidden?
                    serialize_face(result, entity)
                elsif (entity.is_a? Sketchup::Group) || (entity.is_a? Sketchup::ComponentInstance)
                    return hidden_entity(entity, incremental_removes) if entity.hidden?
                    @group_observer.track(entity)
                    return hidden_entity(entity, incremental_removes) if group_layer_hidden(entity, toplevel)
                    serialize_group(result, entity)
                elsif entity.is_a? Sketchup::Image
                    return hidden_entity(entity, incremental_removes) if entity.hidden?
                    @group_observer.track(entity)
                    return hidden_entity(entity, incremental_removes) if group_layer_hidden(entity, toplevel)
                    result.push(4.0 + (19 << 10), entity.entityID, mat_filename_index(entity.path))
                    result.concat(entity.transformation.to_a)
                elsif entity.is_a? Sketchup::Text
                    return hidden_entity(entity, incremental_removes) if entity.hidden?
                    return hidden_entity(entity, incremental_removes) if entity.leader_type == 0
                    #          ignore Texts that are purely screen-space ^^^
                    serialize_text(result, entity)
                else
                    puts("Ignoring entity of type " + entity.typename)
                end
            end

            def group_layer_hidden(entity, toplevel)
                if (!(entity.layer.visible?))
                    # same rule as SketchUp: if a non-toplevel-group is on Layer0 and that Layer0
                    # is hidden, we still show the group
                    if toplevel || (entity.layer != @model.layers[0])
                        return true
                    end
                end
                return false
            end

            def serialize_notes()
                dict = @model.attribute_dictionary("Baroque_VRSketch_notes", false)
                old_version = @notes_version
                @notes_version = nil
                @notes_version = dict["version"] if dict != nil
                return if @notes_version == old_version
                #
                notes = []
                data = []
                if dict != nil
                    dict.each_pair { |key, value|
                        next if key[0] != "p"
                        next if !(value.is_a? Array)
                        next if value.length == 0
                        # much more efficient to send the long array of floats inside 'data'
                        # than write it directly inside 'notes', where it would be JSON-encoded
                        color = dict["c" + key[1...key.length]]
                        notes.push({
                            key: key[1...key.length].to_i,
                            n: value.length,
                            color: color,
                        })
                        data.concat(value)
                    }
                end
                raw = data.pack("f*")
                emit_rawdata(raw, {
                    cmd: "notes",
                    notes: notes,
                })
            end

            def emit(cmd)
                data = JSON.generate(cmd)
                Baroque::VRSketch.sendall(@sock, data + "\x00")
            end

            def _emit_rawdata_direct(raw, cmd)
                cmd[:rawlength] = raw.bytesize
                emit(cmd)
                Baroque::VRSketch.sendall(@sock, raw)
            end

            def emit_rawdata(raw, cmd)
                if raw.bytesize >= 10000
                    if @cache_speedup_out
                        @cache_speedup_out.push(raw)
                        raw_checksum = Digest::SHA1.digest(raw)
                        throw "bad checksum length" if raw_checksum.bytesize != 20
                        rawcmd = JSON.generate(cmd)
                        rawcmd.force_encoding "ASCII-8BIT"
                        _emit_rawdata_direct(raw_checksum + rawcmd, {cmd: "cache"})
                        _process_incoming_cache_requests
                        return
                    end
                end
                _emit_rawdata_direct(raw, cmd)
            end

            def add_material(material)
                if !(@invalid_materials.has_key? material)
                    @invalid_materials[material] = true
                    _enable_updates
                end
            end

            def _make_sure_we_have_model_definitions
                return if @model.definitions
                # occasionally we see a model where @model.definitions == nil.  No clue when,
                # but maybe it's after we convert from an old file format and there are no
                # components or groups.  Blind attempt to fix; I have no clue if it works.
                # Report to the user if it doesn't work.
                begin
                    @model.start_operation("fix the definitions list", true)
                    ctx = @model.active_entities
                    ctx.add_group    # will disappear again because empty
                    @model.commit_operation
                ensure
                    if !@model.definitions
                        warn("This model is missing a 'definitions' list. Maybe you need to create " +
                             "a component to force it to exist? It is also possible that there " +
                             "are already components in this model, but the SketchUp Ruby API " +
                             "does not report them because of a bug. Try to save and open a copy " +
                             "of your model to fix the problem.")
                    end
                end
            end

            def _try_receive_early_message
                loop do
                    return nil if !@early_message_mode
                    begin
                        # reading at most 8 bytes at a time, we're sure we never read the end of
                        # a message together with the complete message that follows.
                        data = @sock.recv_nonblock(8)
                        raise if data == nil
                    rescue IO::WaitReadable
                        return nil
                    rescue
                        @early_message_mode = false
                        return nil
                    end
                    @sock_buffer += data
                    break if data.index("\x00") != nil
                end
                
                limit = @sock_buffer.index("\x00")
                cmd = parse_json(@sock_buffer[0...limit])
                @sock_buffer = @sock_buffer[(limit + 1)...(@sock_buffer.length)]
                
                if cmd["cmd"] == "my_window"
                    @early_message_mode = false
                    got_my_window_info(cmd)
                    return nil
                end
                return cmd
            end

            def _wait_for_initial_message
                return if @sock_buffer.bytesize > 0
                Sketchup.set_status_text(WAITINITIAL_TEXT)

                # For caching, we need some limited data exchange while sending the model.
                # But we only process regular incoming message after the whole model has
                # been sent.  So we have another code path to process "early" incoming messages
                # until we get a "my_window" command, which is the marker for the regular
                # path.
                @early_message_mode = true

                n_timeout = 300      # 15 seconds
                loop do
                    cmd = _try_receive_early_message
                    if cmd != nil
                        if cmd["cmd"] == "fcache"
                            @cache_speedup_out = []
                            @cache_speedup_in = @cache_speedup_out   # same list
                            @cache_speedup_in_index = 0
                        end
                        return
                    else
                        return if !@early_message_mode
                        return if n_timeout <= 0
                        sleep(0.05)
                        n_timeout -= 1
                    end
                end
            end

            def _process_incoming_cache_requests
                loop do
                    cmd = _try_receive_early_message
                    return if cmd == nil
                    process_cache_request(cmd) if cmd["cmd"] == "fcache"
                end
            end

            def process_cache_request(cmd)
                skip = cmd["skip"]
                if skip
                    skip.times { |n|
                        @cache_speedup_in[@cache_speedup_in_index] = nil   # free memory
                        @cache_speedup_in_index += 1
                    }
                end
                get = cmd["get"]
                if get
                    get.times { |n|
                        raw = @cache_speedup_in[@cache_speedup_in_index]
                        @cache_speedup_in[@cache_speedup_in_index] = nil   # free memory
                        @cache_speedup_in_index += 1
                        _emit_rawdata_direct(raw, {cmd: "fcache"})
                    }
                end
                if ((@cache_speedup_out == nil) && (@cache_speedup_in_index == @cache_speedup_in.length))
                    @cache_speedup_in = nil
                    @cache_speedup_in_index = nil
                end
            end

            def upload_whole_model(for_export: false)
                @invalid_materials = {}
                # NOTE: materials.each{} skips the ones used for images apparently
                i = @model.materials.length
                while i > 0
                    i -= 1
                    @invalid_materials[@model.materials[i]] = true
                end
                @seen_materials = {}
                _make_sure_we_have_model_definitions
                total_defs = @model.definitions.length
                bounds = @model.bounds
                emit({
                    cmd: "clearmodel",
                    id: RB_VERSION_ID,
                    v: [total_defs, @tutorial ? 1 : 0, Sketchup.version_number],
                    transform: bounds.min.to_a + bounds.max.to_a,
                })
                if (bounds.max.z - bounds.min.z) > 40000
                    warn("This model is more than 1000 meters high.  You might hit rendering issues " +
                         "and not be able to meaningfully use the teleporter.  Check that the scale " +
                         "is correct.")
                end

                _wait_for_initial_message if !for_export

                push_all_entities_from(@model)
                send_textures = emit_components(total_defs)

                emit({
                    cmd: "modelcomplete",
                    v: [send_textures.length],
                })
                shadow_info_changed(false)
                units_options_changed
                allowed_actions_changed(nil)
                send_scenes_list
                send_layers_list
                send_components_list
                emit({
                    cmd: "setconfig",
                    name: "snap_scale",
                    v: [Baroque::VRSketch.snap_scale * 0.001]
                })

                emit_textures(send_textures)

                msg = {cmd: "texturescomplete"}
                if !for_export
                    msg[:names] = get_base_skm_directories
                end
                emit(msg)
                
                @cache_speedup_out = nil
            end

            def active_path_maybe_changed()
                new_path = @model.active_path
                new_path = [] if new_path == nil
                return if new_path == @active_path
                #
                # say we just went from [A, B, C] to [A, D, E].  Then:
                #
                # - @model entities never move;
                #
                # - "A" entities don't move: they were and are still at coordinates expressed in
                #   the absolute coordinate system
                #
                # - "B" and "C" entities move: they used to be in the absolute coordinate system,
                #   but are now relative ("B" inside "A", and "C" inside "B").
                #
                # - "D" and "E" entities move, in opposite way.
                #
                # - subgroups of any of the A, B, C, D, E groups don't move: their entities are
                #   always relative to their parent.
                #
                # Note that the component instances inside "A" that contain "B" and "D" also
                # both need to be updated, as well possibly as other instances of the same
                # components.  If "D" or "E" is a group, opening it will un-share it with other
                # group instances that used to contain the same group definition; this is done
                # by SketchUp by making a new copy of the group definition and assigning it to
                # all the *other* group instances.  The EntityObserver appears enough for the
                # latter case but not the former case, so we need to do it manually here.
                #
                i = 0
                while (i < @active_path.length) && (i < new_path.length) && (@active_path[i] == new_path[i])
                    i += 1
                end
                @active_path.drop(i).each { |grp| push_update_group(grp) }
                new_path.drop(i).each { |grp| push_update_group(grp) }

                @active_path = new_path
            end

            def push_update_group(grp)
                return if grp.deleted?
                return if grp.definition.deleted?
                push_update(grp)
                push_all_entities_from(grp.definition)
                grp.definition.instances.each { |grp| push_update(grp) }
            end

            WAITINITIAL_TEXT = "Connecting to VR Sketch..."
            EMITTING_TEXT = "Sending data to VR Sketch..."
            READY_TEXT = "Ready"
			FROMCLOUD_TEXT = "Downloading from cloud server..."

            def emit_components(total_defs = 0)
                if @emit_components_rec
                    throw "recursive invocation of emit_components!"
                end
                @emit_components_rec = true
                begin
                    _emit_components(total_defs)
                ensure
                    @emit_components_rec = nil
                end
            end

            def report_invalid_texture_size
                names = ["Materials and images with invalid texture size:"]
                i = 0
                while i < @model.materials.length
                    mat = @model.materials[i]
                    i += 1
                    tex = mat.texture
                    next if !tex

                    if tex.image_width <= 16384 && tex.image_height <= 16384
                        if tex.width >= 0 && tex.height >= 0   # and not NaN
                            next
                        else
                            error = "invalid width/height (%sx%s)" % [tex.width, tex.height]
                        end
                    else
                        error = "too large (%dx%d)" % [tex.image_width, tex.image_height]
                    end
                    names.push("\n    '" + mat.display_name + "': " + error)
                end
                return names.join
            end

            def _emit_components(total_defs = 0)
                # Emit the geometry that changed according to @pending_updates and
                # @pending_removes.  Returns a list of textures left to send.

                Sketchup.set_status_text(EMITTING_TEXT)
                serialize_notes()

                material_names = []
                @material_to_index = {}
                @material_filename_to_index = {}
                # NOTE: materials.each{} skips the ones used for images apparently
                i = 0
                while i < @model.materials.length
                    mat = @model.materials[i]
                    @material_to_index[mat] = material_names.length
                    if mat.texture
                        @material_filename_to_index[mat.texture.filename] = material_names.length
                    end
                    material_names.push(mat.object_id.to_s)
                    i += 1
                end
                #puts({:cmd => "materials", :names => material_names})
                emit({
                    cmd: "materials",
                    names: material_names,
                    transform: @model.edit_transform.to_a,
                })
                send_textures = []
                if @invalid_materials.length > 0
                    @invalid_materials.each_key {
                        |material|
                            next if material.deleted?
                            material_def = {
                                display_name: material.display_name,
                                alpha: material.alpha,       # overall alpha, entered as "Opacity" by the user
                                color: material.color.to_i,  # either the user-entered color; or if there is a texture, the computed mean color.  "to_i" strips the alpha.
                                width: -1.0,                 # "no texture", may be overwritten below
                            }
                            tex = material.texture
                            if tex != nil
								if tex.image_width <= 16384 && tex.image_height <= 16384
									if tex.width >= 0 && tex.height >= 0   # and not NaN
										send_textures.push(material)
										material_def[:width] = tex.width
										material_def[:height] = tex.height
										material_def[:pixel_width] = tex.image_width
										material_def[:pixel_height] = tex.image_height
									else
										warn("some textures have got invalid width/height (sometimes, copying the model in a new file fixes it)") {
                                            report_invalid_texture_size
                                        }
									end
								else
									warn("textures larger than 16384 pixels are not rendered in Unity") {
                                        report_invalid_texture_size
                                    }
								end
                            end
                            # NB. we used to use 'material.name' for :name, but we shouldn't do it any
                            # more, because it is not unique after all.
                            emit({
                                cmd: "material",
                                name: material.object_id.to_s,
                                material_def: material_def,
                            })
                            #puts({:material => material.name, :color => material.color.to_i})
                    }
                    @invalid_materials = {}
                end

                # Remove cdefs that are completely gone
                if @pending_removes_cdef_id != nil
                    emit({
                        cmd: "remove_cdefs",
                        id_list: @pending_removes_cdef_id,
                    })
                    @pending_removes_cdef_id = nil
                end

                # Check if the active path changed, and if so, invalidate the intervening groups
                active_path_maybe_changed

                # Update the @pending_updates, just in case we have some non-nil entry but
                # the cdef has never been seen before
                id_list = []
                @pending_updates.keys.each { |cdef|
                    if ((cdef != @model) && cdef.deleted?)
                        @pending_updates.delete(cdef)
                    else
                        register_cdef(cdef)
                        id_list.push(entityID_of(cdef))
                    end
                }

                # emit "start_update", collecting a few stuff before uploading the structure
                path_id = []
                @active_path.each { |grp| path_id.push(grp.entityID) }

                history = @undo_history
                start = history.length - @undone_operations - 10
                start = 0 if start < 0
                stop = history.length - @undone_operations + 10
                stop = history.length if stop > history.length
                #puts({:history => history.slice(start, stop - start), :undone_ops => @undone_operations})
                emit({
                    cmd: "start_update",
                    id_list: id_list,
                    v: path_id,
                    transform: @model.axes.transformation.to_a,
                    history: history.slice(start, stop - start).map { |h| h[:msg] },
                    undone_ops: @undone_operations
                })

                # Main loop

                counter = 0
                while @pending_updates.length > 0
                    cdef, entities = @pending_updates.shift
                    toplevel = (cdef == @model)
                    #puts({:cdef => cdef})

                    data = []
                    if entities == nil
                        cdef.entities.each { |entity| serialize_entity(data, entity, nil, toplevel) }
                        msg = {cmd: "setentities"}
                        if cdef != @model && cdef.behavior.always_face_camera?
                            msg[:v] = [1]
                        end
                    else
                        if @pending_removes.has_key?(cdef)
                            incremental_removes = @pending_removes[cdef]
                        else
                            incremental_removes = []
                        end
                        # Even moving a single face might require invalidation of adjacent faces,
                        # like if we do a push/pull.  So we compute the whole set of vertices, and
                        # then from there go back to the whole set of edges and faces.
                        extra_faces = []
                        entities = entities.select { |e| !(e.deleted?) }
                        entities.reverse_each { |entity|
                            if ((entity.is_a? Sketchup::Face) || (entity.is_a? Sketchup::Edge))
                                entity.vertices.each { |vertex|
                                    entities.concat(vertex.edges)      # grow the lists a lot, but it will
                                    extra_faces.concat(vertex.faces)   #    be reduced with .uniq below
                                }
                            end
                        }
                        extra_faces.uniq.each { |face|
                            entities.push(face)
                            entities.concat(face.edges)    # face => all edges, for Unity
                        }
                        entities.uniq.each { |entity|
                            serialize_entity(data, entity, incremental_removes, toplevel)
                        }
                        #if incremental_removes.length > 0
                        #    s = "INCREMENTAL, REMOVES:"
                        #    incremental_removes.each { |entity_id|
                        #        s += " " + entity_id.to_s
                        #    }
                        #    puts(s)
                        #else
                        #    puts("INCREMENTAL, NO REMOVES")
                        #end
                        data = [incremental_removes.length] + incremental_removes + data
                        msg = {cmd: "addentities"}
                    end

                    #-- Disabled, because on Windows 10, after clicking somewhere else and back in the
                    #-- SketchUp window, Windows thinks the SketchUp program is stuck and no longer even
                    #-- updates the status text...  so we get a random "component 401 / 1782" that stays
                    #-- even while we're really progressing, which makes the user think the plugin crashed.
                    #-- Bah.
                    #counter += 1
                    #counter_limit = ""
                    #counter_limit = (" / " + total_defs.to_s) if counter <= total_defs
                    #Sketchup.set_status_text(EMITTING_TEXT + " component " + counter.to_s + counter_limit)

                    raw = data.pack("d*")
                    data = nil
                    #puts({:cmd => msg, :id => cdef.entityID, :rawlength => raw.length})
                    msg[:id] = entityID_of(cdef)
                    msg[:name] = cdef.name if cdef.is_a? Sketchup::ComponentDefinition
                    cdefbounds = cdef.bounds
                    msg[:transform] = cdefbounds.min.to_a + cdefbounds.max.to_a
                    emit_rawdata(raw, msg)
                end
                @pending_updates = nil
                @pending_removes = nil

                #puts("===== UPDATED =====")

                return send_textures
            end

            def emit_textures(send_textures)
                # now we load and send all the textures.  There is messy logic for SketchUp 2017,
                # and much simpler logic for SketchUp >= 2018.
                if send_textures.length > 0
                    # materials.each{} skips the ones used for images.  Use that to learn
					# if a given texture is used for an image or not...
					not_for_image = {}
					@model.materials.each { |material|
						not_for_image[material] = true 
					}

                    failures = []
                    send_textures.each { |material|
                        msg = {
                            cmd: "texture",
                            name: material.object_id.to_s,
                        }
                        raw = nil
                        begin
                            # on Windows we get bytes in the BGR or BGRA order; on Mac we get them
                            # in the RGB or RGBA order.  We could convert like this:
                            # (0...raw.length).step(bpp/8) { |j| raw[j],raw[j+2] = raw[j+2],raw[j] }
                            # However, in some cases on Mac it seems that image_rep.data is just
                            # garbage; at least I've seen the case with a 24bpp texture, 450x450,
                            # with 2 bytes of padding per row.  The bytes start reasonably, up to
                            # 20 or 100, and then they are nonsense.  Shrug, just don't use
                            # image_rep on Mac.
                            #
                            # In fact, we shouldn't use image_rep in any case that involves copying
                            # the data over the network, because the uncompressed images are much,
                            # much bigger.
                            raise if !Baroque::VRSketch::IS_WINDOWS
                            raise if !@connected_to_localhost

                            image_rep = material.texture.image_rep(true)   # get colorized version
                            bpp = image_rep.bits_per_pixel
                            if ((bpp == 24) || (bpp == 32))
                                raw = image_rep.data
                                if raw != nil
                                    msg[:id_list] = [image_rep.width, image_rep.height, image_rep.row_padding]
                                end
                            end
                        rescue
                            # this might be either SketchUp 2017 which does not implement image_rep,
                            # or image_rep failing for some other (rare and obscure) cause.
                            raw = nil
                        end
                        
                        if raw == nil
                            # SketchUp 2017 (or, unusual bits_per_pixel)

                            # We make up a filename inside temp_dir, save the texture to disk,
                            # reload it, and send that over the socket.  Oh well.  We also try to
                            # use the jpg format instead of png, because SketchUp is incredibly slow
                            # when writing png; but jpg doesn't support per-pixel alpha...
                            # TIFF is faster to write than png and supports alpha, but Unity does
                            # not natively support TIFF.
                            filename_png = Baroque::VRSketch.temporary_filename(".png")
                            begin
                                # an additional level of fun: we can't rely on 'average_color.a' to tell
                                # us if the texture uses alpha or not.  Sometimes it returns 255 even though
                                # the texture is largely transparent.  Found this obscure workaround:
                                # emit a 8x8 thumbnail png image (apparently 1x1 sometimes fails), and
                                # check if that image contains an alpha channel or not.
                                opacity = 128
                                if material.write_thumbnail(filename_png, 8)
                                    raw = File.binread(filename_png)
                                    i = raw.index("IHDR")
                                    colortype = raw[i + 13].ord
                                    opacity = 255 if colortype < 4
                                end
                                if opacity < 255
                                    filename1 = filename_png
                                else
                                    filename_jpg = Baroque::VRSketch.temporary_filename(".jpg")
                                    filename1 = filename_jpg
                                end
                                #puts("#{material.name} ===> #{filename1}")
                                material.texture.write(filename1, true)
                                raw = File.binread(filename1)
                                File.delete(filename1)
                            rescue
                                failures.push(material.display_name)
                                next
                            end
                            msg[:undone_ops] = opacity
                        end
                        msg[:v] = [1] if !not_for_image[material]
                        emit_rawdata(raw, msg)
                        Baroque::VRSketch.stat_event("texture", raw.length)
                    }
                    if failures.length > 0
                        faillength = failures.length
                        UI.messagebox("Cannot read texture files for #{faillength} materials:\n#{failures}")
                    end
                end
                Sketchup.set_status_text(READY_TEXT)
            end

            def flush_transaction_changes()
                return if @pending_updates == nil
                return if @sock == nil
                return if @model.materials == nil
                
                upload_transaction_changes_now
            end

            def upload_transaction_changes_now()
                #sleep(4) -- to test feedback-in-large-models
                send_textures = emit_components
                emit({
                    cmd: "updatecomplete"
                })

                if @keep_selection != nil
                    picked = @keep_selection
                    @keep_selection = nil
                    id_list = []
                    v = []
                    if ((picked.length > 0) && (picked[0].is_a? Integer))
                        picked = @model.find_entity_by_id(picked)
                    end
                    picked.each { |entity|
                        if (in_current_group(entity))
                            if (entity.is_a? Sketchup::Face) || (entity.is_a? Sketchup::Group) || (entity.is_a? Sketchup::ComponentInstance)
                                id_list.push(entity.entityID)
                            elsif entity.is_a? Sketchup::Edge
                                v.concat(entity.start.position.to_a)
                                v.concat(entity.end.position.to_a)
                            end
                        end
                    }
                    emit({
                        cmd: "selection",
                        id_list: id_list,
                        v: v,
                    })
                end

                emit_textures(send_textures)
                @pending_mat_msgs.each { |msg| emit(msg) }
                @pending_mat_msgs = []

                if @cancel_wait_id > 0
                    emit({cmd: "cancelwait", id: @cancel_wait_id})
                    @cancel_wait_id = 0
                end

                Baroque::VRSketch.stat_event("model update")
            end


            # ----- OTHER CONTROL LOGIC -----
            # This part is for answering to commands from the Unity side

			def read_rawdata(rawlength)
				while @sock_buffer.length < rawlength
					begin
						data = @sock.recv(rawlength - @sock_buffer.length)
					rescue
						data = ""
					end
					return nil if data.length == 0
					@sock_buffer += data
				end
				result = @sock_buffer[0...rawlength]
				@sock_buffer = @sock_buffer[rawlength...(@sock_buffer.length)]
				return result
			end

			def parse_json(buffer)
				begin
                    return JSON.parse(buffer)
                rescue Exception => e
                    buffer = buffer[0, 1000].inspect
                    raise e.class, e.message + "\n#{buffer}"
                end
			end

            def run_timer()
                next_size = read_data_block(8192)
                while next_size
                    next_size = read_data_block(next_size)
                end
            end

            def read_data_block(next_size)
                if @no_recursion_in_read_data_block
                    return nil     # can occur with edit_download_model, notably
                end
                if @sock == nil
                    UI.stop_timer(@timer) if @timer
                    @timer = nil
                    return nil
                else
                    begin
                        data = @sock.recv_nonblock(next_size)
                    rescue IO::WaitReadable
                        return nil
                    rescue    # usually Errno::ECONNRESET, but after .shutdown it is ECONNABORTED and ESHUTDOWN
                        data = ""
                    end
                    return nil if data == nil   # e.g. for class WebSock
                end
                if data.length == 0
                    close_socket
                    UI.stop_timer(@timer) if @timer
                    @timer = nil
                    return nil
                end
                begin
                    @no_recursion_in_read_data_block = true
                    next_size *= 2 if data.length == next_size   # full buffer, there might be (much) more coming
                    @sock_buffer += data
                    loop do
                        limit = @sock_buffer.index("\x00")
                        if limit == nil
                            break
                        end
                        rawdata = nil
                        cmd = parse_json(@sock_buffer[0...limit])
                        @sock_buffer = @sock_buffer[(limit + 1)...(@sock_buffer.length)]
                        if cmd["rawlength"] != nil
                            rawdata = read_rawdata(cmd["rawlength"])
                        end

                        typ = cmd["cmd"]
                        if typ == "pos"
                            Baroque::VRSketch.teleport_tool.update_position_to_display(cmd, @model)
                            #Baroque::VRSketch.teleport_tool.report_mouse
                        elsif typ.start_with? "edit_"
                            @sock_edit_cmd.push(cmd)
                        else
                            case typ
                            when "stat"
                                record_stat(cmd)
                            when "apply"
                                geom_edit_apply(cmd)
                            when "feedback"
                                emit_feedback(cmd["id"])
                            when "fcache"
                                process_cache_request(cmd)
                            when "undo_history"
                                emit_undo_history(cmd)
                            when "open_close_group"
                                open_close_group(cmd)
                            when "warning"
                                warn(cmd["text"])
                            when "send_model_now"
                                send_model_now
                            when "my_window"
                                got_my_window_info(cmd)
                            when "component_image"
                                send_component_image(cmd)
                            when "refresh_cdef"
                                refresh_cdef(cmd)
                            when "change_to_scene"
                                change_to_scene(cmd)
                            when "shadow_time"
                                change_shadow_time(cmd)
                            when "list_files"
                                cmd_list_files(cmd)
                            when "get_file"
                                cmd_get_file(cmd)
                            when "save"
                                save_model(cmd)
                            when "save_photo"
                                save_photo(cmd, rawdata)
                            when "skm_load"
                                skm_load(cmd)
                            when "point_cloud_read"
                                point_cloud_read(cmd)
                            when "audio"
                                cmd_audio(cmd, rawdata)
                            when "private_warehouse"
                                cmd_private_warehouse(cmd)
							when "tutorial"
                                # hack: start_tutorial() will open a new ModelTracker but we want
                                # it to continue reading/writing from the same socket
                                UI.stop_timer(@timer) if @timer
                                @timer = nil
                                sock = @sock
                                sock_buffer = @sock_buffer
                                @sock = nil
                                @sock_buffer = ''
								Baroque::VRSketch.start_tutorial(cmd["index"], @on_quest,
                                    [@hwnd, sock, sock_buffer])
                                return nil
                            else
                                puts("Unknown message..:")
                                puts(cmd)
                            end
                        end
                    end
                    return next_size
                ensure
                    @no_recursion_in_read_data_block = nil
                end
            end

            def _forget_socket
                # Unsure where it comes from, but occasionally there's a timer running a callback
                # on the ModelTracker used to upload to cloud.  This is used to mark it as closed
                # so that the callback won't do anything
                @sock = nil
            end

            def close_socket
                close_session_stats
                if @sock != nil
                    puts("Closing VR")
                    Baroque::VRSketch.teleport_tool.finish_now(@model)
                    @model.remove_observer(@model_observer)
                    # no clue when, but we've seen a @model.definitions being nil.
                    @model.definitions.remove_observer(@definitions_observer) if @model.definitions != nil
                    @model.materials.remove_observer(@materials_observer) if @model.materials != nil
                    @model.shadow_info.remove_observer(@shadow_info_observer) if @model.shadow_info != nil
                    @model.layers.remove_observer(@layers_observer) if @model.layers != nil
                    @model.options["UnitsOptions"].remove_observer(@units_options_observer) if ((@model.options != nil) && (@model.options["UnitsOptions"] != nil))
                    @model.rendering_options.remove_observer(@rendering_options_observer) if @model.rendering_options != nil
                    @model.pages.remove_observer(@pages_observer) if @model.pages != nil
                    #
                    @sock.close
                    @sock = nil
                    #
                    Baroque::VRSketch.deselect_teleport_tool
                    if @clipboard_file
                        begin
                            File.delete(@clipboard_file)
                        rescue
                        end
                        @clipboard_file = nil
                    end
                    if @notify_closing_queue
                        @notify_closing_queue.enqueue("set_status_error('Disconnected.')")
                    end
                    @wh_dialog.close if @wh_dialog
                end
            end

            def close_session_stats
                if @sock != nil
                    Baroque::VRSketch.stat_event("session", Time.now.to_f - @tracker_initialized)
                end
                Baroque::VRSketch.write_statistics
            end

            def record_stat(cmd)
                # "t" may be absent, in which case we only record one for "ev".  If "t" is present
                # we also record "ev total" and "ev max".
                Baroque::VRSketch.stat_event(cmd["ev"], cmd["t"])
            end

            def _pick_vertex(v_list, vx)
                i = v_list.length
                while i >= 3
                    i -= 3
                    if vx.position.distance(v_list.slice(i, 3)) < EPSILON_LENGTH
                        return true
                    end
                end
                return false
            end

            def _pick_entity_ids(entities)
                result = []
                entities.each { |entity|
                    result.push(entity.entityID) if !entity.deleted?
                }
                return result.uniq
            end

            def _grp2component(group)
                cinst = group.to_component
                cinst.definition.name = @model.definitions.unique_name("Component#1")
                return cinst
            end

            def _getscene(cmd)
                i = cmd["index"]
                page = @model.pages[i]
                return page if ((page != nil) && (page.name == cmd["old_name"]))
                # allow +1/-1 to occur
                if i > 0
                    page = @model.pages[i - 1]
                    return page if ((page != nil) && (page.name == cmd["old_name"]))
                end
                page = @model.pages[i + 1]
                return page if ((page != nil) && (page.name == cmd["old_name"]))
                #
                puts("edit_page_xxx: scenes list was modified in the background, not applying change")
                return nil
            end

            def in_current_group(entity)
                return ((entity != nil) && (entity.valid?) &&
                        (entity.parent == @model.active_entities.parent))
            end

            def _try_to_attach_text_label(text)
                pt = text.point
                bb = Geom::BoundingBox.new
                bb.add(pt - EPSILON_VECTOR, pt + EPSILON_VECTOR)
                ph = @model.active_view.pick_helper
                return if ph.boundingbox_pick(bb, Sketchup::PickHelper::PICK_CROSSING) == 0
                path = ph.path_at(0)
                return if path.length == 0
                begin
                    text.attached_to = [path, pt]
                rescue
                    # either SU < 2019 not supporting this attribute,
                    # or https://github.com/SketchUp/api-issue-tracker/issues/465
                end
            end

            def geom_edit_apply(cmd_apply)
                edit_cmd = @sock_edit_cmd
                @sock_edit_cmd = []
                op_name = cmd_apply["op_name"]
                layer0 = @model.layers[0]
                old_layer = @model.active_layer
                @model.start_operation("[VR] " + op_name, true)
                begin
                    Baroque::VRSketch.stat_event("apply " + op_name)
                    old_selection = nil
                    picked = []   # duplicates here will be automatically ignored by Sketchup
                    picked_from_back = 0
                    @model.active_layer = layer0 if layer0 != old_layer
                    ctx = @model.active_entities
                    edit_cmd.each { |item|
                        case item["cmd"]
                        when "edit_pick_face", "edit_pick_group"
                            entity = @model.find_entity_by_id(item["id"])
                            if !in_current_group(entity)
                                puts("pick_face: entity not in current group")
                            elsif entity != nil
                                picked.push(entity)
                            else
                                puts("pick_face: bad face/group id")
                            end
                        when "edit_pick_vertices"
                            entity = @model.find_entity_by_id(item["id"])
                            v_list = item["v_list"]
                            if !in_current_group(entity)
                                puts("pick_vertices: entity not in current group")
                            elsif entity.is_a? Sketchup::Face
                                entity.loops.each { |loop|
                                    loop.vertices.each { |vx| 
                                        picked.push(vx) if _pick_vertex(v_list, vx)
                                    }
                                }
                            elsif entity.is_a? Sketchup::Edge
                                got1 = _pick_vertex(v_list, entity.start)
                                got2 = _pick_vertex(v_list, entity.end)
                                if got1 && got2
                                    picked.push(entity)
                                elsif got1
                                    picked.push(entity.start)
                                elsif got2
                                    picked.push(entity.end)
                                end
                            else
                                puts("pick_vertices: bad face or edge id")
                            end
                        when "edit_pick_new_edge", "edit_pick_edge"   # xxx
                            edge = ctx.add_line(item["v1"], item["v2"])
                            if edge
                                picked.push(edge)
                            else
                                puts("pick_new_edge: add_line failed")
                            end
                        when "edit_move"
                            picked.reject! { |x| x.deleted? }
                            tr = Geom::Transformation.new(item["delta"])
                            ncopies = item["copy"]
                            if ncopies > 0
                                grp1 = ctx.add_group(picked)
								if grp1 != nil
									grp2 = grp1.copy
									grp1.explode
                            
									tr1 = tr
									ncopies0 = ncopies
									while ncopies > 1
										grp3 = grp2.copy
										grp3.transform!(tr)
										grp3.explode
										tr = tr1 * tr
										ncopies -= 1
									end
									grp2.transform!(tr)
									picked = grp2.explode
									if ncopies0 > 1
										picked = []
									end
								else
									puts("edit_move: add_group failed")
								end
                            else
                                transform_me = picked
                                # mess, we need to call transform_entities in two steps if the list
                                # contains both Drawingelements (Faces, Edges, Group instances, etc.)
                                # and Vertex, which is not a Drawingelement.
                                transform1 = transform_me.grep(Sketchup::Drawingelement)
                                transform2 = transform_me.grep(Sketchup::Vertex)
                                if (transform2.length > 0) && (transform1.length > 0)
                                    # We risk double-moving a vertex if it is both a vertex and part of
                                    # an edge/face.  Avoid that by detecting the case and not moving such
                                    # vertices individually.
                                    vertices = {}
                                    transform2.each { |vx| vertices[vx] = true }
                                    transform1.each { |entity|
                                        next if entity.deleted?
                                        if entity.is_a? Sketchup::Face
                                            entity.loops.each { |loop|
                                                loop.vertices.each { |vx| vertices.delete(vx) }
                                            }
                                        elsif entity.is_a? Sketchup::Edge
                                            vertices.delete(entity.start)
                                            vertices.delete(entity.end)
                                        end
                                    }
                                    transform2 = vertices.keys
                                end
                                if (((transform1.length > 0) && !ctx.transform_entities(tr, transform1)) ||
                                    ((transform2.length > 0) && !ctx.transform_entities(tr, transform2)))
                                    puts("move: operation failed")
                                else
                                    # ok, now we need to merge edges and vertices that end up being
                                    # on or identical to another edge...  the way I found to do it
                                    # is to make a group, add a copy of all edges that we moved to it,
                                    # and explode that group.  If there are two identical edges in the
                                    # current ctx, then this third copy will merge with both of them
                                    # during explosion, and only one will be left.  That's a hack.
                                    make_copy = {}
                                    transform1.each { |entity|
                                        next if entity.deleted?
                                        if entity.is_a? Sketchup::Edge
                                            make_copy[entity] = true
                                        elsif entity.is_a? Sketchup::Face
                                            entity.loops.each { |loop|
                                                loop.edges.each { |edge| make_copy[edge] = true }
                                            }
                                        end
                                    }
                                    transform2.each { |vx|
                                        next if vx.deleted?
                                        vx.edges.each { |edge| make_copy[edge] = true }
                                    }
                                    grp = ctx.add_group
                                    make_copy.each_key { |edge|
                                        e2 = grp.entities.add_line(edge.start.position, edge.end.position)
                                        if e2 != nil
                                            e2.smooth = edge.smooth?
                                            e2.soft = edge.soft?
                                            e2.hidden = edge.hidden?
                                            e2.material = edge.material
                                            e2.casts_shadows = edge.casts_shadows?
                                        end
                                    }
                                    exploded_edges = grp.explode
                                    # aaaaaand... the hack needs another hack on top of it, because
                                    # of the edges in the 'picked' list.  These end up marked as
                                    # deleted, because they are replaced with the 'exploded_edges'.
                                    # But not consistently.  Oh well, just take everything together
                                    # and then remove the deleted bits.
                                    picked.concat(exploded_edges)
                                    picked = picked.select { |x| !x.deleted? }
                                end
                            end
                        when "edit_add_edge"
                            grp = ctx.add_group
                            edge = grp.entities.add_line(item["v1"], item["v2"])
                            if edge
                                found_faces = 0
                                grp.explode.each { |e|
                                    if e.is_a? Sketchup::Edge
                                        found_faces += e.find_faces
                                    end
                                }
                                if ((found_faces > 0) && (item["signal"] == 1))
                                    emit({cmd: "found_faces"})
                                end
                            else
                                puts("add_edge: add_line() failed")
                            end
                        when "edit_add_face"
                            ptarray = Baroque::VRSketch._vlist2ptarray(item["v_list"])
                            grp = ctx.add_group
                            begin
                                face = grp.entities.add_face(ptarray)
                            rescue     # e.g. "not enough points, at least 3 required
                                face = nil
                            end
                            grp.explode if face
                        when "edit_delete"
                            picked.each { |entity|
                                begin
                                    entity.erase!
                                rescue TypeError     # item already deleted
                                end
                            }
                            picked = nil
                        when "edit_push_pull"
                            face = @model.find_entity_by_id(item["id"])
                            if face.is_a? Sketchup::Face
                                face.pushpull(item["distance"], item["copy"] > 0)
                            else
                                puts("push_pull: bad face id")
                            end
                        when "edit_explode"
                            grp = @model.find_entity_by_id(item["id"])
                            if !in_current_group(grp)
                                puts("explode: entity not in current group")
                            elsif ((grp.is_a? Sketchup::Group) ||
                                   (grp.is_a? Sketchup::ComponentInstance) ||
                                   (grp.is_a? Sketchup::Image))
                                picked.concat(grp.explode)
                            else
                                puts("edit_explode: bad group id")
                            end
                        when "edit_make_group"
                            picked.reject! { |x| x.deleted? }
                            grp = ctx.add_group(picked)
                            if grp
                                picked = [grp]
                            else
                                puts("edit_make_group: failed")
                            end
                        when "edit_make_component"
                            picked.reject! { |x| x.deleted? }
                            grp = ctx.add_group(picked)
                            if grp
                                picked = [_grp2component(grp)]
                            else
                                puts("edit_make_component: failed")
                            end
                        when "edit_turn_group_into_component"
                            if picked.length == 1
                                grp = picked[0]
                                if (grp && (grp.is_a? Sketchup::Group) && (!grp.deleted?))
                                    picked = [_grp2component(grp)]
                                else
                                    puts("edit_turn_group_into_component: not a group")
                                end
                            else
                                puts("edit_turn_group_into_component: selection count != 1")
                            end
                        when "edit_circle"
                            grp = ctx.add_group
                            n = item["numsegs"]
                            if item["smooth"] > 0
                                edges = grp.entities.add_circle([0, 0, 0], [0, 0, 1], 1, n)
                            else
                                edges = grp.entities.add_ngon([0, 0, 0], [0, 0, 1], 1, n)
                            end
                            grp.entities.add_face(edges)
                            tr = Geom::Transformation.new(item["delta"])
                            grp.transform!(tr)
                            picked = grp.explode
                            #puts({:picked => picked})
                            if (picked != nil) && (picked.length > 0)
                                picked = picked.grep(Sketchup::Drawingelement)
                            else
                                puts("edit_circle: sequence of operations failed")
                            end
                        when "edit_arc"
                            grp = ctx.add_group
                            grp.entities.add_arc(item["center"], item["xaxis"], item["normal"],
                                                 item["radius"], item["start_angle"], item["end_angle"],
                                                 item["numsegs"])
                            picked = grp.explode
                            if (picked != nil) && (picked.length > 0)
                                picked = picked.grep(Sketchup::Drawingelement)
                                picked.each { |e| e.find_faces if ((!e.deleted?) && (e.is_a? Sketchup::Edge)) }
                            else
                                puts("edit_arc: sequence of operations failed")
                            end
                        when "edit_new_material"
                            mat = emit_new_material(get_material_by_name(item["name"]))
                            @pending_mat_msgs.push({cmd: "new_mat", name: mat.object_id.to_s})
                        when "edit_delete_material"
                            mat = get_material_by_name(item["name"])
                            if mat != nil
                                @model.materials.remove(mat)
                            else
                                puts("edit_color_material: material not found")
                            end
                        when "edit_color_material"
                            mat = get_material_by_name(item["name"])
                            if mat != nil
                                mat.color = item["color"]    # integer 0x000000 - 0xffffff
                                mat.alpha = item["alpha"]    # float 0.0 - 1.0
                            else
                                puts("edit_color_material: material not found")
                            end
                        when "edit_rename_material"
                            mat = get_material_by_name(item["name"])
                            if mat != nil
                                if mat.name != item["new_display_name"]
                                    mat.name = unique_mat_display_name(item["new_display_name"])
                                end
                            else
                                puts("edit_rename_material: material not found")
                            end
                        when "edit_set_material_texture"
                            mat = get_material_by_name(item["name"])
                            if mat != nil
                                texture_file_name = item["texture_file_name"]
                                if (@on_quest && texture_file_name && (Baroque::VRSketch.allowed_actions[12] != 1))
                                    puts("edit_set_material_texture: 'Enable browsing image files from the Quest' is not set")
                                else
                                    mat.texture = texture_file_name    # possibly nil
                                end
                            else
                                puts("edit_set_material_texture: material not found")
                            end
                        when "edit_was_from_back"
                            picked_from_back = picked.length
                        when "edit_paint"
                            if item["matid"] >= 0
                                begin
                                    mat = @model.materials[item["matid"]]
                                rescue
                                    mat = nil    # index out of range?
                                end
                            elsif item["matid"] == -3
                                mat = @model.materials.load(item["matextfile"])
                                @pending_mat_msgs.push({cmd: "select_mat", name: mat.object_id.to_s})
                            elsif item["matid"] == -2
                                mat = Sketchup::Color.new(item["matcolor"])   # deprecated
                            else
                                mat = nil
                            end
                            # We assume that Sketchup::Material objects can be assigned to faces only
                            # if they are in the enumeration of .materials, which skips over images.
                            # Some versions of SketchUp complain explicitly when trying to assign
                            # image materials to faces.  In general it is a bad idea to do so.
                            if mat.is_a? Sketchup::Material
                                found = false
                                @model.materials.each { |m1| found |= (m1 == mat) }
                                mat = nil if !found
                            end
                            picked[0...picked_from_back].each { |entity| entity.back_material = mat }
                            picked[picked_from_back...picked.length].each { |entity| entity.material = mat }
							if ((mat.is_a? Sketchup::Material) && (item["v_list"] != nil) && (item["v_list"].length > 0))
								#
								# v_list is three pairs of 3D points, like the ptarray argument
								# of Face.position_material
								ptarray1 = Baroque::VRSketch._vlist2ptarray(item["v_list"])
								pts_on_plane = []
								ptarray1.each_slice(2) { |pt, uv| pts_on_plane.push(pt) }
								source_plane = Geom.fit_plane_to_points(pts_on_plane)
								source_normal = source_plane[0...3]
								#
								picked.each_with_index { |entity, index|
									next if !(entity.is_a? Sketchup::Face)
									on_front = (index >= picked_from_back)
									ptarray = ptarray1.clone
									if ((entity.normal.angle_between(source_normal).abs > 1e-5) &&
									    ((Geom::Vector3d.new(0,0,0) - entity.normal).angle_between(source_normal).abs > 1e-5))
										fnormal = entity.normal
										fnormal = Geom::Vector3d.new(0,0,0) - fnormal if !on_front
										line = Geom.intersect_plane_plane(source_plane, entity.plane)
										# rotate ptarray along the line to make source_plane equal to entity.plane
										source_transform = Geom::Transformation.new(line[0], line[1], source_normal)
										target_transform = Geom::Transformation.new(line[0], line[1], fnormal)
										transform = target_transform * (source_transform.inverse)
										#puts({source_transform: source_transform, target_transform: target_transform, transform: transform})
										ptarray.each_slice(2).with_index { |slice, index|
											ptarray[index * 2] = transform * ptarray[index * 2]
										}
									end
									#puts({ptarray: ptarray})
									entity.position_material(mat, ptarray, on_front)
								}
							end
                        when "edit_setuv"
                            entity = @model.find_entity_by_id(item["id"])
                            if !in_current_group(entity)
                                puts("setuv: entity not in current group")
                            elsif entity.is_a? Sketchup::Face
                                if item["front"] != 0
                                    mat = entity.material
                                else
                                    mat = entity.back_material
                                end
                                ptarray = Baroque::VRSketch._vlist2ptarray(item["v_list"])
                                ptarray = [ptarray[0], [0, 0, 0],
                                           ptarray[1], [1, 0, 0],
                                           ptarray[2], [0, 1, 0]]
								#puts({ptarray: ptarray})
                                if !entity.position_material(mat, ptarray, item["front"] != 0)
									puts("setuv: failed to set the texture position")
								end
                            else
                                puts("setuv: bad face id")
                            end
                        when "edit_add_curve"
                            ptarray = Baroque::VRSketch._vlist2ptarray(item["v_list"])
                            grp = ctx.add_group
                            grp.entities.add_curve(ptarray)
                            grp.explode
                            # XXX with real curve support, could add the result into 'picked'
                        when "edit_hide"
                            picked.each { |entity|
                                begin
                                    entity.hidden = true
                                rescue TypeError     # item already deleted
                                end
                            }
                            picked = nil
                        when "edit_unhide_all"
                            ctx.each { |entity|
                                if ((entity.is_a? Sketchup::Drawingelement) && (entity.hidden?))
                                    entity.hidden = false
                                end
                            }
                        when "edit_intersect"
                            # first, try "intersect faces with selection"
                            picked.reject! { |x| x.deleted? }
                            grp = ctx.add_group(picked)
                            identity = Geom::Transformation.new
                            new_edges = nil
                            if picked.length > 1
                                new_edges = grp.entities.intersect_with(true, identity, grp.entities, grp.transformation, false, grp)
                            end
                            if ((new_edges != nil) && (new_edges.length > 0))
                                # success
                            else
                                # did nothing, so now try "intersect faces with model"
                                new_edges = @model.entities.intersect_with(true, identity, grp.entities, grp.transformation, false, grp)
                            end
                            picked = grp.explode
                        when "edit_reverse"
                            picked.each { |entity|
                                next if entity.deleted?
                                entity.reverse! if entity.is_a? Sketchup::Face
                            }
                        when "edit_add_component"
                            cdef1 = @model.find_entity_by_id(item["cdef_id"])
                            if cdef1.is_a? Sketchup::ComponentDefinition
                                tr = Geom::Transformation.new(item["center"])
                                cinst = ctx.add_instance(cdef1, tr)
                                if cinst
                                    picked = [cinst]
                                else
                                    puts("add_component: failed")
                                end
                            else
                                puts("add_component: invalid cdef_id")
                            end
                        when "edit_download_model"
                            cdef1 = @model.definitions.load_from_url(item["model_url"])
                            if cdef1.is_a? Sketchup::ComponentDefinition
                                tr = Geom::Transformation.new(item["center"])
                                cinst = ctx.add_instance(cdef1, tr)
                                if cinst
                                    picked = [cinst]
                                else
                                    puts("edit_download_model: failed")
                                end
                            else
                                puts("download_model: download failed")
                            end
                        when "edit_set_note"
                            dict = @model.attribute_dictionary("Baroque_VRSketch_notes", true)
                            key = item["key"].to_s
                            if item["erase"] != 0
                                dict.delete_key("p" + key)
                                dict.delete_key("c" + key)
                            else
                                dict["p" + key] = item["pts"]
                                dict["c" + key] = item["color"]
                            end
                            dict["version"] = Time.now.to_f
                        when "edit_save_old_selection"
                            old_selection = _pick_entity_ids(picked)
                            picked = []
						when "edit_move_axes"
							# Mess: axes.set() has lasting effect only if we don't have any group or component
							# open.  If we do, then it does changes the axes but the changes are lost if we close
							# and reopen the component.  The way to change to axes in a component is to move all
							# instances of this component as required, and then compensate by moving all entities
							# in the component an opposite amount.
							definition = @model.active_entities.parent
							if definition.is_a? Sketchup::ComponentDefinition
								origin1 = @model.axes.origin
								xaxis1, yaxis1, zaxis1 = @model.axes.axes
								origin2 = item["org"]
								xaxis2 = Geom::Vector3d.new(item["x"]).normalize
								yaxis2 = Geom::Vector3d.new(item["y"]).normalize
								zaxis2 = Geom::Vector3d.new(item["z"]).normalize
								
								tr1 = Geom::Transformation.new(xaxis1, yaxis1, zaxis1, origin1)
								tr2 = Geom::Transformation.new(xaxis2, yaxis2, zaxis2, origin2)
								tr_direct = tr2 * tr1.inverse
								tr_inverse = tr1 * tr2.inverse
				                push_all_entities_from(definition)
								definition.entities.transform_entities(tr_inverse, definition.entities.to_a)
							    definition.instances.each { |i|
									push_update(i)
									i.transformation = i.transformation * tr_direct
								}
								# now the change is done, but sketchup is in a confused state.  Fix it by
								# closing and re-opening the group...  This shows up as additional undo steps,
								# but it's the best I could figure out so far.
								current_group = @model.active_path[-1]
		                        @model.selection.clear
		                        @model.close_active
			                    Baroque::VRSketch.hack_open_group(@model, current_group)
							else
								@model.axes.set(item["org"], item["x"], item["y"], item["z"])
							end
                        when "edit_rename_component"
                            cdef = @model.definitions[item["name"]]
                            cdef.name = item["new_name"]
                        when "edit_delete_component"
                            cdef = @model.definitions[item["name"]]
                            if cdef.is_a? Sketchup::ComponentDefinition
                                erase_component_definition(cdef)
                            else
                                puts("edit_delete_component: unknown name")
                            end
                        when "edit_add_scene"
                            page = @model.pages.add
                            page.camera.set(item["eye"], item["target"], Z_AXIS)
                        when "edit_rename_scene"
                            page = _getscene(item)
                            page.name = item["new_name"] if page != nil
                        when "edit_delete_scene"
                            page = _getscene(item)
                            @model.pages.erase(page) if page != nil
                        when "edit_change_layer"
                            change_layer(item, picked)
                        when "edit_add_layer"
                            name = @model.layers.unique_name("Layer")
                            layer = @model.layers.add(name)
                            if name == "Layer"       # we always want Layer<Number>, not just Layer
                                name = @model.layers.unique_name("Layer")
                                @model.layers.remove(layer)
                                @model.layers.add(name)
                            end
                        when "edit_rename_layer"
                            target_layer = _getlayer(item)
                            if target_layer != nil
                                target_layer.name = item["new_name"]
                            end
                        when "edit_delete_layer"
                            target_layer = _getlayer(item)
                            if target_layer != nil
                                @model.layers.remove(target_layer, item["remove_geometry"] > 0)
                            end
                        when "edit_set_animation"
                            set_animation(item)
                        when "edit_clipboard_cut"
                            clipboard_cut(ctx, picked)
                        when "edit_clipboard_paste_in_place"
                            clipboard_paste_in_place(ctx)
                        when "edit_add_text"
                            arrowend = Geom::Point3d.new(item["arrowend"])
                            textpivot = Geom::Point3d.new(item["textpivot"])
                            text = ctx.add_text(item["text"], arrowend, textpivot - arrowend)
                            _try_to_attach_text_label(text)
                            picked = [text]
                        when "edit_move_text"
                            text = @model.find_entity_by_id(item["id"])
                            cmd_audio_record(text, item)
                            if text.is_a? Sketchup::Text
                                default_text = ""
                                if ((item["settext"] != nil) && (item["settext"] != 0))
                                    newtext = item["text"]
                                else
                                    newtext = text.text
                                    default_text = item["text"] if item["record"] != nil
                                end
                                newtext = newtext[1...newtext.length] if newtext.start_with? SPEAKER_UNICODE
                                if text.attribute_dictionary("Baroque_VRSketch_audio") != nil
                                    newtext = "" if newtext == default_text
                                    newtext = SPEAKER_UNICODE + newtext
                                else
                                    newtext = default_text if newtext == ""
                                end
                                text.text = newtext
                                textpivot = item["textpivot"]
                                textpivot = [] if textpivot == nil
                                arrowend = item["arrowend"]
                                arrowend = [] if textpivot == nil
                                if arrowend.length == 3
                                    textpivot = (text.point + text.vector).to_a if textpivot.length != 3
                                    text.point = arrowend
                                    _try_to_attach_text_label(text)
                                end
                                if textpivot.length == 3
                                    text.vector = Geom::Point3d.new(textpivot) - text.point
                                end
                                flags = item["flags"]
                                if ((flags != nil) && (flags != 0))
                                    text.display_leader = ((flags & 32) != 0)
                                    if (flags & 128) != 0
                                        text.arrow_type = 2
                                    elsif (flags & 64) != 0
                                        text.arrow_type = 3 if text.arrow_type < 3
                                    else
                                        text.arrow_type = 0
                                    end
                                end
                                text.erase! if item["erase"] == 1
                            else
                                puts("edit_move_text: no Text with this id")
                            end
                        else
                            puts("geom_edit_apply: unknown " + item["cmd"])
                        end
                    }
                    @model.active_layer = old_layer if layer0 != old_layer
                rescue
                    @model.abort_operation
                    @model.active_layer = old_layer if layer0 != old_layer
                    emit({cmd: "cancelwait"})
                    raise
                end
                @transaction_committed = false
                @model.commit_operation
                if @transaction_committed
                    h = { msg: op_name }
                    if old_selection != nil
                        h[:old_sel] = old_selection
                    end
                    @undo_history[-1] = h
                    if cmd_apply["selected"] > 0
                        @keep_selection = picked
                        if picked != nil
                            h[:new_sel] = _pick_entity_ids(picked)
                        end
                    else
                        @keep_selection = nil
                    end
                end
            end

            def emit_feedback(id)
                if @pending_updates == nil
                    emit({cmd: "cancelwait", id: id})
                else
                    @cancel_wait_id = id if @cancel_wait_id < id
                end
            end

            def new_mat_from_scratch()
                mat = @model.materials.add
                # seems that 'mat' is in an invalid state in SU2018.  It displays as white in
                # the materials list, but as black when applied to a face or sent to VRSketch.
                mat.color = "white"
                return mat
            end

            def emit_new_material(mat)
                # now, how do we duplicate the material 'mat'?  Take a guess...
                filename_skm = nil
                if mat != nil
                    filename_skm = Baroque::VRSketch.temporary_filename(".skm")
                    if (!(mat.save_as(filename_skm)))
                        puts("edit_new_material: cannot write material to " + filename_skm)
                        return new_mat_from_scratch
                    end
                    matextfile = filename_skm
                else
                    return new_mat_from_scratch   # make a new default material
                end

                # here, 'mat' is the current material in the model, and 'matextfile' is the
                # name of a '.skm' file that would reimport the very same material.  We want
                # instead to import a copy of it.  If we mess with 'mat.name', it will work.
                newmat = @model.materials.add
                old_name = mat.name
                new_name = newmat.name
                @model.materials.remove(newmat)

                begin
                    mat.name = old_name + "XXXVRSketchTMP~"
                    newmat = @model.materials.load(matextfile)
                    newmat.name = new_name   # a fresh name, != old_name
                    mat.name = old_name
                rescue
                    mat.name = old_name
                    puts("failed to load " + matextfile)
                    newmat = new_mat_from_scratch
                end
                File.delete(filename_skm) if filename_skm != nil
                return newmat
            end

            def get_material_by_name(name)
                # NB. the "name" of a material here is mat.object_id.to_s, something unique
                # and constant.  It is not "material.name".
                i = @model.materials.length
                while i > 0
                    i -= 1
                    mat = @model.materials[i]
                    return mat if mat.object_id.to_s == name
                end
                return nil
            end

            def unique_mat_display_name(name)
                # NB. the "name" parameter here is the new user-defined name
                loop do
                    i = @model.materials.length
                    while i > 0
                        i -= 1
                        mat = @model.materials[i]
                        if mat.name == name
                            name = name + " (dup)"
                            i = -1
                        end
                    end
                    break if i == 0
                end
                return name
            end

            def transaction_commit()
                while @undone_operations > 0
                    @undo_history.pop      # no-op if already empty
                    @undone_operations -= 1
                end
                @undo_history.push({msg: ""})
                @transaction_committed = true
                _enable_updates
            end

            def transaction_undo()
                @keep_selection = nil
                @undone_operations += 1
                if @undone_operations > @undo_history.length
                    @undo_history.insert(0, {msg: ""})
                else
                    h = @undo_history[@undo_history.length - @undone_operations]
                    @keep_selection = h[:old_sel]    # nil if not found; otherwise, a list of IDs
                end
                push_update_whole_current_group
            end

            def transaction_redo()
                @keep_selection = nil
                if @undone_operations > 0
                    index = @undo_history.length - @undone_operations
                    if index >= 0
                        h = @undo_history[index]
                        @keep_selection = h[:new_sel]    # nil if not found; otherwise, a list of IDs
                    end
                    @undone_operations -= 1
                end
                push_update_whole_current_group
            end

            def emit_undo_history(cmd)
                Baroque::VRSketch.stat_event("undo redo")
                steps = cmd["delta"]
                while steps < 0
                    Sketchup.send_action("editUndo:")
                    steps += 1
                end
                while steps > 0
                    Sketchup.send_action("editRedo:")
                    steps -= 1
                end
                _enable_updates   # because send_action() above doesn't have an immediate effect,
                                  # but we want the following "feedback" message to not think
                                  # there is nothing to do
            end

            def open_close_group(cmd)
                Baroque::VRSketch.stat_event("open close group")
                @model.selection.clear
                @transaction_committed = false
                if cmd["id"] == 0
                    count = cmd["close_group"]
                    count = 1 if count == nil
                    while ((count > 0) && (@model.active_entities != @model.entities))
                        break if !@model.close_active
                        count -= 1
                    end
                    op_name = "Close"
                else
                    grp = @model.find_entity_by_id(cmd["id"])
                    emit_me = Baroque::VRSketch.hack_open_group(@model, grp)
                    _enable_updates   # same as emit_undo_history(): this is done by send_action()
                    op_name = "Open"
                    emit(emit_me) if emit_me != nil
                end
                # We don't have @transaction_committed immediately, because of send_action().
                # Wait a little bit
                Baroque::VRSketch.delay(0.0) {
                    if @transaction_committed
                        @undo_history[-1] = {msg: op_name}
                    end
                }
            end

            def refresh_cdef(cmd)
                cdef = @model.find_entity_by_id(cmd["id"])
                if cdef.is_a? Sketchup::ComponentDefinition
                    return if cdef.deleted?
                    push_all_entities_from(cdef)
                end
            end

            def got_my_window_info(cmd)
                @hwnd = cmd["hwnd"] if ((cmd["hwnd"] != 0) && (@hwnd == nil))
                @vrsketch_version = cmd["vrsketch_version"]
            end

            def shadow_info_changed(datetime_only)
                return if !@model.shadow_info   # occurs on Mac?
                msg = {cmd: "shadow_info"}
                shadow_v = @model.shadow_info["SunDirection"].to_a
                if !datetime_only
                    shadow_v.push(0.0, @model.shadow_info["ShadowTime_time_t"])
                    fog_info = [0.0]
                    if @model.rendering_options["DisplayFog"]
                        if @model.rendering_options["FogUseBkColor"]
                            fog_color = @model.rendering_options["BackgroundColor"]
                        else
                            fog_color = @model.rendering_options["FogColor"]
                        end
                        fog_info = [3.0, fog_color.to_i,
                            @model.rendering_options["FogStartDist"],
                            @model.rendering_options["FogEndDist"]]
                    end
                    msg[:transform] = fog_info
                end
                msg[:v] = shadow_v
                emit(msg)
            end

            def change_shadow_time(cmd)
                # we might receive many of these messages in succession.  We'll only answer to
                # the last one pending now
                must_start_timer = (@shadow_time_changed == nil)
                @shadow_time_changed = cmd["time_t"]    # integer

                if must_start_timer
                    Baroque::VRSketch.delay(0.05) {
                        new_time = @shadow_time_changed
                        @shadow_time_changed = nil
                        # update this setting, which calls 'shadow_info_changed(datetime_only: true)'
                        @model.shadow_info["ShadowTime_time_t"] = new_time if new_time != nil
                    }
                end
            end

            def units_options_changed()
                emit({
                    cmd: "setconfig",
                    name: "length_format",
                    v: [@model.options["UnitsOptions"]["LengthFormat"]],
                })
            end

            def allowed_actions_changed(change_index)
                # 'allowed_actions' is global, but items 6 and 7 are ignored and overwritten with
                # the model-settings-specific values here, for backward-compatibility
                aa = Baroque::VRSketch.allowed_actions.dup
                aa[7] = 0      # hide edges
                aa[7] = 1 if @model.rendering_options["EdgeDisplayMode"] == 0
                aa[6] = 0      # hide rest of model
                aa[6] = 1 if @model.rendering_options["InactiveHidden"]
                emit({
                    cmd: "setconfig",
                    name: "allowed_actions",
                    v: aa,
                })
                if change_index == 6    # hide rest of model
                    _enable_updates
                end
            end

            def _push_updates_from_scene(scene)
                if ((scene != nil) && (!scene.deleted?) && (scene.hidden_entities != nil))
                    upd = _enable_updates_cdef(@model)
                    if upd != nil
                        scene.hidden_entities.each { |obj| upd.push(obj) }
                    end
                end
            end

            def start_frame_change(active_model, prev_scene, to_scene)
                return if active_model != @model
                v = Baroque::VRSketch.compute_teleport_destination(to_scene, @frame_change_speed)
                emit({
                    cmd: "teleport",
                    v: v,
                })
                _push_updates_from_scene(prev_scene)
                _push_updates_from_scene(to_scene)
            end

            def change_to_scene(cmd)
                # use index-based scene because the names are not unique at all
                begin
                    target_page = @model.pages[cmd["index"]]
                rescue IndexError
                    return   # never raises IndexError so far, but you never know.  It does for @model.layers
                end
                return if !(target_page.is_a? Sketchup::Page)
                old_show_transition = @model.options["PageOptions"]["ShowTransition"]
                begin
                    @model.options["PageOptions"]["ShowTransition"] = false
                    @frame_change_speed = 1
                    @model.pages.selected_page = target_page
                ensure
                    @frame_change_speed = 0
                    @model.options["PageOptions"]["ShowTransition"] = old_show_transition
                end
            end

            def send_scenes_list()
                labels = []
                @model.pages.each { |page| labels.push(page.label) }
                emit({
                    cmd: "scenes",
                    names: labels,
                })
                # Now send the location of each scene, too.  Note that the location will only
                # be used if the connection to Sketchup is severed---or in cloud models
                i = 0
                @model.pages.each { |page|
                    emit({
                        cmd: "set_scene",
                        id: i,
                        v: Baroque::VRSketch.compute_teleport_destination(page, 1),
                    })
                    i = i + 1
                }
            end

            def _getlayer(cmd)
                begin
                    target_layer = @model.layers[cmd["name"]]
                rescue IndexError
                    target_layer = nil
                end
                return target_layer if target_layer.is_a? Sketchup::Layer
                puts("edit_layer_xxx: layers list was modified in the background, not applying change")
                return nil
            end

            def change_layer(cmd, picked)
                target_layer = _getlayer(cmd)
                return if target_layer == nil
                # use name-based layers because the names seem to be unique, and on the other hand
                # the indexes seem to change sometimes..?
                case cmd["op"]
                #when "activate"
                #    target_layer.visible = true if !target_layer.visible?   # may record undo/redo
                #    @model.active_layer = target_layer    # no undo/redo
                when "hide"
                    target_layer.visible = false    # records undo/redo
                when "show"
                    target_layer.visible = true     # records undo/redo
                when "apply"
                    picked.each { |grp|
                        grp.layer = target_layer if !grp.deleted?
                    }
                end
            end

            def send_layers_list()
                names = []
                visibles = []
                @model.layers.each { |layer|
                    names.push(layer.name)
                    visible = 0
                    visible = 1 if layer.visible?
                    visibles.push(visible)
                }
                emit({
                    cmd: "layers",
                    names: names,
                    id_list: visibles,
                })
            end

            def call_later(attr)
                if @general_pending_ops.size == 0
                    Baroque::VRSketch.delay(0.01) {
                        ops = @general_pending_ops
                        @general_pending_ops = {}
                        ops.each_key { |attr|
                            self.send(attr)
                        }
                    }
                end
                @general_pending_ops[attr] = true
            end

            def send_components_list()
                # this is also called via call_later, so we must check that the model is still valid
                return if !@model.valid?
                return if @model.definitions == nil
                ids = []
                names = []
                bounds = []
                @model.definitions.each { |cdef|
                    next if ((cdef.group?) || (cdef.hidden?) || (cdef.image?))
                    ids.push(cdef.entityID)
                    names.push(cdef.name)
                    bounds.concat(cdef.bounds.min.to_a)
                    bounds.concat(cdef.bounds.max.to_a)
                    bounds.concat(cdef.insertion_point.to_a)
                }
                emit({cmd: "components", id_list: ids, names: names, transform: bounds})
            end

            def send_component_image(cmd)
                cdef_id = cmd["cdef_id"]
                cdef = @model.find_entity_by_id(cdef_id)
                return if !cdef.is_a? Sketchup::ComponentDefinition
                #
                filename_jpg = Baroque::VRSketch.temporary_filename(".jpg")
                raw = nil
                begin
                    if cdef.save_thumbnail(filename_jpg)
                        raw = File.binread(filename_jpg)
                        File.delete(filename_jpg)
                    end
                rescue
                    # ignore errors
                end
                return if !raw
                emit_rawdata(raw, {
                    cmd: "component_thumbnail",
                    id: cdef_id,
                })
            end

            def set_animation(cmd)
                grp_id = cmd["group_id"]
                mask = cmd["mask"]
                grp = @model.find_entity_by_id(grp_id)
                return if !((grp.is_a? Sketchup::ComponentInstance) || (grp.is_a? Sketchup::Group))

                _set_dynamic_attr(grp, "_lastmodified", DateTime.now.strftime("%Y-%m-%d %H:%M"))
                push_update(grp)

                if (mask & 4) != 0     # M_REMOVE
                    # delete_attribute("dynamic_attributes", "onclick") appears to have no effect.
                    # No clue why.  Setting an empty string seems to be enough to remove the
                    # ability to click on it with SketchUp's own "interact" tool, though.
                    _set_dynamic_attr(grp, "onclick", "")
                    return
                end

                func, varname, values = get_animation_func(grp)
                func = "ANIMATE" if func == nil
                initial_values = ""
                positions = values
                if func.start_with?("ANIMATECUSTOM")
                    if values.length > 3
                        positions = values[3...values.length]
                        initial_values = values[0...3].join(",")
                        initial_values += ","
                    else
                        func = "ANIMATE"
                    end
                end

                if (mask & 1) != 0     # M_SET_AMOUNT
                    varindex = cmd["varindex"]
                    amount = cmd["amount"]
                    varname = ["X", "Y", "Z", "ROTX", "ROTY", "ROTZ"][varindex]
                    if varindex < 3
                        current = grp.transformation.origin[varindex].to_f
                        unit = get_animation_unit(grp, true)
                    else
                        current = grp.transformation.send(varname.downcase)   # .rotx, .roty, .rotz
                    end

                    if amount > 0
                        state1 = 0
                        low_bound = current
                        high_bound = current + amount
                    else
                        state1 = 1
                        low_bound = current + amount
                        high_bound = current
                    end

                    positions = [low_bound, high_bound].map { |b|
                        # pfff it seems that angles must be integers, because SketchUp always
                        # assign integers to them, and if the bounds are not integers they will
                        # never be considered equal and SketchUp won't know the state
                        if varindex < 3
                            b = b / unit
                            '%.4f' % b
                        else
                            b.round.to_s
                        end
                    }

                    _set_dynamic_attr(grp, "_onclick_state1", state1)
                    _set_dynamic_attr(grp, "_" + varname.downcase + "_formula", nil)
                    _set_dynamic_attr(grp, varname.downcase, current)
                end

                if (mask & 2) != 0     # M_SET_SPEED
                    speed = cmd["speed"]
                    easein = cmd["easein"]
                    easeout = cmd["easeout"]
                    func = "ANIMATECUSTOM"
                    if ((easein == 0) && (easeout == 100))
                        func = "ANIMATE" if speed == 0.5
                        func = "ANIMATEFAST" if speed == 0.25
                        func = "ANIMATESLOW" if speed == 1
                    end
                    initial_values = ""
                    initial_values = speed + "," + easein + "," + easeout + "," if func == "ANIMATECUSTOM"
                end

                _set_default_dynamic_attr(grp, "_formatversion", 1.0)
                _set_default_dynamic_attr(grp, "_has_movetool_behaviors", 0.0)
                _set_default_dynamic_attr(grp, "_hasbehaviors", 1.0)
                _set_default_dynamic_attr(grp, "_iscollapsed", "false")
                _set_default_dynamic_attr(grp, "_name", "Group")
                _set_default_dynamic_attr(grp, "_onclick_label", "onClick")
                #
                _set_dynamic_attr(grp, "onclick", func + '("' + varname + '",' + initial_values +
                                                  positions.join(",") + ')')
            end

            def model_filename_changed()
                return if @tutorial
                emit({
                    cmd: "model_filename",
                    name: @model.path,    # possibly an empty string
                })
            end

            def save_model(cmd)
                if Baroque::VRSketch.allowed_actions[11] == 0
                    suggested = ""
                    ok = -2
                else
                    suggested = @model.path
                    if ((!suggested) || (suggested == ""))
                        dir = Dir.pwd
                        i = 1
                        while true
                            suggested = File.join(dir, "Untitled VR " + i.to_s + ".skp")
                            break if !File.exist? suggested
                            i += 1
                        end
                    end
                    #puts({suggested: suggested, cmdfilename: cmd['filename']})
                    if cmd['filename'] == suggested
                        ok = 0
                        begin
                            if @model.save(suggested)
                                ok = 1
                                suggested = @model.path    # should be equal, but '/' vs '\' issues
                            end
                        rescue
                        end
                    else
                        ok = -1
                    end
                end
                emit({
                    cmd: "model_filename_save",
                    name: suggested,
                    id: ok,
                })
            end

            if Baroque::VRSketch::IS_WINDOWS
                def _windows_generate_filename(video)
                    sketchlib = Baroque::VRSketch.open_sketchlib
                    result_filename = Baroque::VRSketch.temporary_filename(".txt")
                    r = -9
                    r = sketchlib[:generate_filename].call(video, result_filename) if sketchlib
                    res = nil
                    if r >= -1
                        begin
                            f = open(result_filename)
                            res = f.read
                            f.close
                            File.delete(result_filename)
                        rescue
                        end
                    end
                    if r < 0
                        puts("generate_filename: error " + r.to_s)
                        puts(res) if res != nil
                        return nil
                    end
                    return res
                end

                def _next_photo_file_name(is_video)
                    video = 0
                    video = 1 if is_video
                    return _windows_generate_filename(video)
                end
            else
                def _next_photo_file_name(is_video)
                    path = File.expand_path("~/Pictures/VR Sketch")
                    Dir.mkdir(path) if !(Dir.exist? path)
                    head = "Photo"
                    head = "Video" if is_video
                    head += " " + Date.today.to_s + " ("
                    fullhead = File.join(path, head)
                    extension = "jpg"
                    extension = "mp4" if is_video
                    tail = ")." + extension

                    if @photo_last_used_number != nil
                        @photo_last_used_number += 1
                        result = "%s%03d%s" % [fullhead, @photo_last_used_number, tail]
                        return result if !(File.exist? result)
                    end

                    # Generate a file name of type "head-NUMBER-tail".  For NUMBER we look at all existing
                    # files with a similarly formatted name, and we pick one more than the largest NUMBER
                    # that already exists.
                    @photo_last_used_number = 0 if @photo_last_used_number == nil
                    Dir.foreach(path) { |fn|
                        fn = fn.downcase
                        next unless fn.start_with?(head.downcase)
                        next unless fn.end_with?(tail.downcase)
                        num = fn[head.length...(fn.length - tail.length)].to_i
                        @photo_last_used_number = num if @photo_last_used_number < num
                    }

                    # If the loop above failed for any reason to list existing files, skip them now
                    while true
                        @photo_last_used_number += 1
                        result = "%s%03d%s" % [fullhead, @photo_last_used_number, tail]
                        return result if !(File.exist? result)
                    end
                end
            end

            def save_photo(cmd, rawdata)
                # hackish: use a temporary file which will contain a short string, which is
                # itself the real name of the photo/video file to generate
                filename = _next_photo_file_name(cmd["video"] == 1)
                if filename == nil
                    Sketchup.set_status_text("Failed to save photo/video!")
                    return
                end

                f = open(filename, "wb")
                f.write(rawdata)
                f.close

                Sketchup.set_status_text("Saved: " + filename)
            end

            def get_base_skm_directories
                names = []
                name2 = Sketchup.find_support_file("Materials")        # C:/Users/...
                names.push(name2) if name2 != nil
                
                # how do we get the Materials folder from C:/ProgramData/...?  We can't directly,
                # but we can enumerate all known folders in the hope that at least one of them is
                # not hidden by another folder in C:/Users/.  The actual directory names appear
                # not to be translated in other languages, too, even if the UI translates them in
                # SketchUp (too bad, we can't translate them in VR Sketch).

                # Note: this is NOT the list of directories that will show up.  This is only a
                # list that is large enough so that we're quasi sure that NOT ALL OF THEM are
                # duplicates from C:/Users/.
                all_names = ["3d Printing",
                             "Asphalt and Concrete",
                             "Brick, Cladding and Siding",
                             "Carpet, Fabrics, Leathers, Textiles and Wallpaper",
                             "Colors",
                             "Colors-Named",
                             "Glass and Mirrors",
                             "Landscaping, Fencing and Vegetation",
                             "Metal",
                             "Patterns",
                             "Roofing",
                             "Stone",
                             "Synthetic Surfaces",
                             "Tile",
                             "Water",
                             "Window Coverings",
                             "Wood"]
                all_names.each { |name|
                    fn = Sketchup.find_support_file(name, "Materials")
                    next if fn == nil
                    fn = File.dirname(fn)
                    names.push(fn) if !(names.member? fn)
                }
                return names
            end

            if Baroque::VRSketch::IS_WINDOWS
                def skm_load(cmd)
                    # Quest: helper to list and load .skm files on the PC side
                    sketchlib = Baroque::VRSketch.open_sketchlib
                    name = cmd["name"]
                    dirs = get_base_skm_directories.join("\n")
                    #puts("SKM_LOAD: " + name)
                    result_filename = Baroque::VRSketch.temporary_filename(".skm_load")
                    r = -9
                    r = sketchlib[:load_skm_directory].call(name, dirs, result_filename) if sketchlib
                    if r < 0
                        Sketchup.set_status_text("Failed to list directory containing SKM files (materials)")
                        puts("skm_load error " + r.to_s)
                        if r == -1
                            begin
                                f = open(result_filename)
                                puts(f.read)
                                f.close
                                File.delete(result_filename)
                            rescue
                            end
                        end
                        return
                    end
                    f = open(result_filename, "rb")
                    raw = f.read
                    f.close
                    #puts("GOT: bytes: " + raw.size.to_s)
                    emit_rawdata(raw, {
                        cmd: "skm_loaded",
                        name: name,
                    })
                end
            else
                # OS X
                def skm_load(cmd)
                    name = cmd["name"]
                    if ((name == nil) || (name == ""))
                        # list all .skm files
                        skm_base_dirs = get_base_skm_directories()
                        seen = Set[]
                        result = []
                        _skm_list_files_rec(skm_base_dirs, seen, result)
                        result.map! { |fullname| fullname + "\n" }
                        raw = result.join
                    else
                        return unless ((name.downcase.end_with? ".skm") && (File.exist? name))
                        raw = nil
                        begin
                            dirname = Baroque::VRSketch.temporary_filename(".d")
                            Dir.mkdir(dirname) if !(Dir.exist? dirname)

                            f = File.open(name, "rb")
                            skm_content = f.read
                            f.close
                            f = File.open(File.join(dirname, "skm.zip"), "wb")
                            f.write(skm_content)
                            f.close
                            skm_content = nil

                            thumbnail = File.join(dirname, "doc_thumbnail.png")
                            File.delete(thumbnail) if File.exist? thumbnail

                            system("cd '#{dirname}' && unzip skm.zip doc_thumbnail.png")

                            f = open(thumbnail, "rb")
                            raw = f.read
                            f.close
                            File.delete(File.join(dirname, "skm.zip"))
                            File.delete(thumbnail)
                            Dir.rmdir(dirname)
                        rescue => e
                            puts e
                        end
                        return if !raw
                    end
                    emit_rawdata(raw, {
                        cmd: "skm_loaded",
                        name: name,
                    })
                end

                def _skm_list_files_rec(directories, seen, result)
                    directories.each { |dir1|
                        next if (seen.add? dir1) == nil    # already seen
                        subdirs = []
                        begin
                            Dir.entries(dir1).each { |basename|
                                next if basename.start_with? "."
                                fullname = File.join(dir1, basename)
                                if File.file? fullname
                                    result.push(fullname) if fullname.downcase.end_with? ".skm"
                                elsif File.directory? fullname
                                    subdirs.push(fullname)
                                end
                            }
                        rescue => e
                            puts e   # and continue
                        end
                        _skm_list_files_rec(subdirs, seen, result)
                    }
                end
            end

            def _lf_add_directory(directories, path)
                i = 0
                while i < directories.length
                    return i if File.identical?(directories[i + 1], path)
                    i += 3
                end
                dirname = File.dirname(path)
                if ((dirname == "") || (dirname == path))
                    new_index = directories.length
                    flags = "s"
                else
                    i = _lf_add_directory(directories, dirname)
                    new_index = i + 3
                    flags = directories[i]
                    flags = "" if flags == "s"
                    flags += ">"
                end
                directories.insert(new_index, flags, path, File.basename(path))
                return new_index
            end

            def cmd_list_files(cmd)
                if Baroque::VRSketch.allowed_actions[12] != 1
                    emit({ cmd: "list_files" })
                    return
                end

                path = cmd["path"]
                if ((path == nil) || (path == ""))
                    path = Baroque::VRSketch.last_used_directory
                    if ((path == nil) || (path == ""))
                        path = File.dirname(@model.path)   # possibly "."
                        path = File.absolute_path(path)
                    end
                end

                directories = []

                if Baroque::VRSketch::IS_WINDOWS
                    # use "\" systematically, on Windows
                    path = path.gsub("/", "\\") 
                    drives = _windows_generate_filename(3)
                    drives = "C:\\" if drives == nil
                    drives.split(";").each { |drive|
                        directories.push("s", drive, drive)
                    }
                    pictures_folder = _windows_generate_filename(2)
                else
                    directories.push("s", File.expand_path("~"), "Home")
                    directories.push("s", "/", "Root")
                    pictures_folder = File.expand_path("~/Pictures")
                end
                #puts("PATH: " + path)
                if Baroque::VRSketch.prev_used_directory == nil
                    Baroque::VRSketch.prev_used_directory = Baroque::VRSketch.last_used_directory
                end
                if Baroque::VRSketch.prev_used_directory != nil
                    newdir_plus_slash = path
                    slash = "/"
                    slash = "\\" if Baroque::VRSketch::IS_WINDOWS
                    newdir_plus_slash += slash if (!(newdir_plus_slash.end_with? slash))
                    #puts("newdir_plus_slash: " + newdir_plus_slash)
                    if (!(Baroque::VRSketch.prev_used_directory.start_with? newdir_plus_slash))
                        Baroque::VRSketch.prev_used_directory = nil
                    end
                end
                #puts("prev_used_directory: " + Baroque::VRSketch.prev_used_directory.to_s)

                _lf_add_directory(directories, path)
                _lf_add_directory(directories, Baroque::VRSketch.prev_used_directory) if Baroque::VRSketch.prev_used_directory

                directories.insert(0, "", pictures_folder, "Pictures") if pictures_folder

                i = 0
                while i < directories.length
                    directories[i] += "c" if File.identical?(directories[i + 1], path)
                    i += 3
                end

                files1 = []
                subdirs1 = []
                begin
                    Dir.entries(path).each { |basename|
                        next if basename.start_with? "."
                        fullname = File.join(path, basename)
                        fullname.sub!("/", "\\") if Baroque::VRSketch::IS_WINDOWS
                        if File.file? fullname
                            next unless ((fullname.downcase.end_with? ".jpg") ||
                                            (fullname.downcase.end_with? ".jpeg") ||
                                            (fullname.downcase.end_with? ".png"))
                            files1.push("", fullname, basename)
                        elsif File.directory? fullname
                            subdirs1.push("d", fullname, basename)
                        end
                    }
                rescue => e
                    puts e   # and continue
                end

                #puts({path: path, directories: directories, subdirs1: subdirs1, files1: files1})

                emit({
                    cmd: "list_files",
                    history: directories,
                    names: subdirs1 + files1,
                })

                Baroque::VRSketch.last_used_directory = path
            end

            def cmd_get_file(cmd)
                raw = ""
                if (@on_quest && (Baroque::VRSketch.allowed_actions[12] != 1))
                    puts("get_file: 'Enable browsing image files from the Quest' is not set")
                else
                    fullname = cmd["filename"]
                    if !(fullname.include? "\0")
                        if ((fullname.downcase.end_with? ".jpg") ||
                            (fullname.downcase.end_with? ".jpeg") ||
                            (fullname.downcase.end_with? ".png"))
                            begin
                                f = open(fullname, "rb")
                                raw = f.read
                                f.close
                                puts("sending skybox: " + fullname)
                            rescue
                                raw = ""
                            end
                        end
                    end
                end
                emit_rawdata(raw, {
                    cmd: "get_file",
                    name: cmd["name"],
                })
            end

            def point_cloud_read(cmd)
                return if @point_cloud_base_path == nil
                return if @point_cloud_base_path != cmd["basepath"]
                relname = cmd["relname"]
                return if ((relname == "") ||
                           (relname.start_with? "/") ||
                           (relname.start_with? "\\") ||
                           (relname.include? "..") ||
                           (relname.include? ":"))

                if relname == "cloud.js"
                    # push all the tiny .hrc files in subdirectories
                    Dir.chdir(@point_cloud_base_path) do
                        Dir.glob("**/*.hrc").each { |hrcfile|
                            hrcfile = hrcfile.sub("\\", "/")
                            point_cloud_send(hrcfile)
                        }
                    end
                end
                point_cloud_send(relname)
            end

            def point_cloud_send(relname)
                filename = @point_cloud_base_path + "/" + relname
                return if !File.exist? filename
                f = open(filename, "rb")
                raw = f.read
                f.close
                emit_rawdata(raw, {
                    cmd: "point_cloud_data",
                    names: [@point_cloud_base_path, relname],
                })
            end

            def cmd_audio(cmd, rawdata)
                #puts(cmd)
                case cmd["subcmd"]

                when nil, ""    # after "play", get next page
                    dict = @record_play_attrdict
                    return if dict == nil
                    return if !dict.valid?
                    p = cmd["p"]
                    raw = dict["p" + p.to_s]
                    #puts({cmd: "recdata", id: p, rawlength: raw.pack("i*").size}) if raw.is_a? Array
                    emit_rawdata(raw.pack("i*"), {cmd: "recdata", id: p}) if raw.is_a? Array
                
                when "rec"    # before "edit_move_text {record=1}", receive one page of audio data
                    @record_pages = [] if @record_pages == nil
                    @record_pages.push(rawdata) if @record_pages != nil

                when "play"
                    @record_play_attrdict = nil
                    text = @model.find_entity_by_id(cmd["text_id"])
                    if text.is_a? Sketchup::Text
                        dict = text.attribute_dictionary("Baroque_VRSketch_audio")
                        v_list = [cmd["text_id"]]
                        if dict != nil
                            @record_play_attrdict = dict
                            p = 0
                            while dict["p" + p.to_s] != nil
                                p += 1
                            end
                            v_list.push(dict["date"], 1.0, p)
                        end
                        #puts({cmd: "recdata", v: v_list})
                        emit({cmd: "recdata", v: v_list})
                    else
                        puts("record+play: Text object not found")
                    end
                end
            end

            def cmd_audio_record(text, cmd)
                record = cmd["record"]
                if ((record != nil) && (record != 0) && (text.is_a? Sketchup::Text))
                    text.delete_attribute("Baroque_VRSketch_audio")   # if already there
                    if ((@record_pages != nil) && (@record_pages.size != 0) && (record > 0))
                        dict = text.attribute_dictionary("Baroque_VRSketch_audio", true)
                        page_count = 0
                        @record_pages.each { |raw|
                            dict["p" + page_count.to_s] = raw.unpack("i*")
                            page_count += 1
                        }
                        dict["date"] = Time.now.to_f
                    end
                end
                @record_pages = nil
            end

            def cmd_private_warehouse(cmd)
                if (@wh_dialog && @wh_dialog.visible?)
                    id = cmd['id']
                    return if !id.is_a? Integer
                    thumb_url = cmd['thumb_url']
                    if ((thumb_url.is_a? String) && (thumb_url != ""))
                        return if thumb_url.include? "'"
                        @wh_dialog.execute_script("download(#{id.to_s}, '#{thumb_url}')")
                        return
                    end
                    v1 = Sketchup.version_number
                    v2 = @vrsketch_version
                    v2 = 0 if v2 == nil
                    page = cmd['page']
                    return if !page.is_a? Integer
                    @wh_dialog.execute_script("search(#{v1.to_s}, #{v2.to_s}, #{id.to_s}, #{page.to_s})")
                    return
                end
                @wh_dialog = UI::HtmlDialog.new({
                    :dialog_title => "Warehouse",
                    :preferences_key => "za.co.VRSketch4_wh",
                    :scrollable => false,
                    :resizable => true,
                    :width => 300,
                    :height => 120,
                    :left => 500,
                    :top => 80,
                    :style => UI::HtmlDialog::STYLE_UTILITY,
                    })
                mydir = File.dirname(__FILE__)
                @wh_dialog.set_file(mydir + "/html/private_warehouse.html")
                @wh_dialog.add_action_callback("onload") { |ctx|
                    Baroque::VRSketch.error_monitor.monitor {
                        cmd_private_warehouse(cmd) if (@wh_dialog && @wh_dialog.visible?)
                    }
                }
                @wh_dialog.add_action_callback("reply") { |ctx, reply|
                    Baroque::VRSketch.error_monitor.monitor {
                        emit({cmd: "private_warehouse_reply", name: reply})
                    }
                }
                @wh_dialog.show
            end

            def check_vrsketch_version(min_version)
                if ((@vrsketch_version == nil) || (@vrsketch_version < min_version))
                    if on_quest
                        UI.messagebox("This command cannot be completed because VR Sketch on your Oculus Quest is too old.  Please upgrade it and try again.")
                    else
                        UI.messagebox("This command cannot be completed because of a version mismatch.  You need to upgrade the 'VR Sketch' executable (the VR renderer part) and restart it.")
                    end
                    return false
                end
                return true
            end

            def get_clipboard_file
                if !@clipboard_file
                    @clipboard_file = File.join(Sketchup.temp_dir, "Skp_VRSketch_clipboard.skp")
                    if File.exist? @clipboard_file
                        begin
                            File.delete(@clipboard_file)
                        rescue
                        end
                    end
                end
                return @clipboard_file
            end

            def clipboard_cut(ctx, picked)
                filename = get_clipboard_file

                picked.reject! { |x| x.deleted? }
                if picked.length > 0
                    grp = ctx.add_group(picked)
                    grp.definition.save_as(filename)
                    @clipboard_transformation = grp.transformation
                    grp.erase!
                end
            end

            def clipboard_paste_in_place(ctx)
                return if !@clipboard_transformation

                cdef = @model.definitions.load(get_clipboard_file)
                cinst = ctx.add_instance(cdef, @clipboard_transformation)
                cinst.explode
                erase_component_definition(cdef)
            end

            def erase_component_definition(cdef)
                # definitions.remove() was added only in SU2018.  There are various
                # workarounds found in the sketchUcation forum.  This is from
                # https://sketchucation.com/forums/viewtopic.php?p=145255
                path = @model.active_path
                if path then
                    path.reverse!.map! { |c| c.definition }
                    if path.include?(cdef) then
                        @model.selection.clear
                        (path.index(cdef)+1).times { @model.close_active }
                    end
                end
                begin
                    @model.definitions.remove(cdef)
                rescue
                    cdef.entities.erase_entities(cdef.entities.to_a)
                    # we rely on this cdef disappearing "naturally" after it has no entities left
                end
            end
        end


        # ----- Observers ------

        def change_sketchup_model
            @model_tracker.close_socket if @model_tracker != nil
            @model_tracker = nil
        end

		def leave_tutorial
			@close_model_without_asking = nil
		end

        class MyAppObserver < Sketchup::AppObserver
            def onNewModel(model)
                Baroque::VRSketch.error_monitor.monitor {
                    if Baroque::VRSketch::IS_WINDOWS
                        Baroque::VRSketch.leave_tutorial
                        Baroque::VRSketch.change_sketchup_model
                    end
                }
            end
            def onOpenModel(model)
                Baroque::VRSketch.error_monitor.monitor {
                    if Baroque::VRSketch::IS_WINDOWS
                        Baroque::VRSketch.leave_tutorial
                        Baroque::VRSketch.change_sketchup_model
                    end
                }
            end
            def onQuit
                Baroque::VRSketch.error_monitor.monitor {
                    model_tracker = Baroque::VRSketch.model_tracker
                    model_tracker.close_session_stats if model_tracker != nil
                }
            end
        end

        class MyModelObserver < Sketchup::ModelObserver
            def initialize(tracker)
                @tracker = tracker
            end
            def onActivePathChanged(model)
                #puts("===== active_path_changed =====")
                Baroque::VRSketch.error_monitor.monitor {
                    @tracker._enable_updates
                }
            end
            def onTransactionCommit(model)
                #puts("===== transaction_commit =====")
                Baroque::VRSketch.error_monitor.monitor {
                    @tracker.transaction_commit
                }
            end
            def onTransactionUndo(model)
                #puts("===== transaction_undo =====")
                Baroque::VRSketch.error_monitor.monitor {
                    @tracker.transaction_undo
                }
            end
            def onTransactionRedo(model)
                #puts("===== transaction_redo =====")
                Baroque::VRSketch.error_monitor.monitor {
                    @tracker.transaction_redo
                }
            end
            def onPostSaveModel(model)
                Baroque::VRSketch.error_monitor.monitor {
                    @tracker.model_filename_changed
                }
            end
        end

        class MyEntitiesObserver < Sketchup::EntitiesObserver
            # This tracks the 'entities' pseudo-list, either of the model or of a
            # ComponentDefinition.  It should respond to all changes both of the list
            # itself and of individual elements.  Of course there are various situation
            # where SketchUp fails to call this observer, so we also need the other
            # observer subclasses below to increase coverage.
            def initialize(tracker)
                @tracker = tracker
            end
            def onElementAdded(entities, entity)
                #puts({:ElementAdded => entity})
                Baroque::VRSketch.error_monitor.monitor {
                    @tracker.push_update(entity)
                }
            end
            def onElementModified(entities, entity)
                #puts({:ElementModified => entity})
                Baroque::VRSketch.error_monitor.monitor {
					# pom pom pom workaround: if we change the position of a texture, we're getting
					# an entity that is really of type Sketchup::Entity, whose parent is an
					# AttributeDictionaries, whose own parent is the face.  Meh.
					# Also, a second case: we get 'entity' being an AttributeDictionary when we
					# modify a group's "dynamic_attributes" dictionary.
					if ((entity.class == Sketchup::Entity) ||    # and not a subclass!
					    (entity.class == Sketchup::AttributeDictionary))
						if entity.parent.class == Sketchup::AttributeDictionaries
							entity = entity.parent.parent    # should get us the face or group...
						end
					end
                    return if entity.is_a? Sketchup::Model
                    return if entity.is_a? Sketchup::ComponentDefinition
                    @tracker.push_update(entity)
                }
            end
            def onElementRemoved(entities, entity_id)
                #puts({:ElementRemoved => entity_id})
                Baroque::VRSketch.error_monitor.monitor {
                    fail if entity_id <= 0
                    # aaa "entities.parent" may fail because that entities list is deleted,
                    # but there is no "entities.deleted?".
                    begin
                        cdef = entities.parent
                    rescue
                        return
                    end
                    @tracker.push_removal_entity(entity_id, cdef)
                }
            end
        end

        class MyDefinitionsObserver < Sketchup::DefinitionsObserver
            # This handles a dict that maps ComponentDefinitions to their entityID in the model.
            # It should be updated to contain all ComponentDefinitions that are actually used.
            # The main purpose of this class is to have onComponentRemoved() to track removal
	        # of a ComponentDefinition.  In theory it could also be handled by onEraseEntities on
            # EntitiesObserver, but it seems that onComponentRemoved is generally more reliable.
            def initialize(tracker, model)
                @tracker = tracker
                @components = {model => 0}
            end
            def registered?(cdef)
                Baroque::VRSketch.error_monitor.monitor {
                    return @components.has_key?(cdef)
                }
            end
            def add_component_definition(cdef)
                Baroque::VRSketch.error_monitor.monitor {
                    @components[cdef] = cdef.entityID
                }
            end
            def onComponentRemoved(definitions, definition)
                Baroque::VRSketch.error_monitor.monitor {
                #puts({:ComponentRemoved => definition})
                    cdef_id = @components.delete(definition)
                    @tracker.push_removal_cdef_id(definition, cdef_id) if cdef_id != nil
                    @tracker.call_later(:send_components_list)
                }
            end
            def onComponentAdded(definitions, definition)
                Baroque::VRSketch.error_monitor.monitor {
                    @tracker.call_later(:send_components_list)
                }
            end
            def onComponentPropertiesChanged(definitions, definition)
                Baroque::VRSketch.error_monitor.monitor {
                    @tracker.call_later(:send_components_list)
                }
            end
            def onComponentTypeChanged(definitions, definition)
                Baroque::VRSketch.error_monitor.monitor {
                    @tracker.call_later(:send_components_list)
                }
            end
            def _dump()
                puts({MyDefinitionObserver: @components})
            end
        end

        class MyVertexObserver < Sketchup::EntityObserver
            # For Sketchup::Vertex used in the model as start or end of a detached Edge
            # (i.e. an edge not part of a face).
            def initialize(tracker)
                @tracker = tracker
                @observed_vertices = Set[]
            end
            def track(vertex)
                Baroque::VRSketch.error_monitor.monitor {
                    if @observed_vertices.add?(vertex)
                        vertex.add_observer(self)
                    end
                }
            end
            def onChangeEntity(vertex)
                Baroque::VRSketch.error_monitor.monitor {
                    return if vertex.deleted?
                    #puts({:VertexChange => vertex})
                    vertex.edges.each { |edge| @tracker.push_update(edge) }
                    #vertex.faces.each { |face| @tracker.push_update(face) }
                }
            end
            def onEraseEntity(vertex)
                Baroque::VRSketch.error_monitor.monitor {
                    @observed_vertices.delete(vertex)
                }
            end
            def _dump()
                puts({MyVertexObserver: @observed_vertices})
            end
        end

        class MyGroupObserver < Sketchup::EntityObserver
            # For groups and component instances used in the model.  We use both an EntityObserver
            # and the EntitiesObserver that is present in the 'entities' list, as an attempt to
            # reduce the risk that one of them is not fired.
            def initialize(tracker)
                @tracker = tracker
                @group2layer = {}
            end
            def track(grp)
                Baroque::VRSketch.error_monitor.monitor {
                    old_layer = @group2layer[grp]     # old_layer.deleted? might be true
                    if old_layer != nil
                        grp.add_observer(self) 
                    end
                    @group2layer[grp] = grp.layer
                }
            end
            def layer_changed(layer)
                Baroque::VRSketch.error_monitor.monitor {
                    #puts("layer_changed: #{layer}, name = #{layer.name}")
                    #groups = @layer2groups[layer]
                    #puts("groups = #{groups}")
                    #return if groups == nil
                    #groups.each { |grp| @tracker.push_update(grp) }
                    @group2layer.each_pair { |grp, laygrp|
                        next if laygrp.deleted?
                        if layer == laygrp
                            @tracker.push_update(grp)
                        end
                    }
                }
            end
            def onChangeEntity(grp)
                Baroque::VRSketch.error_monitor.monitor {
                    @tracker.push_update(grp)
                }
            end
            def onEraseEntity(grp)
                Baroque::VRSketch.error_monitor.monitor {
                    old_layer = @group2layer.delete(grp)
                    #@layer2groups[old_layer].delete(grp) if old_layer != nil
                }
            end
            def _dump()
                puts({MyGroupObserver: @group2layer})
            end
        end


        class MyMaterialsObserver < Sketchup::MaterialsObserver
            def initialize(tracker)
                @tracker = tracker
            end
            def onMaterialAdd(materials, material)
                Baroque::VRSketch.error_monitor.monitor {
                    #puts "onMaterialAdd " + material.name
                    @tracker.add_material(material)
                }
            end
            def onMaterialChange(materials, material)
                Baroque::VRSketch.error_monitor.monitor {
                    #puts "onMaterialChange " + material.name
                    @tracker.add_material(material)
                }
            end
            def onMaterialRemove(materials, material)
                Baroque::VRSketch.error_monitor.monitor {
                    #puts "onMaterialRemove " + material.name
                    # -- we add the name to the list of materials to update-or-remove
                    @tracker.add_material(material)
                }
            end
            def onMaterialUndoRedo(materials, material)
                Baroque::VRSketch.error_monitor.monitor {
                    #puts "onMaterialUndoRedo " + material.name
                    @tracker.add_material(material)
                }
            end
        end

        
        class MyShadowInfoObserver < Sketchup::ShadowInfoObserver
            def initialize(tracker)
                @tracker = tracker
            end
            def onShadowInfoChanged(shadow_info, type)
                Baroque::VRSketch.error_monitor.monitor {
                    datetime_only = (type == 0)
                    @tracker.shadow_info_changed(datetime_only)
                }
            end
        end


        class MyLayersObserver < Sketchup::LayersObserver
            # Note about layers: VRSketch doesn't follow exactly SketchUp but instead it follows the
            # recommendations of https://help.sketchup.com/en/sketchup/controlling-visibility-layers.
            # This means that any geometry you draw is always on Layer0, but you can set individual
            # groups and component instances to other layers and show/hide these other layers.  The
            # layer property is ignored on other entities than groups and component instances.
            def initialize(tracker, group_observer)
                @tracker = tracker
                @group_observer = group_observer
            end
            def onLayerChanged(layers, layer)
                Baroque::VRSketch.error_monitor.monitor {
                    @group_observer.layer_changed(layer)
                    @tracker.call_later(:send_layers_list)
                }
            end
            #def onCurrentLayerChanged(layers, layer)
            #    Baroque::VRSketch.error_monitor.monitor {
            #        @tracker.call_later(:send_layers_list)
            #    }
            #end
            def onLayerAdded(layers, layer)
                Baroque::VRSketch.error_monitor.monitor {
                    @tracker.call_later(:send_layers_list)
                }
            end
            def onLayerRemoved(layers, layer)
                Baroque::VRSketch.error_monitor.monitor {
                    @tracker.call_later(:send_layers_list)
                }
            end
        end


        class MyUnitsOptionsObserver < Sketchup::OptionsProviderObserver
            def initialize(tracker)
                @tracker = tracker
            end
            def onOptionsProviderChanged(provider, name)
                Baroque::VRSketch.error_monitor.monitor {
                    @tracker.units_options_changed
                }
            end
        end


        class MyRenderingOptionsObserver < Sketchup::RenderingOptionsObserver
            def initialize(tracker)
                @tracker = tracker
            end
            def onRenderingOptionsChanged(rendering_options, type)
                Baroque::VRSketch.error_monitor.monitor {
                    #puts "onRenderingOptionsChanged(#{rendering_options}, #{type})"
                    case type

                    when Sketchup::RenderingOptions::ROPSetEdgeDisplayMode
                        @tracker.allowed_actions_changed(7)

                    when Sketchup::RenderingOptions::ROPEditComponent   # "hide rest of model"
                        @tracker.allowed_actions_changed(6)
                    
                    when Sketchup::RenderingOptions::ROPSetDisplayFog, \
                         Sketchup::RenderingOptions::ROPSetFogColor, \
                         Sketchup::RenderingOptions::ROPSetFogDist, \
                         Sketchup::RenderingOptions::ROPSetFogUseBkColor, \
                         Sketchup::RenderingOptions::ROPSetBackgroundColor
                        @tracker.shadow_info_changed(false)
                    
                    end
                }
            end
        end


        class MyFrameChangeObserver
            def frameChange(from_scene, to_scene, percent_done)
                # called either several times with percent_done growing from 0.0 to 1.0 (if scene
                # transitions are enabled), or only once with percent_done equal to 1.0 (if not).
                # We want to call start_frame_change() the first time only.
                Baroque::VRSketch.error_monitor.monitor {
                    #puts(sel: Sketchup.active_model.pages.selected_page, prev_scene: @prev_scene, to_scene: to_scene)
                    return if !Sketchup.active_model
                    prev_scene = @prev_scene
                    prev_scene = Sketchup.active_model.pages.selected_page if prev_scene == nil
                    skip = ((@prev_scene == to_scene) && (percent_done > @prev_percentage))
                    @prev_scene = to_scene
                    @prev_percentage = percent_done
                    return if skip

                    # OBSCURE: if we call view.invalidate regularly, then the timer behind Sketchup's
                    # own animation feature resets to 0 every time.  But we get about 10 updates
                    # every second of the beret guy.  That basically locks the animation to never
                    # progress.  I didn't find a better workaround than deselecting the teleport tool,
                    # which hides the beret guy.
                    Baroque::VRSketch.deselect_teleport_tool

                    return if !Baroque::VRSketch.connected("Server")
                    Baroque::VRSketch.model_tracker.start_frame_change(Sketchup.active_model, prev_scene, to_scene)
                }
            end
        end


        class MyPagesObserver < Sketchup::PagesObserver
            def initialize(tracker)
                @tracker = tracker
            end
            def onContentsModified(pages)
                Baroque::VRSketch.error_monitor.monitor {
                    @tracker.call_later(:send_scenes_list)
                }
            end
            def onElementAdded(pages, page)
                Baroque::VRSketch.error_monitor.monitor {
                    @tracker.call_later(:send_scenes_list)
                }
            end
            def onElementRemoved(pages, page)
                Baroque::VRSketch.error_monitor.monitor {
                    @tracker.call_later(:send_scenes_list)
                }
            end
        end


        # ----- CONNECTION & UI ------

        DEFAULT_PORT_PC = 17352

        def try_open_connection()
            if @waiting_sock != nil
                sock = @waiting_sock
                @waiting_sock = nil
                return sock
            end
            begin
                addr = "127.0.0.1"
                port = DEFAULT_PORT_PC
                sock = TCPSocket.new(addr, port)
            rescue Errno::ECONNREFUSED
                return nil
            rescue
                UI.messagebox "#{$!.class}: #{$!}"
                raise
            end
            return sock
        end

        if Baroque::VRSketch::IS_WINDOWS
            def launch_vrsketch_process()
                return nil if !verify_checksums?(true)
                cmd = result = procinfo = nil   # force these variables to be declared here, not only inside the "do" block
                Dir.chdir(File.dirname(__FILE__)) do
                    cmd = 'VRSketch' + VERSION + ".exe"
                    # BOOL CreateProcessA(LPCSTR, PVOID, PVOID, PVOID, BOOL, DWORD, PVOID, PVOID, PVOID, PVOID)
                    startinfo = ([0] * 18).pack('Q' * 18)
                    procinfo = Kernel32::PROCESS_INFORMATION.malloc
                    procinfo.hProcess = 0
                    procinfo.hThread = 0
                    procinfo.dwProcessId = 0
                    procinfo.dwThreadId = 0
                    result = Kernel32.CreateProcessA(cmd, nil, nil, nil, 0, 0, nil, nil, startinfo, procinfo)
                end
                if result == 0
                    # should not really get there any more, now that we have verify_checksums?()
                    UI.messagebox("This is VRSketch " + VERSION + " not finding the executable #{cmd}. " +
                                  "If you just upgraded to the next version, you need to restart " +
                                  "SketchUp now. Note also that some antivirus are known to silently " +
                                  "remove this file when installing; try to disable your antivirus " +
                                  "and reinstall.")
                    return nil
                end
                #puts(CreateProcessA: [procinfo.hProcess, procinfo.hThread])
                result = Kernel32.CloseHandle(procinfo.hThread)
                #puts(CloseHandle: result)
                return procinfo.hProcess, cmd
            end

            def process_still_alive(hprocess)
                exitcode = [0].pack('L')
                result = Kernel32.GetExitCodeProcess(hprocess, exitcode)
                #puts(GetExitCodeProcess: [result, exitcode.unpack('L').first])
                return true if ((result != 0) && (exitcode.unpack('L').first == 259))   # STILL_ACTIVE
                result = Kernel32.CloseHandle(hprocess)
                #puts(CloseHandle: result)
                return false
            end
        else
            # OS X
            def command_only_works_on_pc()
                UI.messagebox("Directly-connected VR headsets are not supported on OS X. " +
                                "This command can only be used on Windows PCs.")
            end
        end

        def async_open_connection(on_quest: false, send_model: true, &block)
            write_statistics
            if !on_quest
                sock = try_open_connection()
                @launched_process = nil
                if sock == nil
                    @launched_process = launch_vrsketch_process()
                    return if @launched_process == nil
                end
                _open_connection_finalize(sock, on_quest, send_model, block)
            else
                propose_quest_connection(send_model, block)
            end
        end

        def _open_connection_finalize(sock, on_quest, send_model, block)
            if read_registry("bypasskey", "").hash != -798388627
                begin
                    Baroque::VRSketch.error_monitor.monitor {
                        if (send_model && (!check_license(true)))
                            show_license_dialog
                            return
                        end
                        submit_telemetry_data
                    }
                rescue
                end
            end
            read_statistics

            if @launched_process != nil
                hproc = @launched_process[0]
                cmd = @launched_process[1]
                @launched_process = nil

                while sock == nil
                    sleep(0.3)
                    if !process_still_alive(hproc)
                        UI.messagebox("Failed to start or connect to the Unity subprocess: #{cmd}")
                        return
                    end
                    sock = try_open_connection()
                end
                result = Kernel32.CloseHandle(hproc)
                #puts(CloseHandle: result)
            end
            block.call(sock) if sock != nil
        end

        def connected(mode="Any")
            return false if @model_tracker == nil
            return @model_tracker.connected(mode)
        end

        def connected_on(on_quest)
            return (connected && (@model_tracker.on_quest == on_quest))
        end

        BASE_URL = "http://api.vrsketch.eu"

        def random_address
            x = Random.new.bytes(10)
            return x.each_byte.map { |b| b.to_s(16) }.join
        end

        if Baroque::VRSketch::IS_WINDOWS
            REGISTRY_PATHNAME = 'SOFTWARE\VRSketch'

            def read_registry(key, default=nil)
                begin
                    Win32::Registry::HKEY_CURRENT_USER.open(REGISTRY_PATHNAME) { |reg|
                        return reg.read_s(key)
                    }
                rescue
                    return default
                end
            end

            def write_registry(key, value)
                begin
                    Win32::Registry::HKEY_CURRENT_USER.create(REGISTRY_PATHNAME) { |reg|
                        reg.write_s(key, value.to_s)
                    }
                rescue
                    puts("write_registry failed to write " + key + " = " + value.to_s)
                end
            end
        else
            # OSX
            def osx_load_config
                if @osx_config == nil
                    appdata = File::expand_path('../../..', Sketchup::find_support_file('Plugins'))
                    @osx_config_path = File::join(appdata, "VR Sketch config")
                    begin
                        f = File.open(@osx_config_path, "r")
                        @osx_config = JSON.parse(f.read)
                        f.close
                    rescue
                        @osx_config = {}
                    end
                end
                return @osx_config
            end
            
            def read_registry(key, default=nil)
                result = osx_load_config[key]
                result = default if result == nil
                return result
            end

            def write_registry(key, value)
                cfg = osx_load_config
                cfg[key] = value
                begin
                    f = File.open(@osx_config_path, "w")
                    f.write(JSON.generate(cfg))
                    f.close
                rescue
                    puts("write_registry failed to write into file " + @osx_config_path)
                end
            end
        end

        def set_cloud_server(server)
            @cloud_server = server
            if @cloud_server.end_with? "/"
                @cloud_server = @cloud_server[0..-2] # chop last character
            end
        end

        def read_config
            cfg = read_registry("cfg_allowed", "")
            i = 0
            while i < @allowed_actions.size && i < cfg.size
                @allowed_actions[i] = 0
                @allowed_actions[i] = 1 if cfg[i] == "1"
                i += 1
            end
            snapscale = read_registry("cfg_snapscale3", "").to_i
            set_cloud_server(read_registry("cfg_cloudserver", "https://vrsketch.eu"))
            @snap_scale = snapscale if snapscale > 0
        end

        def write_config
            write_registry("cfg_allowed", @allowed_actions.map { |c| c.to_s }.join)
            write_registry("cfg_snapscale", @snap_scale.to_s)
            write_registry("cfg_snapscale3", @snap_scale.to_s)
            write_registry("cfg_cloudserver", @cloud_server.to_s)
        end

        def read_license_file
            # returns "" if it cannot be read
            return read_registry("license_key", "")
        end

        def write_to_license_file(contents)
            write_registry('license_key', contents)
        end

        def read_statistics
            return if @statistics != nil
            begin
                @statistics = JSON.parse(read_registry("statistics", "."))
            rescue
                clear_statistics
            end
        end

        def write_statistics
            return if @statistics == nil
            @statistics = Hash[ @statistics.sort_by { |key, value| key } ]
            write_registry("statistics", JSON.generate(@statistics))
        end

        def clear_statistics
            @statistics = {"date initialized" => Time.now.to_i}
            write_statistics
        end

        def stat_event(event_name, time=nil)
            return if @statistics == nil
            #
            key = event_name
            key += " N" if time != nil
            current = @statistics[key]
            current = 0 if current == nil
            @statistics[key] = current + 1
            #
            if time != nil
                time = time.to_i if time == time.to_i
                total = @statistics[event_name + " total"]
                total = 0 if total == nil
                @statistics[event_name + " total"] = total + time
                #
                maximum = @statistics[event_name + " max"]
                if ((maximum == nil) || (time > maximum))
                    @statistics[event_name + " max"] = time
                end
            end
        end

        def submit_telemetry_data
            last_used = read_registry('last_used_timestamp')
            if last_used == nil
                write_registry('last_used_timestamp', Time.now.to_i.to_s)
                write_registry('last_used_count', "1")
                return
            end
            last_used = last_used.to_i
            if Time.now.to_i - last_used > 3600 * 24
                r = nil
                use_count = read_registry('last_used_count', "0").to_i
                use_count = use_count + 1
                last_notify = read_registry('last_used_notify')   # last time we contacted the server
                if last_notify == nil
                    last_notify = last_used.to_s    # if unspecified, use last_used and write it immediately
                    write_registry('last_used_notify', last_notify)  # so that this path is only followed once
                end
                last_notify = last_notify.to_i
                begin
                    if Time.now.to_i - last_notify > 3600 * 24 * 3  # contact every three days
                        uri = URI(BASE_URL + '/users/refresh')
                        d = read_license_file
                        if !d
                            key = ""
                        elsif d[0] == 'T'
                            key = d[1...d.length]
                        else
                            key = d[1...65]
                        end
                        read_statistics
                        @statistics["date submitted"] = Time.now.to_i
                        stats = JSON.generate(@statistics)
                        body = Net::HTTP.post_form(uri, {key: key, count: (use_count - 1).to_s,
                                                         ver: RB_VERSION_ID.to_s, stats: stats}).body
                        # update 'last_used_notify' when the http request succeeded; if it fails, we keep the
                        # old value and will try again to contact the server the next time, i.e. after 24 hours
                        write_registry('last_used_notify', Time.now.to_i.to_s)
                        use_count = 1
                        r = JSON.parse(body)
                        clear_statistics      # only after we check that the answer was valid JSON
                        if key.length > 0 && r['ends_at']
                            write_to_license_file("P" + key + r['ends_at'].to_s)
                        end
                    end
                ensure
                    write_registry('last_used_count', use_count.to_s)
                    write_registry('last_used_timestamp', Time.now.to_i.to_s)
                end
                
                if ((r != nil) && r["note"])
                    warn(r["note"], "Note", 400)
                end
                if ((r != nil) && r["tips"])
                    write_registry("tips", r["tips"])
                end
            end
        end

        def check_license(contact_server)
            mydir = File.dirname(__FILE__)
            data = read_license_file
            if data == "" || data == nil
                uri = URI(BASE_URL + '/users/new')
                begin
                    r = Baroque::VRSketch.error_monitor.monitor {
                        JSON.parse(Net::HTTP.post_form(uri, {mac: random_address, ver: RB_VERSION_ID}).body)
                    }
                rescue
                    write_to_license_file("T" + (Time.now.to_i + 3600 * 24 * 14).to_s)
                    return true
                end
                exp_tstamp = r['exp_tstamp'].to_i
                @license_valid_timestamp = exp_tstamp
                @current_license_description = "temporary license"
                @current_license_key = "no valid key"
                if exp_tstamp > Time.now.to_i
                    write_to_license_file("T" + exp_tstamp.to_s)
                    return true
                else
                    expired_license("Couldn't get initial temporary license.") if contact_server
                    return false
                end
            else
                if data[0] == "T"
                    @license_valid_timestamp = data[1...data.length].to_i
                    @current_license_description = "temporary license"
                    @current_license_key = "no valid key"
                    if @license_valid_timestamp > Time.now.to_i
                        return true
                    else
                        expired_license("Your temporary license expired.  You can get a license from https://vrsketch.eu/.  This applies to all categories: student, hobbyist or professional.") if contact_server
                        return false
                    end
                elsif data[0] == "P"
                    @current_license_key = data[1...65]
                    @current_license_description = "subscription license"
                    @license_valid_timestamp = data[65...data.length].to_i
                    if Time.now.to_i <= @license_valid_timestamp.to_i + 3600 * 120
                        return true    # accept an old key for 120 more hours (5 days)
                    else
                        if contact_server && try_to_renew_license(@current_license_key)
                            return true
                        end
                        expired_license("Your license is no longer valid.  Please check https://vrsketch.eu/.") if contact_server
                        return false
                    end
                end
                return false
            end
        end

        def expired_license(msg)
            UI.messagebox("No valid license found.\n\n" + msg)
        end

        def fill_template(content, h)
            h.each do |k, v|
                # may contain the template several times
                content = content.gsub("{{ " + k.to_s + " }}", v)
            end
            return content
        end

        def show_license_dialog()
            @license_dialog.close if @license_dialog != nil
            @license_dialog = UI::HtmlDialog.new({
                :dialog_title => "VRSketch " + VERSION + " License",
                :preferences_key => "za.co.VRSketch2_Li",
                :scrollable => true,
                :resizable => true,
                :width => 700,
                :height => 400,
                :left => 400,
                :top => 250,
                :style => UI::HtmlDialog::STYLE_DIALOG,
                })
            mydir = File.dirname(__FILE__)
            @license_dialog.set_file(mydir + "/html/license.html")

            @license_dialog.add_action_callback("onload") { |ctx|
                Baroque::VRSketch.error_monitor.monitor {
                    license_type = @current_license_description
                    license_key = @current_license_key
                    date_valid = @license_valid_timestamp
                    if date_valid == nil
                        date_valid = "n/a"
                    else
                        begin
                            date_valid = DateTime.strptime(date_valid.to_s, '%s').strftime('%e %B %Y')
                        rescue
                            date_valid = "(unknown)"
                        end
                    end
                    @license_dialog.execute_script(
                        "set_current_license('#{license_type}', '#{license_key}', '#{date_valid.to_s}')")
                }
            }
            @license_dialog.add_action_callback("save_key") { |_ctx, key|
                Baroque::VRSketch.error_monitor.monitor {
                    uri = URI(BASE_URL + '/users/get_license')
                    begin
                        r = JSON.parse(Net::HTTP.post_form(uri, {key: key}).body)
                    rescue
                        UI.messagebox("Cannot contact license server", MB_OK)
                    else
                        if !r['success']
                            UI.messagebox("Invalid license key")
                        else
                            write_to_license_file("P" + key + r['ends_at'].to_s)
                            check_license(false) # populates the fields
                            show_license_dialog
                            UI.messagebox("successfully updated license")
                        end
                    end
                }
            }
            @license_dialog.show
        end

        def try_to_renew_license(key)
            uri = URI(BASE_URL + '/users/get_license')
            begin
                r = JSON.parse(Net::HTTP.post_form(uri, {key: key}).body)
            rescue
                return false
            else
                if !r['success']
                    return false
                end
                write_to_license_file("P" + key + r['ends_at'].to_s)
                if r['ends_at'].to_i > Time.now.to_i
                    return true
                end
            end
            return false
        end

        def verify_checksums_error(all)
            return 'wrong version of baroquesoftware_vrsketch\\main.rb(e)' if MAIN_VERSION != VERSION
            mydir = File.dirname(__FILE__) + "/"
            DIGEST_FILES.each { |inp|
                if inp == nil
                    next if all
                    break
                end
                filename, expected_checksum = inp
                return ("missing file: " + filename) if !File.exist?(mydir + filename)
                f = File.open(mydir + filename, "rb")
                checksum = Digest::MD5.hexdigest(f.read)
                f.close
                return ("corrupted or wrong version: " + filename) if checksum != expected_checksum
            }
            return nil
        end

        def verify_checksums?(all)
            return true # disable it for now as we don't know why it's failing
            return true if @checksums_verified
            error = verify_checksums_error(all)
            if error
                # keep the following three lines in sync with the same three lines in baroquesoftware_vrsketch.rb
                dlg = UI::WebDialog.new("VR Sketch installation", true, nil, 750, 400, 200, 200)
                dlg.set_html(HTML_INSTALLATION_BROKEN.sub('<!---tag>', 'reason: ' + error))
                dlg.show
                return false
            end
            @checksums_verified = all
            return true
        end

        def create_toolbar
            if true    # for indentation
                mydir = File.dirname(__FILE__)

                @toolbar = UI::Toolbar.new "VR Sketch"
                [false, true].each { |on_quest|
                    where = "on this PC"
                    where = "on Oculus Quest" if on_quest
                    cmd = UI::Command.new("VR Sketch " + name) {
                        Baroque::VRSketch.error_monitor.monitor {
                            return warn_model_needed if !Sketchup.active_model
                            Sketchup.active_model.select_tool(@teleport_tool)
                            if !connected_on(on_quest)
                                send_to_unity_vr(propose_tutorial: true, on_quest: on_quest)
                            end
                        }
                    }
                    iconname = mydir + "/BeretGuy"
                    iconname = mydir + "/BeretGuyQ" if on_quest
                    cmd.small_icon = iconname + "16.png"
                    cmd.large_icon = iconname + "24.png"
                    cmd.tooltip = "Send to VR " + where + " and teleport tool"
                    cmd.status_bar_text = "Send the model to VR " + where + ".  View where the person in VR is, and teleport by clicking on model."
                    cmd.menu_text = "VR"
                    cmd.set_validation_proc {
                        if ((@teleport_tool != nil) && @teleport_tool.now_active && connected_on(on_quest))
                            result = MF_CHECKED
                        else
                            result = MF_UNCHECKED
                        end
                        result
                    }
                    @toolbar.add_item cmd
                }

                cmd2 = UI::Command.new("VR Sketch copy loc") {
                    Baroque::VRSketch.error_monitor.monitor {
                        center_view
                    }
                }
                cmd2.small_icon = mydir + "/CopyVRloc16.png"
                cmd2.large_icon = mydir + "/CopyVRloc24.png"
                cmd2.tooltip = "Copy VR location"
                cmd2.status_bar_text = "Copy viewpoint from VR to SketchUp"
                cmd2.set_validation_proc {
                    if !Baroque::VRSketch.connected_and_can_center_view
                        result = MF_GRAYED
                    else
                        result = MF_UNCHECKED
                    end
                    result
                }
                @toolbar = @toolbar.add_item cmd2

                cmd3 = UI::Command.new("VR Sketch config") {
                    Baroque::VRSketch.error_monitor.monitor {
                        show_config
                    }
                }
                cmd3.small_icon = mydir + "/Config16.png"
                cmd3.large_icon = mydir + "/Config24.png"
                cmd3.tooltip = "Configure VR view"
                cmd3.status_bar_text = "Open the configuration dialog box for VR Sketch"
                @toolbar.add_item cmd3
            end
        end

        class OggPage
            BITSTREAM_SERIAL_NUM = -1 - (Random.rand * 0x7fffffff).to_i
            SAMPLES_PER_BLOCK = 2880

            def initialize(header_type, page_seq_num)
                @header_type = header_type
                @page_seq_num = page_seq_num
                @granule_position = 0
            end

            def write_single_block(f, block_data)
                num_segments = block_data.size / 255 + 1
                segments_header = num_segments.chr
                segments_header += "\xff" * (num_segments - 1)
                segments_header += (block_data.size - 255 * (num_segments - 1)).chr
                segments_header.force_encoding("ASCII-8BIT")
                write_segmented(f, segments_header + block_data)
            end

            def write_ogg_data(f, int_list, granule_position)
                block_data = int_list.pack("i*")
                num_segments = block_data[0].ord
                total_size = 1 + num_segments
                num_segments.times { |i|
                    segment_size = block_data[i + 1].ord
                    total_size += segment_size
                    granule_position += SAMPLES_PER_BLOCK if segment_size < 255
                }
                if total_size > block_data.size
                    throw "invalid Opus data"
                end
                @granule_position = granule_position
                write_segmented(f, block_data[0...total_size])
                return granule_position
            end

            def write_segmented(f, segmented_page_data)
                header = "OggS\x00"    # signature, version=0
                header += @header_type.chr
                header += [@granule_position].pack("<q")
                header += [BITSTREAM_SERIAL_NUM, @page_seq_num, 0].pack("<iii")    # checksum=0
                header.force_encoding("ASCII-8BIT")
                f.write(header + segmented_page_data)
            end
        end

        def write_ogg_file(dict, ogg_filename)
            f = open(ogg_filename, 'wb')
            # version=1, channels=1, pre-skip=0, frame-rate=48000, output-gain=0, mapping_family=0
            header = "OpusHead\x01\x01\x00\x00\x80\xbb\x00\x00\x00\x00\x00"
            header.force_encoding("ASCII-8BIT")
            OggPage.new(2, 0).write_single_block(f, header)
            vendor = "VRSketch"
            vendor.force_encoding("ASCII-8BIT")
            OggPage.new(0, 1).write_single_block(f, "OpusTags" + [vendor.size].pack("<i") + vendor)
            page_seq_num = 2
            gpos = 0
            i = 0
            while dict["p" + i.to_s]
                gpos = OggPage.new(0, page_seq_num).write_ogg_data(f, dict["p" + i.to_s], gpos)
                page_seq_num += 1
                i += 1
            end
            f.close
        end

        def play_audio_recording(obj)
            if (@ogg_dialog && @ogg_dialog.visible?)
                @ogg_dialog.close
                @ogg_dialog = nil
                delay(0.27) {
                    play_audio_recording(obj)
                }
                return
            end

            Dir.glob(Sketchup.temp_dir + "/Skp_VRSketch_*.ogg").each { |filename|
                begin
                    File.delete(filename)
                rescue
                end
            }

            return if obj.deleted?
            dict = obj.attribute_dictionary("Baroque_VRSketch_audio")
            return if dict == nil

            ogg_filename = temporary_filename(".ogg")
            write_ogg_file(dict, ogg_filename)

            @ogg_dialog = UI::HtmlDialog.new({
                :dialog_title => "Play",
                :preferences_key => "za.co.VRSketch3_ogg",
                :scrollable => false,
                :resizable => true,
                :width => 220,
                :height => 170,
                :left => 700,
                :top => 200,
                :style => UI::HtmlDialog::STYLE_UTILITY,
                })
            mydir = File.dirname(__FILE__)
            @ogg_dialog.set_file(mydir + "/html/play_ogg.html")
            @ogg_dialog.add_action_callback("onload") { |ctx|
                Baroque::VRSketch.error_monitor.monitor {
                    enc_filename = Base64.strict_encode64(ogg_filename)
                    @ogg_dialog.execute_script("play_file('#{enc_filename}')")
                }
            }
            @ogg_dialog.set_on_closed {
                Baroque::VRSketch.error_monitor.monitor {
                    begin
                        File.delete(ogg_filename)
                    rescue
                    end
                }
            }
            @ogg_dialog.show
        end

        def create_menu_handler
            UI.add_context_menu_handler { |menu|
                sel = Sketchup.active_model.selection
                next if sel.size != 1
                obj = sel[0]
                next unless obj.is_a? Sketchup::Text
                
                item = menu.add_item("Play") {
                    play_audio_recording(obj)
                }
                menu.set_validation_proc(item) {
                    if obj.attribute_dictionary("Baroque_VRSketch_audio")
                        MF_ENABLED
                    else
                        MF_GRAYED
                    end
                }
            }
        end

        def deselect_teleport_tool
            # if the "teleport tool" is currently selected, deselect it
            return if !Sketchup.active_model
            if ((@teleport_tool != nil) && @teleport_tool.now_active)
                Sketchup.active_model.select_tool(nil)
            end
        end

        def show_upgrade_notice(cfg)
			dialog = UI::WebDialog.new("Welcome to VR Sketch " + VERSION, true,
                "za.co.VRSketch5d_prop_tut", 690, 440, 210, 150, true)

            dialog.set_html '
                <html><head><style>
                    body { font-family: Arial, Helvetica, sans-serif; background-color: "f8f8ff"; }
                    p { margin: 0.5em; font-size: 120%; }
                    p.vskip { margin: 0.5em; }
                    p.small { font-size: 100%; }
                    a.h { color: black; text-decoration:none; }
                    span.slightlysmall { font-size: 90%; }
                    form { border-style: outset; border-color: white; border-width: 1;
                            margin-left: 0.5em; margin-right: 0.5em; padding: 0.4em;
                            padding-left: 2.2em; padding-right: 2.1em; }
                    input.text { height: 2em; }
					
					.button {
						background-color: #4CAF50;
						border: none;
						color: white;
						padding: 15px 25px;
						text-align: center;
						font-size: 16px;
						cursor: pointer;
					}
					.button:hover {
						background-color: green;
					}
					.otherbutton {
						background-color: #A0C0A0;
						border: none;
						color: white;
						padding: 15px 25px;
						text-align: center;
						font-size: 16px;
						cursor: pointer;
					}
					.otherbutton:hover {
						background-color: #C0E0C0;
					}
                </style>
                </head>
                <body>
                &nbsp;
				<form>
                <p>&nbsp;</p>
				<p align="center">
                Welcome to VR Sketch ' + MAIN_VERSION + '!</p>
                <p>&nbsp;</p>
                <p><b>Where are my tools?</b></br>&nbsp;</p>
                <li>In the tools panel, non-editing tools are now over the Teleport icon.
                <br>&nbsp;
                <li>We are now using the ' + "'menu'" + ' button on the controller.  On Vive, it
                is the top small button; on Oculus, it is the B or Y button.
                <p>&nbsp;</p>
				<p align="center">
				<button class="button" onclick="window.location.href=\'skp:ok\'">OK</button>&nbsp;&nbsp;
                <button class="button" onclick="window.location.href=\'skp:ok@MORE\'">Read more...</button>
				</p>
				</form>
                </body></html>'

            dialog.add_action_callback("ok") { |dlg, param|
                Baroque::VRSketch.error_monitor.monitor {
					dialog.close
                    show_whatsnew if param == "MORE"
                    write_registry("tutorials", cfg + "5")
                }
            }
            dialog.show
			return false
		end

		def do_propose_tutorial(on_quest)
			return false if @tutorial_proposed
			cfg = read_registry("tutorials", "")
			return false if cfg.length >= 2
            return show_upgrade_notice(cfg) if cfg.length == 1

			dialog = UI::WebDialog.new("Welcome to VR Sketch " + VERSION, true,
                "za.co.VRSketch5d_prop_tut", 690, 440, 210, 150, true)

            dialog.set_html '
                <html><head><style>
                    body { font-family: Arial, Helvetica, sans-serif; background-color: "f8f8ff"; }
                    p { margin: 0.3em; font-size: 120%; }
                    p.vskip { margin: 0.5em; }
                    p.small { font-size: 100%; }
                    a.h { color: black; text-decoration:none; }
                    span.slightlysmall { font-size: 90%; }
                    form { border-style: outset; border-color: white; border-width: 1;
                            margin-left: 0.5em; margin-right: 0.5em; padding: 0.4em; }
                    input.text { height: 2em; }
					
					.button {
						background-color: #4CAF50;
						border: none;
						color: white;
						padding: 15px 25px;
						text-align: center;
						font-size: 16px;
						cursor: pointer;
					}
					.button:hover {
						background-color: green;
					}
					.otherbutton {
						background-color: #A0C0A0;
						border: none;
						color: white;
						padding: 15px 25px;
						text-align: center;
						font-size: 16px;
						cursor: pointer;
					}
					.otherbutton:hover {
						background-color: #C0E0C0;
					}
                </style>
                </head>
                <body>
                &nbsp;
				<form>
                <p>&nbsp;</p>
				<p align="center">
				Welcome to VR Sketch!<br>
				We have some very quick tutorials to get you started.
				</p>
				<p class="small" align="center">You can also access the tutorials from the menu Extensions > VR Sketch.</p>
                <p>&nbsp;</p>
				</form>
				&nbsp;
                <form>
				<p>&nbsp;</p>
				<p align="center">
				<button type="button" class="button" onclick="window.location.href=\'skp:start\'">Start tutorials</button>&nbsp;&nbsp;
				<button type="button" class="otherbutton" onclick="window.location.href=\'skp:later\'">Remind me later</button>&nbsp;&nbsp;
				<button type="button" class="otherbutton" onclick="window.location.href=\'skp:dont\'">Don\'t ask again</button>
				</p>
				<p>&nbsp;</p>
                </form>
                </body></html>'

            dialog.add_action_callback("start") { |dlg, param|
                Baroque::VRSketch.error_monitor.monitor {
					dialog.close
                    Baroque::VRSketch.delay(0.1) {
						start_tutorial(0, on_quest, nil)
						write_registry("tutorials", "s5") if @model_tracker != nil
                    }
                }
            }
            dialog.add_action_callback("later") { |dlg, param|
                Baroque::VRSketch.error_monitor.monitor {
					@tutorial_proposed = true
					dialog.close
                    Baroque::VRSketch.delay(0.1) {
						send_to_unity_vr(on_quest: on_quest)
					}
                }
            }
            dialog.add_action_callback("dont") { |dlg, param|
                Baroque::VRSketch.error_monitor.monitor {
					@tutorial_proposed = true
					dialog.close
                    Baroque::VRSketch.delay(0.1) {
						write_registry("tutorials", "d5")
						send_to_unity_vr(on_quest: on_quest)
					}
                }
            }
            dialog.show
			return true
		end

        def send_to_unity_vr(tutorial: false, propose_tutorial: false, send_model: true, on_quest: false,
                             dont_reopen_if_connected: false, &block)
            if (dont_reopen_if_connected && connected_on(on_quest))
                hack_SetForegroundWindow
                block.call if block
                return
            end

            if ((!Baroque::VRSketch::IS_WINDOWS) && (!on_quest))
                command_only_works_on_pc     # gives the messagebox()
                return
            end

			if propose_tutorial
				return if do_propose_tutorial(on_quest)
			end
            deselect_teleport_tool
            @teleport_tool = TeleportTool.new

            if !@initialized
                @initialized = true
                hack_prepare_open_group

                Sketchup::Pages.add_frame_change_observer(MyFrameChangeObserver.new)
            end
            @toolbar.show

            hwnd = nil
            hwnd = @model_tracker.hwnd if connected_on(on_quest)
            change_sketchup_model

            #puts("===== open_new_model =====")
            async_open_connection(on_quest: on_quest, send_model: send_model) { |sock|
                change_sketchup_model

                model = Sketchup.active_model
                return warn_model_needed if !model
                tracker = ModelTracker.new(model, sock, on_quest)
                @model_tracker = tracker
                tracker.hwnd = hwnd
                tracker.tutorial = tutorial
                hack_SetForegroundWindow
                model.active_view.invalidate

                tracker.emit({
                    cmd: "setconfig",
                    name: "cloud_server",
                    names: [@cloud_server]
                })

                if send_model
                    tracker.send_model_now
                    tracker.model_filename_changed
                    debug_autoreconnect
                else
                    # turn off "debug: auto-reconnect" when connecting to a cloud model, otherwise we
                    # will immediately afterwards re-send the local model
                    debug_autoreconnect(force_off: true)
                end
                block.call if block
            }
        end

        def connected_and_can_center_view()
            return false if !connected
            return false if @teleport_tool == nil
            return (@teleport_tool.guy_transform != nil)
        end

        def center_view()
            return if !connected_and_can_center_view
            origin = @teleport_tool.guy_transform.origin
            #origin = origin + @teleport_tool.guy_transform * Geom::Vector3d.new(0, 0, 1.7 * (1/0.0254))
            target = origin + @teleport_tool.guy_transform * Geom::Vector3d.new(0, -2.5 * (1/0.0254), 0.5 * (1/0.0254))
            #move_by = scale * 2
            #target = origin + Geom::Vector3d.new(move_by, move_by, move_by)
            @model_tracker.model.active_view.camera = [Sketchup::Camera.new(target, origin, [0, 0, 1]), 0.5]
        end

        def compute_teleport_destination(scene, frame_change_speed)
            pos, on_floor, dir = Baroque::VRSketch.compute_position_to_send_to_vr(scene.camera)
            #puts({teleport: [pos, dir, on_floor, frame_change_speed]})
            return pos.to_a + dir.to_a + [on_floor, frame_change_speed]
        end

        def compute_position_to_send_to_vr(camera)
            dir = camera.direction.normalize
            center = camera.eye
            # move 'center' back 20cm.  The goal is to avoid SketchUp displaying the Modulor man
            # on the complete screen from very close
            step_distance = dir
            step_distance.length = 0.20/0.0254
            center -= step_distance
            # if possible, adjust 'center' so that it ends up on the ground.  Adjust only by up
            # to 2.6 meters, and at least 1.1 meter.
            ray = [center - Geom::Vector3d.new(0, 0, 1.1 * (1/0.0254)), Geom::Vector3d.new(0, 0, -1)]
            target = nil
            target = Sketchup.active_model.raytest(ray) if Sketchup.active_model
            if ((target != nil) && (center.z - target[0].z < 2.6 * (1/0.0254)))
                return [target[0], 1, dir]
            else
                return [center, 0, dir]
            end
        end

        def update_radio()
            return if ((@config_dialog == nil) || (!@config_dialog.visible?))
            i = 0
            while i < @allowed_actions.length
                if ((i != 6) && (i != 7))
                    allow = @allowed_actions[i]
                    @config_dialog.execute_script(
                        "document.getElementById('act" + i.to_s + "').checked = " + allow.to_s)
                end
                i += 1
            end
            @config_dialog.execute_script(
                "document.getElementById('snap_scale').value = " + @snap_scale.to_s)
            @config_dialog.execute_script(
                "document.getElementById('cloud_server').value = '" + @cloud_server.to_s + "'")
        end

        def debug_autoreconnect(force_off: false)
            debug_reconnect = (read_registry("devmode", "") == "1")
            debug_reconnect = false if force_off
            if debug_reconnect
                if @debug_auto_reconnect == nil
                    @debug_auto_reconnect = UI.start_timer(2.7, true) {
                        Baroque::VRSketch.error_monitor.monitor {
                            if !connected
                                if @waiting_sock == nil
                                    Sketchup.set_status_text("DEVMODE")
                                    @waiting_sock = try_open_connection()
                                    if @waiting_sock != nil
                                        send_to_unity_vr
                                    else
                                        debug_autoreconnect   # re-check the registry
                                    end
                                end
                            end
                        }
                    }
                end
            elsif @debug_auto_reconnect != nil
                UI.stop_timer(@debug_auto_reconnect)
                @debug_auto_reconnect = nil
            end
        end

        def temporary_filename(extension)
            name1 = Process.pid.to_s
            return File.join(Sketchup.temp_dir, "Skp_VRSketch_#{name1}" + extension)
        end

        class Sock2File
            def initialize(filename)
                @z = Zlib::Deflate.new(9)
                @f = File.new(filename, "wb")
                @f.write("VRSketch 1.#{RB_VERSION_ID}\n\r\n\0")
            end
            def send(data, ignored)
                @z.deflate(data) { |chunk| @f.write(chunk) }
                return data.bytesize
            end
            def close
                if @z != nil
                    @f.write(@z.flush())
                    @total = @f.tell()
                    @f.close
                    @z = nil
                end
                return @total
            end
        end

        def open_sketchlib
            if !@vrsketch_sketchlib
                if !Baroque::VRSketch::IS_WINDOWS
                    warn("This operation could not be completed for now (it requires Windows-specific code)")
                    return nil
                end
                # indirect: use LoadLibraryExW() to give the full path as utf16, and then
                # use the regular Fiddle.dlopen() with only the module name, which has been
                # loaded already by LoadLibraryExW().  I think Fiddle.dlopen() uses internally
                # LoadLibraryExA()...
                path = File.dirname(__FILE__) + '/vrsketch-sketchlib' + VERSION + '.dll'
                strutf16 = (path + "\0").encode(Encoding::UTF_16LE)
                strptr = Fiddle::Pointer.malloc(strutf16.bytesize)
                strptr[0, strptr.size] = strutf16
                handle = Kernel32.LoadLibraryExW(strptr, 0, 0)
                puts(LoadLibraryExW: path, result: handle)
                if handle == 0
                    warn("Cannot load " + path + ".  Check your installation.")
                    return nil
                end
                lib = Fiddle.dlopen('vrsketch-sketchlib' + VERSION + '.dll')
                @vrsketch_sketchlib = {
                    #upload_via_https: Fiddle::Function.new(
                    #    lib['upload_via_https'], [Fiddle::TYPE_VOIDP,  # url
                    #                              Fiddle::TYPE_VOIDP,  # filename
                    #                              Fiddle::TYPE_VOIDP,  # basename
                    #                              Fiddle::TYPE_VOIDP,  # description
                    #                              Fiddle::TYPE_VOIDP,  # key
                    #                              Fiddle::TYPE_VOIDP   # result_filename
                    #    ], Fiddle::TYPE_VOID),
                    #upload_via_https_progress: Fiddle::Function.new(
                    #    lib['upload_via_https_progress'], [], Fiddle::TYPE_LONG),
                    #upload_via_https_cancel: Fiddle::Function.new(
                    #    lib['upload_via_https_cancel'], [], Fiddle::TYPE_VOID),
                    generate_filename: Fiddle::Function.new(
                        lib['generate_filename'], [Fiddle::TYPE_LONG,  # video
                                                   Fiddle::TYPE_VOIDP  # result_filename
                        ], Fiddle::TYPE_LONG),
                    load_skm_directory: Fiddle::Function.new(
                        lib['load_skm_directory'], [Fiddle::TYPE_VOIDP,  # name
                                                    Fiddle::TYPE_VOIDP,  # dirs
                                                    Fiddle::TYPE_VOIDP   # result_filename
                        ], Fiddle::TYPE_LONG),
                }
                #puts(@vrsketch_sketchlib)
            end
            return @vrsketch_sketchlib
        end

        def open_cloud_page
            license_key = read_license_file
            if (!license_key) || (license_key[0] != 'P')
                url = @cloud_server + CLOUD_PAGE
            else
                url = @cloud_server + CLOUD_PAGE + '#' + license_key
            end
            if !UI.openURL(url)
                UI.messagebox("Error parsing " + @cloud_server + " as URL")
            end
        end

        def view_cloud_model
            if !@cloud_server.start_with? "http"
                UI.messagebox("misconfigured cloud server - should start with http:// or https://")
                return
            end
            begin
                url = Clipboard.paste_latin1.strip
                url = "" if !url.start_with? (@cloud_server + "/m/")
            rescue
                url = ""
            end
            template = {paste_buffer: url}
            dialog = show_dialog('open_cloud_model', 'View cloud model', 750, 400, template: template)
            def reset_url(dlg, url)
                return if url.index("'") != nil
                Baroque::VRSketch.delay(0.05) {
                    dlg.execute_script("document.getElementById('url').value = '" + url + "'")
                }
            end
            dialog.add_action_callback("go") { |dlg, param|
              Baroque::VRSketch.error_monitor.monitor {
                url = dlg.get_element_value("url").strip
                if url.start_with? (@cloud_server + "/m/")
                    load_remote_model(url, param == "quest")
                else
                    UI.messagebox('Invalid model URL: should be "' + @cloud_server + '/m/..."', MB_OK)
                end
                reset_url(dlg, url)
              }
            }
            dialog.add_action_callback("pub") { |dlg, param|
              Baroque::VRSketch.error_monitor.monitor {
                if !PUBLIC_CLOUD_SERVER.start_with? @cloud_server
                    UI.messagebox("Public model only works if cloud server is set to default (https://vrsketch.eu), please change in settings")
                else
                    load_remote_model(PUBLIC_CLOUD_SERVER + param, false)
                    dialog.close
                end
              }
            }
        end

		def upload_vrsketch_file
            if @upload_vrsketch_dialog != nil
                @upload_vrsketch_dialog.close
                @upload_vrsketch_dialog = nil
            end

            license_key = read_license_file
            if (!license_key) || (license_key[0] != 'P')
                UI.messagebox("You need a license in order to get cloud storage", MB_OK)
                return false
            end

            basename = nil
            basename = Sketchup.active_model.title if Sketchup.active_model
            basename = "Untitled" if ((!basename) || (basename.length == 0))

			@upload_vrsketch_dialog = UI::HtmlDialog.new({
                :dialog_title => "Upload to cloud",
                :preferences_key => "za.co.VRSketch13_u2cld",
                :scrollable => true,
                :resizable => true,
                :width => 820,
                :height => 360,
                :left => 200,
                :top => 100,
                :style => UI::HtmlDialog::STYLE_DIALOG,
                })
            mydir = File.dirname(__FILE__)
            @upload_vrsketch_dialog.set_file(mydir + "/html/cloud_upload.html")
            
            @upload_vrsketch_dialog.add_action_callback("upload") { |ctx|
                Baroque::VRSketch.error_monitor.monitor {
                    start_upload_vrsketch_file(@upload_vrsketch_dialog)
                }
            }
            @upload_vrsketch_dialog.add_action_callback("close") { |ctx|
                Baroque::VRSketch.error_monitor.monitor {
                    @upload_vrsketch_dialog.close
                    @upload_vrsketch_dialog = nil
                }
            }
            @upload_vrsketch_dialog.add_action_callback("viewpage") { |ctx|
                Baroque::VRSketch.error_monitor.monitor {
                    open_cloud_page
                }
            }
            @upload_vrsketch_dialog.add_action_callback("save") { |ctx|
                Baroque::VRSketch.error_monitor.monitor {
                    return if !Sketchup.active_model
                    Sketchup.active_model.save if Sketchup.active_model.modified?
                }
            }
            @upload_vrsketch_dialog.add_action_callback("onload") { |ctx|
                Baroque::VRSketch.error_monitor.monitor {
                    value = Base64.strict_encode64(basename)
                    @upload_vrsketch_dialog.execute_script("set_filename('#{value}')")
                }
            }
            @upload_vrsketch_dialog.add_action_callback("delete_file") { |ctx|
                Baroque::VRSketch.error_monitor.monitor {
                    begin
                        File.delete(@_upload_vrsketch_delete_filename)
                    rescue
                    end
                }
            }

            warning_visible = ["none", "none"]
            @upload_vrsketch_dialog.add_action_callback("ontimer") { |ctx|
                Baroque::VRSketch.error_monitor.monitor {
                    should_be_visible = ["none", "none"]
                    if Sketchup.active_model && Sketchup.active_model.modified?
                        if Sketchup.active_model.path == ""
                            should_be_visible[1] = "block"
                        else
                            should_be_visible[0] = "block"
                        end
                    end
                    if should_be_visible != warning_visible
                        dlg = @upload_vrsketch_dialog
                        dlg.execute_script("document.getElementById('save_option').style.display = '#{should_be_visible[0]}'")
                        dlg.execute_script("document.getElementById('save_suggestion').style.display = '#{should_be_visible[1]}'")
                        warning_visible = should_be_visible
                    end
                }
            }

            @upload_vrsketch_dialog.show
        end

        def warn_model_needed
            UI.messagebox("Please open a model first.")
            return nil
        end

        def start_upload_vrsketch_file(dlg)
            # close currently opened groups and components before uploading a cloud model
            model = Sketchup.active_model
            return warn_model_needed if !model
            model.selection.clear
            while model.active_entities != model.entities
                break if !model.close_active
            end

            Baroque::VRSketch.delay(0.1) {
                do_upload_vrsketch_file(dlg, model)
            }
        end

        def do_upload_vrsketch_file(dlg, model)
            return warn_model_needed if !model.valid?
            filename = temporary_filename(".vrsketch")
            @_upload_vrsketch_delete_filename = filename

            sock = Sock2File.new(filename)
            tracker = ModelTracker.new(model, sock, false)
            tracker.upload_whole_model(for_export: true)
            tracker._forget_socket
            total_size = sock.close()

            cloud_server = @cloud_server
            license_key = read_license_file()
            dlg.execute_script("start_upload('#{cloud_server}', '#{license_key}', '#{filename}')")
		end


		def propose_quest_connection(send_model, block)
            if @quest_conn_dialog != nil
                @quest_conn_dialog.close
                @quest_conn_dialog = nil
            end

			@quest_conn_dialog = UI::HtmlDialog.new({
                :dialog_title => "VR Sketch Remote",
                #:preferences_key => "za.co.VRSketch9_quest_conn",
                :scrollable => true,
                :width => 700,     # value overridden below
                :height => 440,    # value overridden below
                :left => 300,
                :top => 90,
                :style => UI::HtmlDialog::STYLE_UTILITY,
                })
            mydir = File.dirname(__FILE__)
            @quest_conn_dialog.set_file(mydir + "/html/quest_connection.html")
            
            quest_conn_dialog = @quest_conn_dialog
            quest_conn_queue = EnqueuedExecuteScripts.new(quest_conn_dialog)
            websock = nil
            broadcast_udp = nil

            quest_conn_dialog.add_action_callback("onload") { |ctx|
                Baroque::VRSketch.error_monitor.monitor {
                    value = read_registry("quest_uid_history", "")
                    quest_conn_dialog.execute_script("set_uid('#{value}')")
                    #
                    # start the timer that pings the UDP broadcast address
                    broadcast_udp.close if broadcast_udp != nil
                    broadcast_udp = nil
                    broadcast_udp = BroadcastUDP.new(quest_conn_queue)
                }
            }
            quest_conn_dialog.add_action_callback("help") { |ctx|
                Baroque::VRSketch.error_monitor.monitor {
                    UI.openURL("https://vrsketch.eu/docs-getting-started-quest.html")
                }
            }
            quest_conn_dialog.add_action_callback("alert") { |ctx, msg|
                Baroque::VRSketch.error_monitor.monitor {
                    Baroque::VRSketch.delay(0) {
                        UI.messagebox(msg)
                    }
                }
            }
            quest_conn_dialog.add_action_callback("go") { |ctx, uid|
                Baroque::VRSketch.error_monitor.monitor {
                    quest_conn_dialog.set_size(700, 118)
                    websock = broadcast_udp.open_link(uid)   # still nil for cloud connections
                    broadcast_udp.close
                    broadcast_udp = nil
                    if websock != nil    # in this case, it's actually a local wifi connection
                        old_model_tracker = @model_tracker
                        _open_connection_finalize(websock, true, send_model, block)
                        if @model_tracker != old_model_tracker
                            quest_conn_dialog.execute_script("set_status_connected('Connected')")
                            @model_tracker.notify_closing_queue = quest_conn_queue
                        end
                    end
                    write_registry("quest_uid_history", uid)
                }
            }
            quest_conn_dialog.add_action_callback("retry") { |ctx|
                Baroque::VRSketch.error_monitor.monitor {
                    Baroque::VRSketch.delay(0.1) {
                        propose_quest_connection(send_model, block)
                    }
                }
            }
            quest_conn_dialog.add_action_callback("ready") { |ctx|
                Baroque::VRSketch.error_monitor.monitor {
                    Baroque::VRSketch.delay(0.1) {
                        websock = WebSock.new(quest_conn_queue)
                        _open_connection_finalize(websock, true, send_model, block)
                    }
                }
            }
            quest_conn_dialog.add_action_callback("recv") { |ctx, base64data|
                Baroque::VRSketch.error_monitor.monitor {
                    websock._push(base64data)
                }
            }
            quest_conn_dialog.add_action_callback("close") { |ctx|
                Baroque::VRSketch.error_monitor.monitor {
                    broadcast_udp.close if broadcast_udp != nil
                    broadcast_udp = nil
                    websock._close_message_from_js if websock != nil
                }
            }
            quest_conn_dialog.add_action_callback("ontimer") { |ctx|
                Baroque::VRSketch.error_monitor.monitor {
                    quest_conn_queue.run_scripts
                    quest_conn_queue.immediate = true
                    begin
                        if @model_tracker
                            if @model_tracker.model.valid?
                                @model_tracker.run_timer
                            else
                                change_sketchup_model
                            end
                        end
                        broadcast_udp.run_timer if broadcast_udp
                    ensure
                        quest_conn_queue.immediate = false
                    end
                }
            }
            quest_conn_dialog.add_action_callback("update_fileaccess_check") { |ctx|
                Baroque::VRSketch.error_monitor.monitor {
                    ok = @allowed_actions[12]
                    quest_conn_dialog.execute_script("set_fileaccess('#{ok}')")
                }
            }
            quest_conn_dialog.add_action_callback("change_fileaccess") { |ctx, state|
                Baroque::VRSketch.error_monitor.monitor {
                    @allowed_actions[12] = state.to_i
                    @model_tracker.cmd_list_files({}) if connected
                    write_config
                }
            }
            quest_conn_dialog.set_can_close {
                Baroque::VRSketch.error_monitor.monitor {
                    broadcast_udp.close if broadcast_udp != nil
                    broadcast_udp = nil
                    websock._close_message_from_js if websock != nil
                }
                true
            }

            quest_conn_dialog.set_size(700, 440)
            quest_conn_dialog.set_position(300, 90)
            quest_conn_dialog.show
        end

        class WebSock
            def initialize(quest_conn_queue)
                @quest_conn_queue = quest_conn_queue
                @queue = nil
            end
            def send(data, ignored)
                # OK so I suspect that various timeouts are reached if we
                # send too large packets.  That's nonsense, but at least it
                # doesn't cause any issue if we send it in pieces.
                start = 0
                #puts({send: data.bytesize})
                while start < data.bytesize
                    stop = start + 32*1024
                    stop = data.bytesize if data.bytesize < stop
                    value = Base64.strict_encode64(data[start...stop])
                    @quest_conn_queue.enqueue("netconn.send('#{value}')")
                    #puts({piece_length: data[start...stop].length, stop: stop})
                    start = stop
                end
                return data.bytesize
            end
            def _push(base64data)
                value = Base64.strict_decode64(base64data)
                return if value.bytesize == 0
                if @queue == nil
                    @queue = []
                    @queue_start = 0
                end
                @queue.push(value)
            end
            def recv_nonblock(ignored)
                return "" if @js_closed
                return nil if @queue == nil
                result = @queue[@queue_start]
                @queue[@queue_start] = nil
                @queue_start += 1
                @queue = nil if @queue_start >= @queue.length
                return result
            end
            def recv(ignored)
                while @queue == nil     #XXX CHECK IF THAT WORKS
                    return "" if @js_closed
                    sleep(0.1)
                end
                return recv_nonblock(ignored)
            end
            def close
                @quest_conn_queue.enqueue("netconn.close()") if !@js_closed
                @js_closed = true
            end
            def _close_message_from_js
                @js_closed = true
                websock = self
                Baroque::VRSketch.delay(0.1) {
                    tracker = Baroque::VRSketch.model_tracker
                    tracker.close_socket if ((tracker != nil) && (tracker.sock == websock))
                }
            end
        end

        class BroadcastUDP
            def initialize(quest_conn_queue)
                begin
                    @udpsock = UDPSocket.new
                    @udpsock.setsockopt(Socket::SOL_SOCKET, Socket::SO_BROADCAST, true)
                    @udpsock.bind("0.0.0.0", 0)    # all interfaces, pick a random port
                rescue
                    puts("cannot create UDP socket for local broadcasting")
                    return
                end

                @request = "\x7b\xfc\xb2\xeb\x1e\x6e\xbb\x53"
                @answer = "\x6a\x68\x6c\xcc\x5e\xbd\xaa\xd6"
                @request.force_encoding("ASCII-8BIT")
                @answer.force_encoding("ASCII-8BIT")

                @local_wifi = {}
                @local_time = 0
                @quest_conn_queue = quest_conn_queue

                @ticks = 0
                run_timer
            end

            def run_timer
                # sends a ping and listens for answers
                @ticks -= 1
                if @ticks < 0
                    @ticks = 8
                    begin
                        @udpsock.send(@request, 0, "255.255.255.255", DEFAULT_PORT_PC)
                        puts("sent")
                    rescue
                        puts("failed to send UDP broadcast request")
                    end
                elsif @ticks == 3
                    @local_time += 1
                    begin
                        while true
                            result = @udpsock.recvfrom_nonblock(32)
                            puts(result)
                            msg = result[0]
                            addr = result[1]
                            if ((msg.start_with? @answer) && (msg.bytesize >= 14))
                                uidb = msg[8...14]
                                puts({UID: uidb, addr: addr})
                                if uidb.bytes.all? { |x| ((x >= 48) && (x <= 57)) }
                                    if @local_wifi[uidb] == nil
                                        @quest_conn_queue.enqueue("udp_pong_activate('#{uidb}')")
                                    end
                                    @local_wifi[uidb] = [@local_time, addr]
                                end
                            end
                        end
                    rescue IO::WaitReadable
                    end
                    @local_wifi.each_pair { |uidb, peer|
                        if ((peer != nil) && ((@local_time - peer[0]) >= 3))
                            @local_wifi[uidb] = nil
                            @quest_conn_queue.enqueue("udp_pong_deactivate('#{uidb}')")
                        end
                    }
                end
            end

            def close()
                UI.stop_timer(@timer) if @timer != nil
                begin
                    @udpsock.close
                rescue
                end
            end

            def open_link(uid)
                peer = @local_wifi[uid]
                if peer == nil
                    wss_url = 'wss://vrsketch.eu/cloudsession/c1?key=' + uid;
                    @quest_conn_queue.enqueue("connect_with_ws('#{wss_url}')")
                    return nil
                end
                begin
                    addr = peer[1][3]
                    port = peer[1][1]
                    sock = TCPSocket.new(addr, port)
                rescue
                    puts("#{$!.class}: #{$!}")
                    @quest_conn_queue.enqueue("set_status_error('Could not establish local connection.')")
                    return nil
                end
                sock.define_singleton_method(:_close_message_from_js) {
                    begin
                        sock.shutdown :RDWR
                    rescue     # probably already closed
                    end
                    Baroque::VRSketch.delay(0.16) {
                        tracker = Baroque::VRSketch.model_tracker
                        tracker.close_socket if ((tracker != nil) && (tracker.sock == sock))
                    }
                }
                return sock
            end
        end


        class EnqueuedExecuteScripts
            # on Mac, if we call dlg.execute_script() when shutdown is in
            # progress, we crash SketchUp.  The JavaScript side is apparently
            # shut down early when we quit---before onQuit is even called.
            # This should not occur if dlg.execute_script() is called in
            # response to some user action from Ruby; similarly, I guess that
            # the javascript engine is still valid when it invokes Ruby
            # callbacks that themselves use dlg.execute_script().  The only
            # case would then be if dlg.execute_script() is called from
            # UI.start_timer { ... }.  To fix this, we only enqueue scripts,
            # and we have another timer running on the JavaScript side which
            # fetches them.
            attr_accessor :immediate
            def initialize(dlg)
                @dlg = dlg
                @immediate = false
            end
            def enqueue(script)
                if @immediate
                    @dlg.execute_script(script) if @dlg.visible?
                else
                    @queue = [] if !@queue
                    @queue.push(script)
                end
            end
            def any_script
                return @queue
            end
            def run_scripts
                return if !@queue
                queue = @queue
                @queue = nil
                return if !@dlg.visible?
                queue.each { |script| @dlg.execute_script(script) }
            end
        end


        def load_remote_model(url, on_quest)
            send_to_unity_vr(send_model: false, on_quest: on_quest, dont_reopen_if_connected: true) {
                @model_tracker.emit({cmd: "load_remote_model", name: url})
                if @model_tracker.notify_closing_queue
                    @model_tracker.notify_closing_queue.enqueue("loaded_remote_model()")
                end
            }
        end

        def show_config()
            if ((@config_dialog == nil) || (!@config_dialog.visible?))
                @config_dialog = UI::WebDialog.new("VR Sketch " + VERSION + " config", true,
                    "za.co.VRSketch5d_config", 560, 660, 300, 150, true)
                @config_dialog.set_html '
                    <html><head><style>
                        body { font-family: Arial, Helvetica, sans-serif; background-color: "f8f8ff"; }
                        p { margin: 0.3em; }
                        p.vskip { margin: 0.5em; }
                        p.small { font-size: 80%; }
                        a.h { color: black; text-decoration:none; }
                        span.slightlysmall { font-size: 90%; }
                        span.tiny { font-size: 55%; }
                        form { border-style: outset; border-color: white; border-width: 1;
                               margin-left: 0.5em; margin-right: 0.5em; padding: 0.4em; }
                        input.text { height: 2em; }
                        h2 { color: red; }
                    </style>
                    </head>
                    <body>
                    <form>
                        <p>Enabled actions in VR:</p>
                        <a class="h" href="skp:act@0">
                            <input type="checkbox" id="act0" onclick="window.location.href=\'skp:act@0\'">
                            Teleporting (trigger)</a><br>
                        <a class="h" href="skp:act@1">
                            <input type="checkbox" id="act1" onclick="window.location.href=\'skp:act@1\'">
                            Move/scale/rotate model ("grip" buttons on the side)</a><br>
                        <a class="h" href="skp:act@2">
                            <input type="checkbox" id="act2" onclick="window.location.href=\'skp:act@2\'">
                            Editing (change tools with the touchpad)</a><br>
                        <a class="h" href="skp:act@11">
                            <input type="checkbox" id="act11" onclick="window.location.href=\'skp:act@11\'">
                            Saving</a><br>
                    </form>
                    <form>
                        <p>Rendering:</p>
                        <a class="h" href="skp:act@8">
                            <input type="checkbox" id="act8" onclick="window.location.href=\'skp:act@8\'">
                            Shadows follow SketchUp sun direction</a><br>
                        <table border=0><tr>
                            <td valign="top" align="right"><p class="small">sent automatically:
                                                               </p></td>
                            <td><p class="small">View &gt; Edge Style &gt; Edges<br>
                                                 View &gt; Component Edit &gt; Hide Rest Of Model<br>
                                                 View &gt; Fog and Fog settings</p></td>
                        </tr></table>
                        <p class="small">These settings and more are also available from VR (cog wheel, left of teleport tool).</p>
                    </form>
                    <form>
                        &nbsp;Alignment:&nbsp;&nbsp;&nbsp;
                        <input type="range" id="snap_scale" min="1" max="100" value="16" onchange="window.location.href=\'skp:snapscale@\'+this.value">
                        <p class="small">This measure is in real-life mm.  It changes how close to a point, axis or plane you must be
                        before snapping occurs, as well as the automatic rounding of distances.  Default value: 22 mm.
                        </p>
                    </form>
                    <h2>Advanced <span class="tiny"><a href="skp:cloudserver@HELP">What is this?</a></span></h2>
                    <form>
                        <table border=0><tr>
                        <td><button type="button" onclick="window.location.href=\'skp:cloudserver@PLAY\'">play VR video</button></td>
                        <td><button type="button" onclick="window.location.href=\'skp:cloudserver@REC\'">record VR video</button></td>
                        </tr></table><br/>
                        Cloud server: <input oninput="window.location.href=\'skp:cloudserver@\' + this.value" id="cloud_server" type="text" size=40></input> <button type="button" onclick="window.location.href=\'skp:cloudserver@RESET\'">reset to default</button>
                    </form>
                    </body></html>'

                @config_dialog.add_action_callback("act") { |dlg, param|
                    Baroque::VRSketch.error_monitor.monitor {
                        i = param.to_i
                        @allowed_actions[i] = 1 - @allowed_actions[i]
                        update_radio
                        if connected
                            @model_tracker.allowed_actions_changed(i)
                        end
                        write_config
                    }
                }
                @config_dialog.add_action_callback("cloudserver") { |dlg, param|
                    Baroque::VRSketch.error_monitor.monitor {
                        if param == "HELP"
                            UI.messagebox(
                                '"VR Videos" are 3D recordings.  They can only be played back in 3D in VR. ' +
                                "This is experimental, and unrelated to the photo/video tool in VR, " +
                                "which captures normal pictures and videos.\n\n" +
                                'The cloud server is the machine(s) where you store and retrieve models ' +
                                'with the "Upload file to cloud" and "View cloud model" commands.  These are ' +
                                'collaborative viewing tools.  Your models are never sent over the ' +
                                'Internet if you do not use them.  If you want to, but do not want ' +
                                'your models to travel over the Internet for security or performance ' +
                                'reasons, please contact us: we can provide a local version of the ' +
                                "cloud server to our business customers.\n\ninfo@baroquesoftware.com")
                            break
                        end
                        if param == "PLAY"
                            play_vr_video
                            break
                        end
                        if param == "REC"
                            record_vr_video
                            break
                        end
                        if param == "RESET"
                            @cloud_server = "https://vrsketch.eu"
                            update_radio()
                        else
                            set_cloud_server(param)
                        end
                        if connected
                            @model_tracker.emit({
                                cmd: "setconfig",
                                name: "cloud_server",
                                names: [@cloud_server]
                            })
                        end
                        write_config
                    }
                }
                @config_dialog.add_action_callback("snapscale") { |dlg, param|
                    Baroque::VRSketch.error_monitor.monitor {
                        @snap_scale = param.to_i
                        if connected
                            @model_tracker.emit({
                                cmd: "setconfig",
                                name: "snap_scale",
                                v: [@snap_scale * 0.001]
                            })
                        end
                        write_config
                    }
                }
                @config_dialog.show
                Baroque::VRSketch.delay(0.0) { update_radio }
                return true
            else
                @config_dialog.close
                @config_dialog = nil
                return false
            end
        end

        def record_vr_video()
            # SketchUp's UI.savepanel() is quite unusable.  Let's ask Unity to do it
            send_to_unity_vr(on_quest: false, dont_reopen_if_connected: true) {
                @model_tracker.emit({
                    cmd: "record_vr_video",
                })
            }
        end

        def play_vr_video()
            # SketchUp's UI.openpanel() is quite unusable.  Let's ask Unity to do it
            send_to_unity_vr(on_quest: false, dont_reopen_if_connected: true) {
                @model_tracker.emit({
                    cmd: "play_vr_video",
                })
            }
        end

        def check_point_cloud_connected
            if !connected
                UI.messagebox("First send your model to VR.  Use an empty model if you want to see the point cloud only.")
                return false
            end
            return @model_tracker.check_vrsketch_version(100200)
        end

        def view_point_cloud()
            if @point_cloud_dialog != nil
                @point_cloud_dialog.close
                @point_cloud_dialog = nil
            end
            return if !check_point_cloud_connected
            @point_cloud_dialog = UI::WebDialog.new("View point cloud", true,
                "za.co.VRSketch4_pointcloud", 700, 500, 200, 150, true)
            @point_cloud_dialog.set_html '
                <html><head><style>
                    body { font-family: Arial, Helvetica, sans-serif; background-color: "f8f8ff"; }
                    p { margin: 0.3em; }
                    p.vskip { margin: 0.5em; }
                    p.small { font-size: 80%; }
                    a.h { color: black; text-decoration:none; }
                    span.slightlysmall { font-size: 90%; }
                    span.tiny { font-size: 55%; }
                    form { border-style: outset; border-color: white; border-width: 1;
                            margin-left: 0.5em; margin-right: 0.5em; padding: 0.4em; }
                    input.text { height: 2em; }
                    h2 { color: red; }

					.button {
						background-color: #4CAF50;
						border: none;
						color: white;
						padding: 15px 25px;
						text-align: center;
						font-size: 16px;
						cursor: pointer;
					}
					.button:hover {
						background-color: green;
					}
                </style>
                </head>
                <body>
                <br>
                <p>You must first convert your point cloud to the POTree format
                (see <a href="skp:help">instructions</a>).
                Then pick the file "cloud.js" in the directory containing your POTree point cloud:</p>
                <br>

                <form>
                <br>
                <p><input type="file" id="cloudjs" size=80 accept=".js"></p>
                <br>
                <p>Number of points to display at once:
                <br>
                <input type="text" id="budget" size="8" value="2.0"> millions (try values like 2.0 on PC or 0.5 on Quest)
                <br>
                <input type="text" id="pixelsize" size="8" value="3.0"> pixels: radius of every point, or 0 to display as real points (forced 0 on Quest)

                <br><br>
                <p align="center">
                <button type="button" class="button" onclick="window.location.href=\'skp:start\'">Display in VR</button>
                &nbsp;&nbsp;
                <button type="button" class="button" onclick="window.location.href=\'skp:stop\'">Remove point cloud from VR</button>
                &nbsp;&nbsp;
                <button type="button" class="button" onclick="window.location.href=\'skp:cancel\'">Cancel</button>
                </p>
                <br>
                </form>

                </body></html>'

            @point_cloud_dialog.add_action_callback("help") { |dlg, param| Baroque::VRSketch.error_monitor.monitor {
                UI.openURL "https://vrsketch.eu/docs-pointcloud.html"
            } }
            @point_cloud_dialog.add_action_callback("start") { |dlg, param| Baroque::VRSketch.error_monitor.monitor {
                next if !check_point_cloud_connected
                filename = @point_cloud_dialog.get_element_value("cloudjs")
                basename = File.basename(filename)
                if basename.downcase != "cloud.js"
                    UI.messagebox("You must pick the file 'cloud.js' inside a POTree directory.")
                    next
                end
                budget = @point_cloud_dialog.get_element_value("budget").to_f
                if !(budget > 0.0)
                    UI.messagebox("Bad value for number of points.")
                    next
                end
                budget *= 1000000.0
                pixelsize = @point_cloud_dialog.get_element_value("pixelsize").to_f
                point_cloud_base_path = File.dirname(filename)
                @model_tracker.point_cloud_base_path = point_cloud_base_path
                @model_tracker.emit({
                    cmd: "point_cloud",
                    name: point_cloud_base_path,
                    v: [budget, pixelsize],
                })
                @point_cloud_dialog.close
                @point_cloud_dialog = nil
            } }
            @point_cloud_dialog.add_action_callback("stop") { |dlg, param| Baroque::VRSketch.error_monitor.monitor {
                if connected
                    @model_tracker.point_cloud_base_path = nil
                    @model_tracker.emit({cmd: "point_cloud"})
                end
                @point_cloud_dialog.close
                @point_cloud_dialog = nil
            } }
            @point_cloud_dialog.add_action_callback("cancel") { |dlg, param| Baroque::VRSketch.error_monitor.monitor {
                @point_cloud_dialog.close
                @point_cloud_dialog = nil
            } }
            @point_cloud_dialog.show
        end

        def show_documentation()
            UI.openURL "https://vrsketch.eu/documentation.html"
        end

        def show_whatsnew()
            UI.openURL "https://vrsketch.eu/docs-whatsnew.html"
        end

        def show_about()
            @about_dialog.close if @about_dialog != nil
            @about_dialog = UI::WebDialog.new("VRSketch " + VERSION + " About", true,
                    "za.co.VRSketchAb", 380, 300, 400, 250, true)
            mydir = File.dirname(__FILE__)
            f = open(mydir + "/about.html")
            @about_dialog.set_html(f.read)
            f.close
            @about_dialog.show
        end

		def load_model_template(template_file_name)
 			length_format = nil
 			length_format = Sketchup.active_model.options["UnitsOptions"]["LengthFormat"] if Sketchup.active_model
			if Sketchup.active_model && Sketchup.active_model.modified?
				if Sketchup.active_model == @close_model_without_asking
					Sketchup.active_model.close(true)    # ignore changes
				else
					Sketchup.active_model.close(false)
					return false if Sketchup.active_model && Sketchup.active_model.modified?
				end
			end
			old_template = Sketchup.template
			begin
				Sketchup.template = template_file_name
				Sketchup.file_new
   				Sketchup.active_model.options["UnitsOptions"]["LengthFormat"] = length_format if length_format != nil
			ensure
				Sketchup.template = old_template
			end
			@close_model_without_asking = Sketchup.active_model
			return true
		end

		def start_tutorial(index, on_quest, running_tracker)
			tutdir = File.dirname(__FILE__) + '/Tutorial'
			line = File.readlines(tutdir + '/Tutorial.txt')[index]
			tutname, title = line.split("|")
			name = tutname
			if !(File.exist?(tutdir + '/' + name + '.skp'))
				name = "Empty"
			end
			return if !load_model_template(tutdir + '/' + name + '.skp')

            vrvideo_filename = tutdir + '/' + tutname + '.vrvideo'
            if running_tracker == nil
                send_to_unity_vr(tutorial: true, on_quest: on_quest) {
                    _start_tutorial_now(index, vrvideo_filename)
                }
            else
                # a subset of send_to_unity_vr() to reuse the existing sock/sock_buffer
                hwnd = running_tracker[0]
                sock = running_tracker[1]
                sock_buffer = running_tracker[2]
                change_sketchup_model
                model = Sketchup.active_model     # should not be nil, we just loaded a model template
                tracker = ModelTracker.new(model, [sock, sock_buffer], on_quest)
                @model_tracker = tracker
                tracker.hwnd = hwnd
                tracker.tutorial = true
                tracker.send_model_now
                _start_tutorial_now(index, vrvideo_filename)
            end
        end

        def _start_tutorial_now(index, vrvideo_filename)
			if @model_tracker != nil
				@model_tracker.emit({
					cmd: "start_tutorial",
					id: index,
				})
                # for Quest, it is necessary to send the whole .vrvideo file via the wire too
                # (previous versions of the code would only send the directory name)
                f = open(vrvideo_filename, "rb")
                raw = f.read
                f.close
                @model_tracker.emit_rawdata(raw, {
                    cmd: "start_tutorial_video",
                })
			end
		end

        def _vlist2ptarray(v_list)
            ptarray = []
            v_list.each_slice(3) { |x, y, z|
                ptarray.push([x, y, z])
            }
            return ptarray
        end

        
        class MiniModel
            
            def initialize(basename)
                @bbox = Geom::BoundingBox.new
                @submeshes = []
                return if basename == nil

                mydir = File.dirname(__FILE__)
                f = open(mydir + "/minimodel/" + basename + ".json")
                value = f.read
                f.close
                minimodel = JSON.parse(value)

                minimodel.each { |minimodel1|
                    scale = minimodel1["scale"]
                    scale = [1, 1, 1] if scale == nil
                    recenter = minimodel1["recenter"]
                    recenter = [0, 0, 0] if recenter == nil
                    v = minimodel1["vertices"]
                    allpoints = []
                    i = 0
                    while i < v.length
                        allpoints.push(Geom::Point3d.new((v[i  ] * scale[0] + recenter[0]).m,
                                                         (v[i+2] * scale[2] + recenter[2]).m,
                                                         (v[i+1] * scale[1] + recenter[1]).m))
                        i += 3
                    end
                    minimodel1["submeshes"].each { |submesh|
                        col = submesh["color"]
                        v = []
                        submesh["triangles"].each { |t|
                            v.push(allpoints[t])
                        }
                        @submeshes.push({
                            color: Sketchup::Color.new(col[0], col[1], col[2], 255),
                            v: v,
                        })
                    }
                    @bbox.add(allpoints)
                }
            end

            def add_to_extents(transform, result_bbox)
                return if @bbox.empty?
                i = 0
                while i < 8
                    result_bbox.add(transform * @bbox.corner(i))
                    i += 1
                end
            end

            def draw(transform, view)
                @submeshes.each { |submesh|
                    view.drawing_color = submesh[:color]
                    view.draw(GL_TRIANGLES, submesh[:v].map { |v| transform * v })
                }
            end
        end

        class TeleportTool
            def initialize
                mydir = File.dirname(__FILE__)
                @cursor_id = UI.create_cursor(mydir + "/BeretGuy24.png", 16, 24)
                @loaded_models = {}
                @models_to_draw = []
                @extra_vlines = []
                
                @notes_bbox = nil
                @notes_version = nil

                @timer_count = 0
                #@mouse_loc = nil
                #@mouse_last_sent = nil
            end

            def guy_transform
                return @models_to_draw[0][1] if @models_to_draw.length > 0
                return nil
            end

            def activate
                @now_active = true
                @invalidate_requested = false

                model = Sketchup.active_model
                if (model != nil) && (model.active_view != nil) && (get_model_notes(model) != nil)
                    model.active_view.invalidate
                end
            end

            def deactivate(view)
                @now_active = false
                view.invalidate       # hide Beret Guy
            end

            def suspend(view)
                @now_active = false
                view.invalidate
            end

            def resume(view)
                @now_active = true
                @invalidate_requested = false

                if get_model_notes(view.model) != nil
                    view.invalidate
                end
            end

            def now_active
                @now_active
            end

            def update_position_to_display(cmd, model)
                @models_to_draw = []
                v_list = cmd["m"]
                i = 0
                while i < v_list.length
                    mdl_index = v_list[i]
                    tr = Geom::Transformation.new(v_list[i+1...i+17])
                    mdl = @loaded_models[mdl_index]
                    if mdl == nil
                        case mdl_index
                        when 0
                            mdl = MiniModel.new("Headset")
                        when 1
                            mdl = MiniModel.new("Body")
                        when 2
                            mdl = MiniModel.new("Teleport")
                        when 3
                            mdl = MiniModel.new("CtrlVive")
                        when 4
                            mdl = MiniModel.new("CtrlOculus")
                        else
                            mdl = MiniModel.new(nil)
                        end
                        @loaded_models[mdl_index] = mdl
                    end
                    @models_to_draw.push([mdl, tr])
                    i += 17
                end

                v_list = cmd["v"]
                v_list = [] if v_list == nil
                @extra_vlines = Baroque::VRSketch._vlist2ptarray(v_list)
                @timer_count += 1

                post_refresh(model)
            end

            #def report_mouse()
                # ------ CURRENTLY DISABLED, until we think of something nicer
                #if (@mouse_loc != @mouse_last_sent)
                #    @mouse_last_sent = @mouse_loc
                #    return if !Baroque::VRSketch.connected("Server")
                #
                #    msg = {
                #        cmd: "mouse",
                #    }
                #    if @mouse_loc != nil
                #        ray = pick_point(@mouse_loc[0], @mouse_loc[1], @mouse_loc[2])
                #        if ray != nil
                #            msg[:v] = ray[0].to_a + ray[1].to_a
                #        end
                #    end
                #    Baroque::VRSketch.model_tracker.emit(msg)
                #end
            #end

            def finish_now(model)
                if ((@models_to_draw.length > 0) || (@extra_vlines.length > 0))
                    @models_to_draw = []
                    @extra_vlines = []
                    post_refresh(model)
                end
            end

            def post_refresh(model)
                if (!@invalidate_requested) && (model != nil) && (model.active_view != nil)
                    @invalidate_requested = true
                    model.active_view.invalidate
                end
            end

            #def onMouseMove(flags, x, y, view)
            #    @mouse_loc = [x, y, view]
            #end

            #def onMouseLeave(view)
            #    @mouse_loc = nil
            #end

            def vector3mul(vector3, scale)
                return Geom::Vector3d.new(vector3.x * scale, vector3.y * scale, vector3.z * scale)
            end

            def pick_point(model, x, y, view)
                ray = view.pickray x, y
                target = model.raytest(ray)
                if target != nil
                    v = target[0]
                elsif ray[0].z > 0 && ray[1].z < 0
                    v = ray[0] - vector3mul(ray[1], ray[0].z / ray[1].z)
                else
                    return nil
                end
                return [ray[0], v]
            end

            def onLButtonDown(_flags, x, y, view)
                return if !Baroque::VRSketch.connected("Server")
                model = Sketchup.active_model
                return if model != Baroque::VRSketch.model_tracker.model

                ray = pick_point(model, x, y, view)
                if ray != nil
                    Baroque::VRSketch.model_tracker.emit({
                        cmd: "teleport",
                        v: ray[1].to_a
                    })
                end
            end

            def get_model_notes(model)
                return model.attribute_dictionary("Baroque_VRSketch_notes", false)
            end

            def getExtents
                bb = Geom::BoundingBox.new
                if @now_active
                    @models_to_draw.each { |mdltr|
                        mdltr[0].add_to_extents(mdltr[1], bb)
                    }
                end

                mtrack = Baroque::VRSketch.model_tracker
                if mtrack != nil
                    dict = get_model_notes(mtrack.model)
                    if dict != nil
                        version = dict["version"]
                        if version != @notes_version
                            @notes_version = version
                            @notes_bbox = Geom::BoundingBox.new
                            dict.each_pair { |key, v_list|
                                next if key[0] != "p"
                                @notes_bbox.add(Baroque::VRSketch._vlist2ptarray(v_list))
                            }
                        end
                        bb.add(@notes_bbox)
                    end
                end
                return bb
            end

            def draw(view)
                if @now_active
                    @models_to_draw.each { |mdltr|
                        mdltr[0].draw(mdltr[1], view)
                    }

                    if (@timer_count % 6) < 3
                        view.drawing_color = "#705000"
                    else
                        view.drawing_color = "#FFC0C0"
                    end
                    @extra_vlines.each { |pt1|
                        pt2 = Geom::Point3d.new(pt1.x, pt1.y, 99999)
                        view.draw_lines(pt1, pt2)
                    }
                end

                dict = get_model_notes(view.model)
                if dict != nil
                    view.line_width = 2.6
                    dict.each_pair { |key, v_list|
                        next if key[0] != "p"
                        color = dict["c" + key[1...key.length]]
                        color = Sketchup::Color.new(color)
                        view.drawing_color = color.blend(0xffffff, 0.7)
                        view.draw_polyline(Baroque::VRSketch._vlist2ptarray(v_list))
                    }
                    view.line_width = 1
                end

                @invalidate_requested = false
            end

            def onSetCursor
                return if !@cursor_id
                UI.set_cursor(@cursor_id)
            end
        end


        def delay(seconds, &block)
            done = false
            UI.start_timer(seconds, false) {
                next if done
                done = true
                Baroque::VRSketch.error_monitor.monitor {
                    block.call
                }
            }
        end


        # Here we add a menu item for the extension. Note that we again use a
        # load guard to prevent multiple menu items from accidentally being
        # created.
        if Sketchup.version_number < 1700000000
            UI.messagebox("VR Sketch does not work on SketchUp 2016 or older.  You need at least SketchUp 2017 (Make or Pro).")
        
        elsif (!file_loaded?(__FILE__))

            initialize

            # We fetch a reference to the top level menu we want to add to. Note that
            # we use "Plugins" here which was the old name of the "Extensions" menu.
            # By using "Plugins" you remain backwards compatible.
            menu = UI.menu('Plugins')

            # We add the menu item directly to the root of the menu in this example.
                # But if you plan to add multiple items per extension we recommend you
            # group them into a sub-menu in order to keep things organized.
            submenu = menu.add_submenu('VR Sketch')

            cmd = UI::Command.new("Send to VR on this PC") { Baroque::VRSketch.error_monitor.monitor {
				send_to_unity_vr(propose_tutorial: true)
			} }
            cmd.status_bar_text = "Start the Virtual Reality view"
            submenu.add_item(cmd)

            cmd = UI::Command.new("Send to VR on Oculus Quest...") { Baroque::VRSketch.error_monitor.monitor {
                send_to_unity_vr(propose_tutorial: true, on_quest: true)
			} }
            cmd.status_bar_text = "Send to an Oculus Quest"
            submenu.add_item(cmd)

            cmd = UI::Command.new("Center view") { Baroque::VRSketch.error_monitor.monitor { center_view } }
            cmd.status_bar_text = "Copy viewpoint from VR to SketchUp"
            cmd.set_validation_proc {
                if connected_and_can_center_view
                    MF_UNCHECKED
                else
                    MF_GRAYED
                end
            }
            submenu.add_item(cmd)

            cmd = UI::Command.new("Configuration...") {
                Baroque::VRSketch.error_monitor.monitor {
                    if (!show_config)
                        show_config   # re-open if closed
                    end
                }
            }
            cmd.status_bar_text = "Open the configuration dialog box for VR Sketch"
            submenu.add_item(cmd)

	        tutorialmenu = submenu.add_submenu('Tutorials')
			cmd = UI::Command.new("on this PC") {
				Baroque::VRSketch.error_monitor.monitor {
					start_tutorial(0, false, nil)
				}
			}
			tutorialmenu.add_item(cmd)
			cmd = UI::Command.new("on Oculus Quest") {
				Baroque::VRSketch.error_monitor.monitor {
					start_tutorial(0, true, nil)
				}
			}
			tutorialmenu.add_item(cmd)

            submenu.add_separator

            cmd = UI::Command.new("View a point cloud...") { Baroque::VRSketch.error_monitor.monitor {
                view_point_cloud }
            }
            cmd.status_bar_text = "View a point cloud (potree) on top of the current model"
            submenu.add_item(cmd)

            submenu.add_separator

            cmd = UI::Command.new("Open cloud page") { Baroque::VRSketch.error_monitor.monitor {
				open_cloud_page
			} }
			cmd.status_bar_text = "Open a browser at the URL showing your cloud files"
			submenu.add_item(cmd)

            cmd = UI::Command.new("Upload file to cloud...") { Baroque::VRSketch.error_monitor.monitor {
				upload_vrsketch_file
			} }
			cmd.status_bar_text = "Upload the current model to a file on the cloud"
			submenu.add_item(cmd)

            cmd = UI::Command.new("View cloud model...") { Baroque::VRSketch.error_monitor.monitor {
                view_cloud_model
			} }
            cmd.status_bar_text = "Download and view a cloud-based model"
            submenu.add_item(cmd)

			#cmd = UI::Command.new("Export .vrsketch file...") { Baroque::VRSketch.error_monitor.monitor {
			#	export_vrsketch_file
			#} }
			#cmd.status_bar_text = "Export the current model to a file in the VRSketch format"
			#submenu.add_item(cmd)

            #submenu.add_separator

            #cmd = UI::Command.new("Custom Sky image...") { Baroque::VRSketch.error_monitor.monitor { custom_sky_image } }
            #cmd.status_bar_text = "Cutomize the image used for the sky"
            #submenu.add_item(cmd)

            #cmd = UI::Command.new("Play a VR video...") { Baroque::VRSketch.error_monitor.monitor {
            #    play_vr_video }
            #}
            #cmd.status_bar_text = "Open a VR video file"
            #submenu.add_item(cmd)
            #
            #cmd = UI::Command.new("Record a VR video...") { Baroque::VRSketch.error_monitor.monitor {
            #    record_vr_video }
            #}
            #cmd.status_bar_text = "Save a VR video file"
            #submenu.add_item(cmd)

            submenu.add_separator

            submenu.add_item('Documentation') {
                Baroque::VRSketch.error_monitor.monitor {
                    show_documentation
                }
            }

            submenu.add_item("What's new") {
                Baroque::VRSketch.error_monitor.monitor {
                    show_whatsnew
                }
            }

            submenu.add_item('About') {
                Baroque::VRSketch.error_monitor.monitor {
                    show_about
                }
            }

            cmd = UI::Command.new("License...") {
                Baroque::VRSketch.error_monitor.monitor {
                    check_license(false) # populates the fields
                    show_license_dialog
                }
            }
            cmd.status_bar_text = "License details"
            submenu.add_item(cmd)

            create_toolbar
            create_menu_handler

            Sketchup.add_observer(MyAppObserver.new)

            file_loaded(__FILE__)
         end

    end # module VRSketch
end # module Baroque
